{************************************************}
{                                                }
{   DBGrid visual component                      }
{                                                }
{   Turbo Pascal 7                               }
{   FreePacal                                    }
{   Delphi 2 - 5                                 }
{   Virtual Pascal 2.1                           }
{                                                }
{   Copyright (c) 1995-2000 by Alexs             }
{                                                }
{************************************************}

{$IFDEF VER70}
{$O+}
{$ENDIF}

{$I Define.inc}

unit DBGrids;

interface

uses Objects, Drivers, Views, DataCtrl, DAE, Clases, Ext_Type, Register,
     Ext_Sys;

const
  CDBGrid       : ShortString = #1#2#3#5#6;
  CDataScroller : ShortString = #4#5#5;

const
  ufOneLine              = $01;
  ufTwoLine              = $02;
  ufThreeLine            = $04;
  ufWorkArea             = $08;
  ufView                 = $10;

type
{ TDataScroller object }

  { Palette layout }
  { 1 = Page areas }
  { 2 = Arrows }
  { 3 = Indicator }

  PDataScroller = ^TDataScroller;
  TDataScroller = object(TView)
    DataSourse:PDataSourse;
    PgStep: Integer;
    ArStep: Integer;
    Chars:TScrollChars;
    constructor Init(var Bounds: TRect; ADS:PDataSourse);
    constructor Load(var S: TStream);
    procedure Draw; virtual;
    procedure DrawPos(Pos: Integer);
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure ScrollDraw; virtual;
    procedure SetStep(APgStep, AArStep: Integer);
    procedure Store(var S: TStream);

    function GetPalette: PPalette; virtual;
    function GetPos: Integer;
    function GetSize: Integer;
    function ScrollStep(Part: Integer): Integer; virtual;
  end;

  { Palette layout }
  { 1 = Active }
  { 2 = Inactive }
  { 3 = Focused }
  { 4 = Selected }
  { 5 = Header }

type
  TDBGridOption = (dgEditing, dgAlwaysShowEditor, dgTitles,
                   dgIndicator, dgColumnResize, dgColLines, dgRowLines,
                   dgTabs, dgRowSelect, dgAlwaysShowSelection,
                   dgConfirmDelete, dgCancelOnExit, dgMultiSelect);
  TDBGridOptions = set of TDBGridOption;
type
   PDBGrid = ^TDBGrid;

   PDBGridField = ^TDBGridField;
   TDBGridField = object(TObject)
     Field:PField;
     Name:PString;
     FieldName:PString;
     Width:integer;
     Align:TAlign;
     ReadOnly:boolean;
     constructor Init(AOwner:PDBGrid; AField:PField);
     constructor Load(var S:TStream);
     procedure Store(var S:TStream);
     destructor Done;virtual;
     function GetText(MaxLen:Byte):string;
     function GetSize:integer;
     procedure SetWidth( AWidth:integer);
     procedure SetTitle( ATitle:string);
   private
     Owner:PDBGrid;
   end;

  TDBGridKeyEvent=procedure(Sender:PDBGrid; KeyCode: Word);

{ TDBGrid }
  TDBGrid = object(TView)
    DataScroller:PDataScroller;
    DataSourse:PDataSourse;
    FieldDefs:PCollection;
    ScrollBar: PScrollBar;
    TopItem: Longint;
    DBGridOptions: TDBGridOptions;
    Delta:TPoint;
    CurPos:TPoint;
    FixedColums:integer;
    constructor Init(var Bounds: TRect; AScrollBar :PScrollBar;
                         ADS:PDataSourse; ADataScroller:PDataScroller);
    constructor Load(var S: TStream);
    destructor Done;virtual;
    procedure ChangeBounds(var Bounds: TRect); virtual;
    procedure Draw; virtual;
    procedure DrawNilBase;
    procedure DrawHeader;
    procedure DrawTableOnly;
    procedure DrawLineOnly( Target:boolean);
    procedure FocusItem(Item: LongInt); virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure NewTable( ADS:PDataSourse);virtual;
    procedure SelectItem(Item: LongInt); virtual;
    procedure SetState(AState: Word; Enable: Boolean); virtual;
    procedure Store(var S: TStream);
    function  GetPalette: PPalette; virtual;
    procedure DeleteField(Name:String);
    procedure AtDelField(ANum:integer);
    function  CountFields:integer;
    procedure PrepareLine(var B:TDrawBuffer);
    procedure PrevRecord;
    procedure NextRecord;
    procedure PageDown;
    procedure PageUp;
    procedure ListHome;
    procedure ListEnd;
    procedure EndTable;
    procedure BeginTable;
    procedure RightCollum;
    procedure LeftCollum;
    procedure ConnectBase(ADS:PDataSourse);
    procedure Lock;
    procedure UnLock;
    procedure Update;
    procedure EditData;
    function CurentField:PDBGridField;
    procedure SetFixedColums(AFixedColums:integer);
  private
    NormalColor, SelectedColor, FocusedColor, Color, Header: Word;
    LockFlag:integer;
    EditPos:integer;
    EditDBPos:Longint;
    UpdateFlag:word;
    SelfEvent:boolean;
    procedure DoUpdate(w:word);
    procedure ScrollBarEvent;
    function TitleHight:integer;
  public
    OnColEnter:TNotifyEvent;
    OnColExit:TNotifyEvent;
    OnColumnMoved:TNotifyEvent;
    OnDblClick:TNotifyEvent;
    OnDragDrop:TNotifyEvent;
    OnDragOver:TNotifyEvent;
    OnEditButtonClick:TNotifyEvent;
    OnEndDrag:TNotifyEvent;
    OnEnter:TNotifyEvent;
    OnExit:TNotifyEvent;
    OnKeyDown:TDBGridKeyEvent;
    OnStartDrag:TNotifyEvent;
  end;

const
  RDBGrid: TStreamRec = (
     ObjType: rcDBGrid;
     VmtLink: Ofs(TypeOf(TDBGrid)^);
     Load:    @TDBGrid.Load;
     Store:   @TDBGrid.Store
  );
const
  RDBGridField: TStreamRec = (
     ObjType: rcDBGridField;
     VmtLink: Ofs(TypeOf(TDBGridField)^);
     Load:    @TDBGridField.Load;
     Store:   @TDBGridField.Store
  );
const
  RDataScroller: TStreamRec = (
     ObjType: rcDataScroller;
     VmtLink: Ofs(TypeOf(TDataScroller)^);
     Load:    @TDataScroller.Load;
     Store:   @TDataScroller.Store
  );


procedure RegisterDBGrid;
implementation
uses Ext_Math, Ext_Str, DBCtrls;

const
  MouseAutosToSkip = 1;
type
  PInplaceEdit = ^TInplaceEdit;
  TInplaceEdit = object(TDBEdit)
    ModalResult:integer;
    procedure HandleEvent(var Event:TEvent);virtual;
  end;

{ TDBGridField }
constructor TDBGridField.Init(AOwner:PDBGrid; AField:PField);
begin
  inherited Init;
  Owner:=AOwner;
  Field:=AField;
  Align:=Field^.Align;
  Name:=NewStr(Field^.Name^);
  FieldName:=NewStr(Field^.Name^);
  Width:=Field^.GetWidth;
  ReadOnly:=Field^.ReadOnly;
end;

destructor TDBGridField.Done;
begin
  DisposeStr(Name);
  DisposeStr(FieldName);
  inherited Done;
end;

constructor TDBGridField.Load(var S:TStream);
begin
  TObject.Init;
  Name:=S.ReadStr;
  FieldName:=S.ReadStr;
  S.Read(Width, SizeOf(Integer));
  S.Read(Align, SizeOf(TAlign));
  S.Read(ReadOnly, SizeOf(boolean));
end;

procedure TDBGridField.Store(var S:TStream);
begin
  S.WriteStr(Name);
  S.WriteStr(FieldName);
  S.Write(Width, SizeOf(Integer));
  S.Write(Align, SizeOf(TAlign));
  S.Write(ReadOnly, SizeOf(boolean));
end;

function TDBGridField.GetText(MaxLen:Byte):string;
begin
  if Field<>nil then
    GetText:=MakeAlign(Align, Field^.DisplayText, Min(GetSize,MaxLen))
  else GetText:='';
end;
{$IFNDEF VER70}
function TDBGridField.GetSize:integer;
begin
  if Width>Length(Name^) then GetSize:=Width
  else GetSize:=Length(Name^);
end;
{$ELSE}
function TDBGridField.GetSize:integer;assembler;
asm
        LES   DI,Self
        XOR   AX,AX
        MOV   BX,ES:[DI].TDBGridField.Width
        LES   DI,ES:[DI].Name
        MOV   AL,ES:[DI]
        CMP   AL,BL
        JA    @End
        MOV   AL,BL
@End:
end;
{$ENDIF}

procedure TDBGridField.SetWidth( AWidth:integer);
begin
  if (Name<>nil) and (Length(Name^)>AWidth) then Width:=Length(Name^)
  else Width:=AWidth;
end;

procedure TDBGridField.SetTitle( ATitle:string);
begin
   AssignStr(Name, ATitle);
   if Owner<>nil then Owner^.DoUpdate(ufView);
end;

{ TDataScroller }

constructor TDataScroller.Init(var Bounds: TRect; ADS:PDataSourse);
begin
  TView.Init(Bounds);
  EventMask := EventMask or evBroadcast;
  PgStep := 1;
  ArStep := 1;
  DataSOurse:=ADS;
  ComponentType:=ctDataScroller;
end;

constructor TDataScroller.Load(var S: TStream);
begin
  TView.Load(S);
  GetPeerViewPtr(S, DataSourse);
  S.Read(PgStep, SizeOf(Integer) * 2);
end;

procedure TDataScroller.Draw;
begin
  if Size.X = 1 then
  begin
    GrowMode := gfGrowLoX + gfGrowHiX + gfGrowHiY;
    Chars := VChars{$IFNDEF VER70}[true]{$ENDIF};
  end else
  begin
    GrowMode := gfGrowLoY + gfGrowHiX + gfGrowHiY;
    Chars := HChars{$IFNDEF VER70}[true]{$ENDIF};
  end;
  DrawPos(GetPos);
end;

procedure TDataScroller.DrawPos(Pos: Integer);
var
  S: Integer;
  B: TDrawBuffer;
begin
  S := GetSize - 1;
  MoveChar(B[0], Chars[0], GetColor(2), 1);
  MoveChar(B[1], Chars[2], GetColor(1), S - 1);
  MoveChar(B[Pos], Chars[3], GetColor(3), 1);
  MoveChar(B[S], Chars[1], GetColor(2), 1);
  WriteBuf(0, 0, Size.X, Size.Y, B);
end;

function TDataScroller.GetPalette: PPalette;
begin
  GetPalette := @CDataScroller;
end;

function TDataScroller.GetPos: Integer;
begin
  if (DataSourse=nil) or not DataSourse^.Active then  GetPos:= GetSize - 1
  else
  if DataSourse^.BOF then GetPos:=1
  else
  if DataSourse^.EOF then GetPos:= GetSize - 2
  else GetPos:=  GetSize  div 2;
end;

function TDataScroller.GetSize: Integer;
var
  S: Integer;
begin
  if Size.X = 1 then S := Size.Y else S := Size.X;
  if S < 3 then GetSize := 4 else GetSize := S;
end;

procedure TDataScroller.HandleEvent(var Event: TEvent);
var
  P, S, ClickPart: Integer;
  Mouse: TPoint;
  Extent: TRect;

function GetPartCode: Integer;
var
  Mark, Part: Integer;
begin
  Part := -1;
  if Extent.Contains(Mouse) then
  begin
    if Size.X = 1 then Mark := Mouse.Y else Mark := Mouse.X;
    if Mark = P then Part := sbIndicator else
    begin
      if Mark < 1 then Part := sbLeftArrow else
        if Mark < P then Part := sbPageLeft else
          if Mark < S then Part := sbPageRight else
            Part := sbRightArrow;
      if Size.X = 1 then Inc(Part, 4);
    end;
  end;
  GetPartCode := Part;
end;

procedure Clicked;
begin
  Message(Owner, evBroadcast, cmDataScrollerClicked, @Self);
end;

begin
  TView.HandleEvent(Event);
  if DataSourse=nil then exit;
  case Event.What of
    evMouseDown:
      begin
        Clicked;
        MakeLocal(Event.Where, Mouse);
        GetExtent(Extent);
        Extent.Grow(1, 1);
        P := GetPos;
        S := GetSize - 1;
        ClickPart := GetPartCode;
        if ClickPart <> sbIndicator then
        repeat
           MakeLocal(Event.Where, Mouse);
           if GetPartCode = ClickPart then
           begin
              DataSourse^.MoveBy( ScrollStep(ClickPart));
              ScrollDraw;
           end;
        until not MouseEvent(Event, evMouseAuto);
        ClearEvent(Event);
      end;
    evBroadcast:
      case Event.Command of
           cmUpDataDS:if Event.InfoPtr=DataSourse then DrawView;
           cmNewTable:if Event.InfoPtr=DataSourse then DataSourse:=Event.InfoPtr;
           cmDestroed:if Event.InfoPtr=DataSourse then DataSourse:=nil;
      end;
  end;
end;

procedure TDataScroller.ScrollDraw;
begin
  Message(Owner, evBroadcast, cmUpDataDS, DataSourse);
end;

function TDataScroller.ScrollStep(Part: Integer): Integer;
var
  Step: Integer;
begin
  if Part and 2 = 0 then Step := ArStep else Step := PgStep;
  if Part and 1 = 0 then ScrollStep := -Step else ScrollStep := Step;
end;

procedure TDataScroller.SetStep(APgStep, AArStep: Integer);
begin
  PgStep := APgStep;
  ArStep := AArStep;
end;

procedure TDataScroller.Store(var S: TStream);
begin
  TView.Store(S);
  PutPeerViewPtr(S, DataSourse);
  S.Write(PgStep, SizeOf(Integer) * 2);
end;

  { TInplaceEdit }
procedure TInplaceEdit.HandleEvent(var Event:TEvent);
procedure _Done(Code:integer);
begin
  ModalResult:=Code;
  EndModal(cmOk);
end;
begin
  inherited HandleEvent(Event);
  case Event.What of
    evKeyDown:
    case Event.KeyCode of
        kbUp:_Done(-1);
        kbDown:_Done(1);
        kbPgDn:_Done(2);
        kbPgUp:_Done(-2);
    else
        exit;
    end;
  else
    exit;
  end;
  ClearEvent(Event);
end;

{ TDBGrid }

constructor TDBGrid.Init(var Bounds: TRect; AScrollBar :PScrollBar;
                         ADS:PDataSourse; ADataScroller:PDataScroller);
begin
  TView.Init(Bounds);
  FieldDefs:=New(PCollection, Init(10,10));
  Options := Options or (ofFirstClick + ofSelectable);
  GrowMode := GrowMode or gfGrowHiX or gfGrowHiY;
  EventMask := EventMask or evBroadcast;
  DBGridOptions:=[dgTitles, dgColumnResize];
  if ADataScroller<>nil then ADataScroller^.SetStep( Size.Y-2, 1);
  ScrollBar := AScrollBar;
  DataScroller := ADataScroller;
  EditPos:=-1;
  NewTable(ADS);
  if ScrollBar<>nil then ScrollBar^.SetParams(0, 0, CountFields-1, 5, 1);
  ComponentType:=ctDBGrid;
end;

destructor TDBGrid.Done;
begin
  NewTable(nil);
  Dispose(FieldDefs,Done);
  inherited Done;
end;

constructor TDBGrid.Load(var S: TStream);
procedure DoAssign(F:PDBGridField);far;
begin
  F^.Owner:=@Self;
end;
begin
  inherited Load(S);
  GetPeerViewPtr(S, ScrollBar);
  GetPeerViewPtr(S, DataScroller);
  GetPeerViewPtr(S, DataSourse);
  S.Read(DBGridOptions, SizeOf(TDBGridOptions));
  S.Read(TopItem, SizeOf(Longint));
  FieldDefs:=PCollection(S.Get);
  LoadProcedure(S,@OnColEnter);
  LoadProcedure(S,@OnColExit);
  LoadProcedure(S,@OnColumnMoved);
  LoadProcedure(S,@OnDblClick);
  LoadProcedure(S,@OnDragDrop);
  LoadProcedure(S,@OnDragOver);
  LoadProcedure(S,@OnEditButtonClick);
  LoadProcedure(S,@OnEndDrag);
  LoadProcedure(S,@OnEnter);
  LoadProcedure(S,@OnExit);
  LoadProcedure(S,@OnKeyDown);
  LoadProcedure(S,@OnStartDrag);
  if FieldDefs<>nil then FieldDefs^.ForEach(@DoAssign);
end;

procedure TDBGrid.Store(var S: TStream);
begin
  inherited Store(S);
  PutPeerViewPtr(S, ScrollBar);
  PutPeerViewPtr(S, DataScroller);
  PutPeerViewPtr(S, DataSourse);
  S.Write(DBGridOptions, SizeOf(TDBGridOptions));
  S.Write(TopItem, SizeOf(Longint));
  S.Put(FieldDefs);
  StoreProcedure(S,@OnColEnter);
  StoreProcedure(S,@OnColExit);
  StoreProcedure(S,@OnColumnMoved);
  StoreProcedure(S,@OnDblClick);
  StoreProcedure(S,@OnDragDrop);
  StoreProcedure(S,@OnDragOver);
  StoreProcedure(S,@OnEditButtonClick);
  StoreProcedure(S,@OnEndDrag);
  StoreProcedure(S,@OnEnter);
  StoreProcedure(S,@OnExit);
  StoreProcedure(S,@OnKeyDown);
  StoreProcedure(S,@OnStartDrag);
end;

procedure TDBGrid.Lock;
begin
  inc(LockFlag);
end;

procedure TDBGrid.UnLock;
begin
  Dec(LockFlag);
  if UpdateFlag<>0 then Update
end;

function TDBGrid.TitleHight:integer;
begin
  TitleHight:=byte(dgTitles in DBGridOptions);
end;

procedure TDBGrid.DeleteField(Name:String);
var P:PDBGridField;
    i:integer;
    PP:Pointer;
begin
  Name:=stUpCase(Name);
  for i:=0 to FieldDefs^.Count-1 do
  begin
      P:=FieldDefs^.At(i);
      if stUpCase(P^.Name^)=Name then
      begin
        FieldDefs^.AtFree(i);
        DoUpdate(ufView);
        Exit;
      end;
  end;
  PP:=@Name;
  DBError(edbFieldNotFound, PP);
end;

procedure TDBGrid.AtDelField(ANum:integer);
begin
  if (ANum>=0) or (ANum<FieldDefs^.Count) then FieldDefs^.AtFree(ANum)
  else SingleError(edbFieldNotFound);
end;


function TDBGrid.CountFields:integer;
begin
  CountFields:=FieldDefs^.Count;
end;

procedure TDBGrid.NewTable( ADS:PDataSourse);
procedure DoNew(P:PField);far;
begin
   FieldDefs^.Insert(New(PDBGridField, Init(@Self, P)));
end;
procedure ReReadTable(P:PDBGridField);far;
begin
  P^.Field:=ADS^.FieldByName(P^.FieldName^);
end;
begin
  if (ADS = nil) then FieldDefs^.FreeAll
  else
  begin
    if (ADS<>DataSourse) then
    begin
      FieldDefs^.FreeAll;
      if ADS^.Active then ADS^.FieldDefs^.ForEach(@DoNew);
    end
    else
    if ADS^.Active then
      if FieldDefs^.Count=0 then ADS^.FieldDefs^.ForEach(@DoNew)
      else FieldDefs^.ForEach(@ReReadTable);
    if ADS^.Active then
    begin
      TopItem:=ADS^.RecNo;
    end;
  end;
  DataSourse:=ADS;
  DrawView;
end;

procedure TDBGrid.ChangeBounds(var Bounds: TRect);
begin
  TView.ChangeBounds(Bounds);
end;

procedure TDBGrid.PrepareLine(var B:TDrawBuffer);
var
  P:PDBGridField;
  I,J,k:integer;
  PrewCOlor:word;
begin
  i:=Delta.X+1;
  MoveChar(B, ' ', NormalColor, Size.X);
  P:=FieldDefs^.At(Delta.X);
  j:=Min(Size.X, P^.GetSize);
  PrewCOlor:=NormalColor;
  if Delta.X+CurPos.X+1=I then
  begin
    MoveChar(B, ' ', Color, J);
    MoveStr(B, P^.GetText(Size.X), Color)
  end
  else
  begin
    if Delta.X<FixedColums then
    begin
      MoveStr(B, P^.GetText(Size.X), Header);
      PrewCOlor:=Header;
    end
    else MoveStr(B, P^.GetText(Size.X), NormalColor);
  end;
  while (i<CountFields) and (j<Size.X) do
  begin
     P:=FieldDefs^.At(i);
     MoveChar(B[j], '', PrewCOlor, 1);
     K:=P^.GetSize;
     if K+J+1>Size.X then K:=Size.X-J-1;
     if I=Delta.X+CurPos.X then
     begin
       MoveChar(B[j+1], ' ', Color, k);
       MoveStr(B[j+1], P^.GetText(k), Color);
       PrewColor:=NormalColor;
     end
     else
     begin
       if i<FixedColums then
       begin
{         MoveStr(B, P^.GetText(Size.X), Header)}
         MoveChar(B[j+1], ' ', Header, k);
         MoveStr(B[j+1], P^.GetText(k), Header);
       end
       else
       begin
         PrewCOlor:=NormalColor;
         MoveChar(B[j+1], ' ', NormalColor, k);
         MoveStr(B[j+1], P^.GetText(k), NormalColor);
       end;
     end;
     inc( j, k+1);
     inc(i);
  end;
end;

procedure TDBGrid.DrawNilBase;
var
  B: TDrawBuffer;
begin
  Color:=GetColor(5);
  MoveChar(B, ' ', Color, Size.X);
  MoveStr(B, 'DBGrid', Color);
  WriteLine(0, 0, Size.X, 1, B);
  MoveChar(B, ' ', GetColor(1), Size.X);
  WriteLine(0, 1, Size.X, Size.Y, B);
end;

procedure TDBGrid.DrawHeader;
var
  I,j: Integer;
  B: TDrawBuffer;
  P:PDBGridField;
begin
  if State and (sfSelected + sfActive) = (sfSelected + sfActive) then
      Header := GetColor(2) else Header := GetColor(1);
  MoveChar(B, ' ', Header, Size.X);
  i:=Delta.X;
  P:=FieldDefs^.At(i);
  MoveStr(B[0], P^.Name^ , Header); j:=P^.GetSize;
  inc(i);
  while (i<=CountFields-1) and (j<Size.X) do
  begin
    P:=FieldDefs^.At(i);
    MoveStr(B[j], #179+P^.Name^ , Header);
    inc( j, P^.GetSize+1);
    inc(i);
  end;
  WriteLine(0, 0, Size.X, 1, B);
end;

procedure TDBGrid.Draw;
begin
  if LockFlag<>0 then exit;
  NormalColor := GetColor(5);
  if State and (sfSelected + sfActive) = (sfSelected + sfActive) then
      Header := GetColor(2) else Header := GetColor(1);
  if State and (sfSelected + sfActive) = (sfSelected + sfActive) then SelectedColor := GetColor(4)
  else  SelectedColor := GetColor(3);
  if (DataSourse=nil) or (not DataSourse^.Active) or (FieldDefs^.Count=0) then
     DrawNilBase
  else
  begin
    if dgTitles in DBGridOptions then DrawHeader;
    DrawTableOnly;
  end;
end;

procedure TDBGrid.DrawTableOnly;
var
  Item: Integer;
  B: TDrawBuffer;
  CurentPos:Longint;
begin
  Owner^.Lock;
  Item:=TitleHight;
  CurentPos:=DataSourse^.RecNo;
  DataSourse^.Go( TopItem);
  while (not DataSourse^.Eof) and (Item<>Size.Y) do
  begin
    if DataSourse^.RecNo=CurentPos then Color:=SelectedColor
    else Color:=NormalColor;
    PrepareLine( B);
    WriteLine(0, Item, Size.X, 1, B);
    DataSourse^.NextRecEx;
    inc(Item);
  end;
  if Item<>Size.Y then
  begin
    MoveChar(B, ' ', NormalColor, Size.X);
    WriteLine(0, Item, Size.X, Size.Y, B);
  end;
  DataSourse^.Go( CurentPos);
  Owner^.UnLock;
end;

procedure TDBGrid.DrawLineOnly( Target:boolean);
var
  B: TDrawBuffer;
  L:longint;
begin
  Owner^.Lock;
  L:=DataSourse^.RecNo;
  if Target then
  begin
    Color:=SelectedColor;
    PrepareLine( B);
    WriteLine(0, CurPos.Y+TitleHight, Size.X, 1, B);
    DataSourse^.NextRecEx;
    Color:=NormalColor;
    PrepareLine( B);
    WriteLine(0, CurPos.Y+1+TitleHight, Size.X, 1, B);
  end
  else
  begin
    Color:=SelectedColor;
    PrepareLine( B);
    WriteLine(0, CurPos.Y+TitleHight, Size.X, 1, B);
    DataSourse^.PriorRecEx;
    Color:=NormalColor;
    PrepareLine( B);
    WriteLine(0, CurPos.Y-1+TitleHight, Size.X, 1, B);
  end;
  DataSourse^.Go(L);
  Owner^.UnLock;
end;

function TDBGrid.GetPalette: PPalette;
begin
  GetPalette := @CDBGrid;
end;

procedure TDBGrid.HandleEvent(var Event: TEvent);
var
  Mouse: TPoint;
  Count: Word;

procedure DragCollumns;
var
  P:PDBGridField;
  I,J:integer;
  OldSize, OldPos, OP:integer;
begin
  if (Mouse.X<0) or (Mouse.X>Size.X) then exit;
  i:=Delta.X+1;
  P:=FieldDefs^.At(Delta.X);
  j:=P^.GetSize;
  while (i<CountFields) and (j<Mouse.X) do
  begin
     P:=FieldDefs^.At(i);
     inc( j, P^.GetSize+1);
     inc(i);
  end;
  if J<>Mouse.X then exit;
  Dec(i);
  P:=FieldDefs^.At(i);
  OldSize:=P^.GetSize;
  OldPos:=Mouse.X;
  OP:=Mouse.X;
  repeat
    if OldPos<>Mouse.X then
    begin
      P^.SetWidth(Mouse.X-OP+OldSize);
      DrawView;
    end;
    MakeLocal(Event.Where, Mouse);
  until not MouseEvent(Event, evMouseMove + evMouseAuto);
  ClearEvent(Event);
end;

procedure DBGridMouseEvent;
var
  L,i,k:integer;
  NewItem, OldItem: Integer;
begin
    OldItem := CurPos.Y;
    K:=CurPos.X;
    MakeLocal(Event.Where, Mouse);
    if (MouseInView(Event.Where)) and ((Mouse.Y<>0) or
                            not (dgTitles in DBGridOptions)) then
    begin
       NewItem := Mouse.Y - TitleHight;
       i:=Delta.X;
       L:=0;
       CurPos.X:=-1;
       repeat
         L:=L+PDBGridField(FieldDefs^.At(i))^.GetSize+1;
         inc(CurPos.X);
         inc(i);
       until (l>Mouse.X) or (i>FieldDefs^.Count-1);
       CurPos.Y:=NewItem;
    end;
    Count := 0;
    if (Mouse.Y=0) and (dgColumnResize in DBGridOptions) then DragCollumns
    else
  if Event.Double then
  begin
    Lock;
    if (dgEditing in DBGridOptions) and
         not (CurentField^.Field^.FieldType in
             [ftBCD, ftBytes, ftVarBytes, ftAutoInc,
              ftBlob, ftMemo, ftGraphic, ftFmtMemo, ftParadoxOle,
              ftDBaseOle, ftTypedBinary]) then
      EditData
    else
      FocusItem(DataSourse^.RecNo);
    UnLock;
  end
  else
    repeat
      if (NewItem <> OldItem) or (K<>CurPos.X) then
      begin
        if (NewItem <> OldItem) then
        begin
          Lock;
          DataSourse^.Go(TopItem);
          while (NewItem>0) and not DataSourse^.EOF do
          begin
            DataSourse^.NextRecEx;
            Dec(NewItem);
          end;
          SelectItem(DataSourse^.RecNo);
          DataSourse^.UpData;
          DoUpdate(ufView);
          UnLock;
        end;
      end;
      MakeLocal(Event.Where, Mouse);
      OldItem := CurPos.Y;
      if (MouseInView(Event.Where)) and ((Mouse.Y<>0) or
                                    not (dgTitles in DBGridOptions)) then
      begin
        NewItem := Mouse.Y - TitleHight;
        CurPos.Y:=NewItem;
        i:=Delta.X;
        L:=0;
        CurPos.X:=-1;
        repeat
          L:=L+PDBGridField(FieldDefs^.At(i))^.GetSize+1;
          inc(CurPos.X);
          inc(i);
        until (l>Mouse.X) or (i>FieldDefs^.Count-1);
      end
      else
      begin
       if Event.What = evMouseAuto then Inc(Count);
          if Count = MouseAutosToSkip then
          begin
            Lock;
            Count := 0;
            if Mouse.Y < byte(dgColumnResize in DBGridOptions) then PrevRecord
            else
            if Mouse.Y >= Size.Y then NextRecord
            else
               if (Mouse.X < 0) then LeftCollum
               else
               if (Mouse.X>=Size.X) then RightCollum;
            DataSourse^.UpData;
            UnLock;
          end;
      end;
      SelectItem(DataSourse^.RecNo);
    until not MouseEvent(Event, evMouseMove + evMouseAuto);
end;

procedure ReReadTable(P:PDBGridField);far;
begin
  P^.Field:=DataSourse^.FieldByName(P^.FieldName^);
end;

procedure UpDateDBGrid;
begin
  if LockFlag<>0 then exit;
  TopItem:=DataSourse^.RecNo;
  CurPos.Y:=0;
  DoUpdate(ufView);
end;

procedure DoDragColumn(Forse:boolean);
var
  P:PDBGridField;
begin
  if not (dgColumnResize in DBGridOptions) then exit;
  P:=FieldDefs^.At(Delta.X+CurPos.X);
  if Forse then P^.SetWidth(P^.GetSize+1)
  else P^.SetWidth(P^.GetSize-1);
  DoUpdate(ufView);
end;

procedure DoPostDelete(Direct:boolean);
var i:integer;
    Pos:Longint;
begin
  if (Event.InfoPtr=DataSourse) and (DataSourse^.HasSorted or Direct) then
  begin
    Pos:=DataSourse^.RecNo;
    i:=TitleHight-1;
    while (i<CurPos.Y) and not DataSourse^.Bof do
    begin
      DataSourse^.PriorRecEx;
      inc(i);
    end;
    CurPos.Y:=i;
    TopItem:=DataSourse^.RecNo;
    DataSourse^.Go(Pos);
    DoUpdate(ufView);
  end;
end;

begin
  TView.HandleEvent(Event);
  if (Event.What=evBroadcast) and (Event.Command=cmDestroed) then
    if Event.InfoPtr=DataSourse then
    begin
       NewTable(nil);
       exit;
     end
     else if Event.InfoPtr=DataScroller then DataScroller:=nil
     else if Event.InfoPtr=ScrollBar then ScrollBar:=nil;
  if (DataSourse = nil) then exit;
  if not DataSourse^.Active then exit;
  case Event.What of
  evMouseDown:
         DBGridMouseEvent;
  evCommand:
    case Event.Command of
      cmDBGridFirst:BeginTable;
      cmDBGridLast:EndTable;
      cmDBGridNext:NextRecord;
      cmDBGridPrev:PrevRecord;
    else
      exit;
    end;
  evKeyDown:begin
    if Assigned(OnKeyDown) then OnKeyDown(@Self, Event.KeyCode);
    case CtrlToArrow(Event.KeyCode) of
      kbUp:PrevRecord;
      kbDown:NextRecord;
      kbPgUp:PageUp;
      kbPgDn:PageDown;
      kbHome:ListHome;
      kbEnd: ListEnd;
      kbCtrlPgDn: EndTable;
      kbCtrlPgUp:BeginTable;
      kbRight:RightCollum;
      kbLeft: LeftCollum;
      kbCtrlRight:DoDragColumn(true);
      kbCtrlLeft: DoDragColumn(false);
    else
      if (dgEditing in DBGridOptions) and
         not (CurentField^.Field^.FieldType in
                  [ftBCD, ftBytes, ftVarBytes, ftAutoInc,
                  ftBlob, ftMemo, ftGraphic, ftFmtMemo, ftParadoxOle,
                  ftDBaseOle, ftTypedBinary]) then
      begin
        if not CurentField^.ReadOnly then
        begin
          if (Event.KeyCode=kbEnter) then EditData
          else
          if Event.CharCode>#31 then
          begin
            PutEvent(Event);
            EditData;
          end else exit;
        end else exit;
      end
      else
      if (Event.CharCode = ' ') then FocusItem(DataSourse^.RecNo) else Exit;
    end;
    SelectItem(DataSourse^.RecNo);
    end;
  evBroadcast:
     begin
       case Event.Command of
           cmNewTable:if Event.InfoPtr=DataSourse then NewTable(Event.InfoPtr);
           cmScrollBarClicked: if Event.InfoPtr=ScrollBar then Select;
           cmDataScrollerClicked: if Event.InfoPtr = DataScroller then Select;
           cmScrollBarChanged: if (ScrollBar=Event.InfoPtr) and not SelfEvent then
                        ScrollBarEvent;
           cmUpDataDS:
             if (Event.InfoPtr=DataSourse) then UpDateDBGrid;
           cmAfterPost:DoPostDelete(true);
           {cmAfterDelete:DoPostDelete(DataSourse^.FirstRecEx, false);}
           cmAfterDelete:DoPostDelete(true);
       end;
       exit;
     end;
  else
    Exit;
  end;
  Lock;
  DataSourse^.UpData;
  UnLock;
  ClearEvent(Event);
end;

procedure TDBGrid.ScrollBarEvent;
var i,l,ii:integer;
begin
  if ScrollBar=nil then exit;
  i:=ScrollBar^.Value-(CurPos.X+Delta.X);
  if ((CurPos.X+Delta.X)+i<FieldDefs^.Count) and
     ((CurPos.X+Delta.X)+i>-1) then
  begin
    if Assigned(OnColExit) then OnColExit(@Self);
    if i>0 then
    begin
      L:=0;
      for ii:=Delta.X+CurPos.X+I downto Delta.X do
      begin
        l:=l+PDBGridField(FieldDefs^.At(ii))^.GetSize+1;
        if L>Size.X then break;
      end;
      if L>Size.X then
      begin
        ii:=ii+1;
        CurPos.X:=(Delta.X+CurPos.X+I)-II;
        Delta.X:=ii;
        DoUpdate(ufView);
      end
      else
      begin
        inc(CurPos.X,i);
        DoUpdate(ufOneLine);
      end;
    end
    else
    begin
      if (I<CurPos.X) and (CurPos.X+I>0) then
      begin
        CurPos.X:=CurPos.X+i;
        DoUpdate(ufOneLine);
      end
      else
      begin
        i:=i+CurPos.X;
        CurPos.X:=0;
        Delta.X:=Delta.X+i;
        DoUpdate(ufView);
      end;
    end;
    if Assigned(OnColEnter) then OnColEnter(@Self);
  end;
end;

procedure TDBGrid.RightCollum;
var
  i,l:integer;
begin
 if (CurPos.X+Delta.X)<FieldDefs^.Count-1 then
 begin
   if Assigned(OnColExit) then OnColExit(@Self);
   L:=0;
   for i:=Delta.X+CurPos.X+1 downto Delta.X do
   begin
     l:=l+PDBGridField(FieldDefs^.At(i))^.GetSize+1;
     if L>Size.X then break;
   end;
   if L>Size.X then
   begin
     if I=Delta.X+CurPos.X+1 then
     begin
       Delta.X:=Delta.X+CurPos.X+1;
       CurPos.X:=0;
     end
     else
     begin
       CurPos.X:=(Delta.X+CurPos.X)-I;
       Delta.X:=I+1;
     end;
     DoUpdate(ufView);
   end
   else
   begin
     inc(CurPos.X);
     DoUpdate(ufOneLine);
   end;
   if Assigned(OnColEnter) then OnColEnter(@Self);
 end;
end;

procedure TDBGrid.SetFixedColums(AFixedColums:integer);
begin
  if AFixedColums>FieldDefs^.Count-1 then exit;
  FixedColums:=AFixedColums;
  if CurPos.X<FixedColums then CurPos.X:=FixedColums;
end;

procedure TDBGrid.LeftCollum;
begin
  if Assigned(OnColExit) then OnColExit(@Self);
  if CurPos.X>FixedColums then
  begin
    Dec(CurPos.X);
    DoUpdate(ufOneLine);
  end
  else
  if Delta.X>0 then
  begin
    Dec(Delta.X);
    DoUpdate(ufView);
  end
  else Exit;
  if Assigned(OnColEnter) then OnColEnter(@Self);
end;

procedure TDBGrid.ListHome;
begin
  DataSourse^.Go( TopItem);
  CurPos.Y:=0;
  DoUpdate(ufWorkArea);
end;

procedure TDBGrid.ListEnd;
begin
  while CurPos.Y<Size.Y-TitleHight-1 do
  begin
    DataSourse^.NextRecEx;
    if DataSourse^.Eof then break;
    inc(CurPos.Y);
  end;
  DoUpdate(ufWorkArea);
end;

procedure TDBGrid.EndTable;
begin
  DataSourse^.LastRecEx;
  DataSourse^.MoveBy(-(Size.Y-TitleHight-1));
  TopItem:=DataSourse^.RecNo;
  CurPos.Y:=0;
  while CurPos.Y<Size.Y-TitleHight-1 do
  begin
    DataSourse^.NextRecEx;
    if DataSourse^.Eof then break;
    inc(CurPos.Y);
  end;
  DoUpdate(ufWorkArea);
end;

procedure TDBGrid.BeginTable;
begin
  TopItem:=DataSourse^.FirstRecEx;
  CurPos.Y:=0;
  DoUpdate(ufWorkArea);
end;

procedure TDBGrid.PageDown;
var I:integer;
begin
  TopItem:=DataSourse^.RecNo;
  CurPos.Y:=0;
  for i:=0 to Size.Y-3 do
  begin
    DataSourse^.NextRecEx;
    if DataSourse^.Eof then break;
    inc(CurPos.Y);
  end;
  DoUpdate(ufWorkArea);
end;

procedure TDBGrid.PrevRecord;
begin
  DataSourse^.PriorRecEx;
  if DataSourse^.BOF then exit;
  Dec(CurPos.Y);
  if CurPos.Y<0 then
  begin
    TopItem:=DataSourse^.RecNo;
    Inc(CurPos.Y);
    DrawTableOnly;
  end
  else DrawLineOnly(true);
end;

procedure TDBGrid.NextRecord;
var C:longint;
begin
  DataSourse^.NextRecEx;
  if DataSourse^.EOF then exit;
  inc(CurPos.Y);
  if CurPos.Y=Size.Y-1 then
  begin
    C:=DataSourse^.RecNo;
    DataSourse^.Go( TopItem);
    DataSourse^.NextRecEx;
    TopItem:=DataSourse^.RecNo;
    DataSourse^.Go(c);
    Dec(CurPos.Y);
    DrawTableOnly;
  end
  else DrawLineOnly(false);
end;

procedure TDBGrid.PageUp;
var I:integer;
begin
  for i:=0 to Size.Y-3 do
  begin
    DataSourse^.PriorRecEx;
    if DataSourse^.Bof then break;
  end;
  TopItem:=DataSourse^.RecNo;
  CurPos.Y:=0;
  DoUpdate(ufWorkArea);
end;

procedure TDBGrid.SelectItem(Item: LongInt);
begin
end;

procedure TDBGrid.FocusItem(Item: LongInt);
begin
  Message(Owner, evBroadcast, cmDBGridSelect, @Self);
end;


procedure TDBGrid.SetState(AState: Word; Enable: Boolean);

procedure ShowSBar(SBar: PView);
begin
  if (SBar <> nil) then
    if GetState(sfActive) and GetState(sfVisible) then SBar^.Show
    else  SBar^.Hide;
end;

begin
  TView.SetState(AState, Enable);
  if AState and (sfSelected + sfActive + sfVisible) <> 0 then
  begin
    ShowSBar(ScrollBar);
    ShowSBar(DataScroller);
    DrawView;
  end;
end;

procedure TDBGrid.DoUpdate(w:word);
begin
  UpdateFlag:=UpdateFlag or W;
  if LockFlag=0 then Update;
end;

procedure TDBGrid.Update;
procedure DrawLine;
var
  B: TDrawBuffer;
begin
  Color:=SelectedColor;
  DataSourse^.Go(DataSourse^.RecNo);
  PrepareLine( B);
  WriteLine(0, CurPos.Y+1, Size.X, 1, B);
end;

begin
  if LockFlag>0 then exit;
  if UpdateFlag and ufView <> 0 then
     DrawView
  else
  if UpdateFlag and ufWorkArea <> 0 then
     DrawTableOnly
  else
  if UpdateFlag and ufThreeLine <> 0 then
     DrawView
  else
  if UpdateFlag and ufTwoLine <> 0 then
     DrawView
  else
  if UpdateFlag and ufOneLine <> 0 then
     DrawLine;
  SelfEvent:=true;
  if ScrollBar<>nil then
  begin
    ScrollBar^.SetValue(CurPos.X+Delta.X);
    if ScrollBar^.Max<>CountFields-1 then
       ScrollBar^.SetRange( 0, CountFields-1);
  end;
  SelfEvent:=false;
  UpdateFlag:=0;
end;

procedure TDBGrid.ConnectBase(ADS:PDataSourse);
procedure ReReadTable(P:PDBGridField);
begin
  P^.Field:=ADS^.FieldByName(P^.FieldName^);
end;
var i:integer;
begin
  DataSourse:=ADS;
  if (ADS<>nil) and (FieldDefs<>nil) then
  begin
     for i:=0 to FieldDefs^.Count-1 do ReReadTable(FieldDefs^.At(i));
     TopItem:=DataSourse^.RecNo;
  end;
end;

function TDBGrid.CurentField:PDBGridField;
begin
  CurentField:=FieldDefs^.At(Delta.X+CurPos.X);
end;

procedure TDBGrid.EditData;
var Edit:PInplaceEdit;
    R:TRect;
    i,j:integer;
begin
  if State and sfDisabled <> 0 then exit;
  if (not DataSourse^.ReadOnly) then
  begin
    if CurentField^.ReadOnly then exit;
    j:=0;
    for i:=Delta.X to Delta.X+CurPos.X-1 do
      inc(j,PDBGridField(FieldDefs^.At(i))^.GetSize+1);
    R.Assign(j,CurPos.Y+1,Min(J+PDBGridField(FieldDefs^.At(Delta.X+CurPos.X))^.GetSize,Size.X),CurPos.Y+2);
    R.Move(Origin.X, Origin.Y);
    Lock;
    Edit:=New(PInplaceEdit, Init(R, DataSourse, CurentField^.Field^.Name^));
    DataSourse^.Edit;
    if Owner^.ExecView(Edit)<>cmCancel then
    begin
      DataSourse^.Post;
      case Edit^.ModalResult of
        -1:PrevRecord;
         1:NextRecord;
      end;
    end
    else DataSourse^.Cancel;
    Edit^.Free;
    DoUpdate(ufWorkArea);
    UnLock;
  end
  else
    DBError(edbDataSetReadOnly, DataSourse^.TableName);
end;

procedure RegisterDBGrid;
begin
   RegisterType(RDBGrid);
   RegisterType(RDBGridField);
   RegisterType(RDataScroller);
end;

end.
