{************************************************}
{                                                }
{   Turbo Pascal 7.0                             }
{   FreePacal                                    }
{   Delphi 4                                     }
{   Data base unit                               }
{   Copyright (c) 1995,99 by Alexs               }
{                                                }
{************************************************}

{$IFDEF VER70}
{$O-}
{$ENDIF}
{$I Define.inc}

unit DBFTable;

interface

uses Objects, Dos, DAE, Ext_Dos, Ext_Type, Clases, DAE_Type, Parser;

const
  MaxFields      = 250;
const
  LockFieldName  = '_DBASELOCK';
  sDbfExt        = '.DBF';
type
  PFieldRecord = ^TFieldRecord;
  TFieldRecord = record
    Name:PString;
    fType:char;
    fLength:byte;
    fDecimal:byte;
    fOffset:word;
  end;
  PFieldArray = ^TFieldArray;
  TFieldArray = array [0..65520 div SizeOf(TFieldRecord)] of TFieldRecord;

  PDBaseHeader = ^TDBaseHeader;
  TDBaseHeader = record {this structure used only in memory}
    Version:byte;
    LastUpdate:TDateTime;
    RecordCount:Longint;
    HeaderSize:word;
    RecordSize:word;
    FieldCount:word;
    MemoFile:boolean;
    LanguageDriver:word;
    MDX:boolean;
    isDataBase:boolean;
    TableLanguage:PString;
    Fields:PFieldArray;
    DBCName:PString; {for VFP files}
  end;

type
  T_WriteEvent = procedure(S:PStream; Header:PDBaseHeader; WriteFields:boolean);
  TReadEvent = procedure (S:PStream; var Header:PDBaseHeader);

type
   PDBFStream = ^TDBFStream;
   TDBFStream = object(TDBStream)
     FileHeader : PDBaseHeader;
     constructor Init(AFileName:FNameStr; AMode:word);
     constructor Create(AFileName:FNameStr; AFileHeader : PDBaseHeader);
     destructor Done;virtual;
     procedure CopyStructure(S:PStream);virtual;
     procedure ReadHeader;
     procedure WriteHeader;
     procedure WriteFields;
   private
     OnRead:TReadEvent;
     OnWrite:T_WriteEvent;
     procedure AssignDrivers(vers:byte);
   end;

type
   PDBFTable = ^TDBFTable;
   TDBFTable = object(TDataSet)
      FileHeader :PDBaseHeader;
      FreePageDbt:Longint;
      DbfFile    :PDBFStream;
      Options    :word;
      DispDeleted:boolean;

      constructor Init(AFileName:FNameStr;AMode:word);
      constructor Create(AFileName:FNameStr;
                            var  AFieldDefs: array of TFieldRecord;
                            fFieldCount:integer;
                            DbfTableType:byte; SubType:byte);
      destructor  Done;virtual;
      function Open:boolean;virtual;
      procedure Close;virtual;
      procedure CreateIndex(AIndexName:FNameStr; AField:string; iOpt:TIndexOptions);virtual;
      procedure OpenIndex(AIndexName:FNameStr);virtual;
      procedure AddIndex( AIndex:PIndexObject);
      procedure SetIndex(IndexName:string);

      procedure Delete;virtual;
      procedure Recall;virtual;
      function DeletedRec:boolean;virtual;
      function SetDeleted(Del:boolean):Boolean;

      procedure Go(ARecNo:longint);virtual;
      procedure MoveBy(Distance:longint);virtual;
      procedure First;virtual;
      procedure Prior;virtual;
      procedure Next;virtual;
      procedure Last;virtual;
      procedure FreeBookmark(Bookmark: PBookmark);virtual;
      function GetBookmark:PBookmark;virtual;
      procedure GotoBookmark(Bookmark: PBookmark);virtual;
      function isPosBookmark(Bookmark: PBookmark):boolean;virtual;

      procedure FieldInfo(NumField:integer;var Rec:TFieldInfoRecord);virtual;
      function RecordCount:Longint;virtual;
      function RecordSize:word;virtual;
      function GetNumberFieldByName(AName:string):integer;virtual;
      function FieldCount:integer;virtual;

      procedure PutFields(AName:byte;Value:string);virtual;
      function GetFields(AName:byte):string;virtual;

      function HeaderSize:word;

      procedure SetFiltred(F:boolean);virtual;
      procedure SetFilter(AFilter:string);virtual;
      procedure SetKey;virtual;
      procedure EditKey;virtual;
      function GotoKey:boolean;virtual;
      function FindKey:boolean;virtual;
      function Location:boolean;virtual;

      procedure Post;virtual;
      procedure Edit;virtual;
      procedure Cancel;virtual;
      procedure Append;virtual;

      procedure EditRangeEnd;
      procedure EditRangeStart;
      procedure SetRangeEnd;
      procedure SetRangeStart;
      procedure ApplyRange;
      procedure CancelRange;
      procedure CancelRangeStart;
      procedure CancelRangeEnd;
      function ValidRange:boolean;virtual;

      function RLock:boolean;
      function RUnLock:boolean;

      procedure EmptyTable;virtual;
      procedure CopyStructure(AFileName:string);virtual;
      procedure Refresh;
      function LastUpdate:TDateTime;
      procedure FieldChanged(Field:PField);virtual;
      function GetMemoFile:PMemoFile;virtual;

      function FiType(ANum:byte):char;
      function FiName(ANum:byte):string;
      function FiLength(ANum:integer):integer;
      function FiOfs(ANum:integer):word;
      function FiDecimals(ANum:integer):integer;
      procedure RenameTable(NewTableName: string);virtual;
      function DoFilter:boolean;virtual;
      function ReadOnly:boolean;virtual;
      procedure ApplyUpdates;virtual;
      procedure CancelUpdates;virtual;
   private
      SetKeyBuf:PByteArray;
      FieldSet:TByteSet;
      CountFiSearch:integer;
      MemoFile:PMemoFile;
      fParser:PParser;
      fParserOPZ:PString;
      procedure CreateFieldList;
      procedure Put(isAppend:boolean);
      procedure Get;
      procedure ClearBuffer;
      procedure ConvertRecBuffer(Target:boolean);
   end;

type
   PStringField = ^TStringField;
   TStringField = object(TField)
     DefaultValue:PString;
     destructor Done;virtual;
     function AsString:string;virtual;
     function AsFloat:double;virtual;
     function AsInteger:Longint;virtual;
     procedure SetAsString( Value:String);virtual;
     procedure SetAsFloat( Value:Double);virtual;
     procedure SetAsInteger( Value:Longint);virtual;
     function DisplayText:string;virtual;
     procedure SetDafault(const s:string);
     procedure Load(S:PReader);virtual;
     procedure Store(S:PWriter);virtual;
     procedure OnNewRecord;virtual;
   end;

type
  PDateField = ^TDateField;
  TDateField = object(TField)
    DisplayFormat:PString;
    function AsString:string;virtual;
    function AsDateTime:TDateTime;virtual;
    procedure SetAsString(D:string);virtual;
    procedure SetAsDateTime(D:TDateTime);virtual;
    function DisplayText:string;virtual;
    function IsValidValue(S:String):boolean;virtual;
    procedure Load(S:PReader);virtual;
    procedure Store(S:PWriter);virtual;
    function GetWidth:word;virtual;
  private
    function GetValue:TDateTime;virtual;
    procedure SetValue(AValue:TDateTime);virtual;
  end;

type
  PVFPDateField = ^TVFPDateField;
  TVFPDateField = object(TDateField)
    function IsNull:boolean;virtual;
    function DisplayText:string;virtual;
    function AsString:string;virtual;
    procedure SetAsString(D:string);virtual;
    function GetWidth:word;virtual;
  private
    function GetValue:TDateTime;virtual;
    procedure SetValue(AValue:TDateTime);virtual;
  end;

type
  PIntegerField = ^TIntegerField;
  TIntegerField = object(TField)
     MinValue:int32;
     MaxValue:int32;
     CheckRange:Boolean;
     constructor Init(AOwner:PDataSet; AFieldNo:integer);
     function AsString:string;virtual;
     function AsInteger:Longint;virtual;
     function AsFloat:double;virtual;
     function AsBoolean:boolean;virtual;
     procedure SetAsString(D:string);virtual;
     procedure SetAsInteger(D:Longint);virtual;
     function DisplayText:string;virtual;
     function IsValidValue(S:String):boolean;virtual;
     procedure Load(S:PReader);virtual;
     procedure Store(S:PWriter);virtual;
  private
     function _CheckRange(Value:Longint):boolean;
     function GetValue:integer;virtual;
     procedure SetValue(AValue:integer);virtual;
  end;

type
  PVFPIntegerField = ^TVFPIntegerField;
  TVFPIntegerField = object(TIntegerField)
    constructor Init(AOwner:PDataSet; AFieldNo:integer);
  private
    function GetValue:integer;virtual;
    procedure SetValue(AValue:integer);virtual;
  end;

type
  PFloatField = ^TFloatField;
  TFloatField = object(TField)
    MinValue:Double;
    MaxValue:Double;
    CheckRange:Boolean;
    constructor Init(AOwner:PDataSet; AFieldNo:integer);
    function AsFloat:double;virtual;
    function AsString:string;virtual;
    function AsBoolean:boolean;virtual;
    procedure SetAsString(Value:string);virtual;
    procedure SetAsFloat(Value:Double);virtual;
    function DisplayText:string;virtual;
    function IsValidValue(S:String):boolean;virtual;
    procedure Load(S:PReader);virtual;
    procedure Store(S:PWriter);virtual;
    function GetWidth:word;virtual;
  private
    function _CheckRange(Value:Double):boolean;
    function GetValue:Double;virtual;
    procedure SetValue(AValue:Double);virtual;
  end;

type
  PFVPFloatField = ^TFVPFloatField;
  TFVPFloatField = object(TFloatField)
    constructor Init(AOwner:PDataSet; AFieldNo:integer);
    function GetWidth:word;virtual;
  private
    function GetValue:Double;virtual;
    procedure SetValue(AValue:Double);virtual;
  end;

  PVFPCurrencyField = ^TVFPCurrencyField;
  TVFPCurrencyField = object(TFVPFloatField)
    constructor Init(AOwner:PDataSet; AFieldNo:integer);
    function GetWidth:word;virtual;
  end;

type
  PBooleanField = ^TBooleanField;
  TBooleanField = object(TField)
    TextValuesTrue:PString;
    TextValuesFalse:PString;
    constructor Init(AOwner:PDataSet; AFieldNo:integer);
    destructor Done;virtual;
    function AsString:string;virtual;
    function AsBoolean:boolean;virtual;
    procedure SetAsBoolean(D:boolean);virtual;
    procedure SetAsString(D:string);virtual;
    function AsInteger:Longint;virtual;
    procedure SetAsInteger(D:Longint);virtual;
    function DisplayText:string;virtual;
    procedure Load(S:PReader);virtual;
    procedure Store(S:PWriter);virtual;
    function GetWidth:word;virtual;
  end;

type
   PLockUpField = ^TLockUpField;
   TLockUpField = object(TIntegerField)
     Items:PStrings;
     ErrorValue:PString;
     constructor Init(AOwner:PDataSet; AFieldNo:integer;
                                       AItems:PSItem;
                                       AErrorValue:String);
     destructor Done;virtual;
     function DisplayText:string;virtual;
   end;

type
  PDBLockUpField = ^TDBLockUpField;
  TDBLockUpField = object(TField)
     MainField:PField;
     IndexField:PField;
     ResultField:PField;
     MainOwner:PDataSEt;
     constructor Init(AOwner, ALockUp:PDataSet;
                  AFieldName,
                  ALockUpIndex, ALockUpName, ACalcName:string);
     function AsString:string;virtual;
     function AsInteger:Longint;virtual;
     function AsFloat:double;virtual;
     function AsBoolean:boolean;virtual;
     function AsDateTime:TDateTime;virtual;
     procedure Update;
     function Calculated:boolean;virtual;
  end;

type
  PExpressionField = ^TExpressionField;
  TExpressionField = object(TField)
     FieldDefs:PCollection;
     constructor Init(AOwner:PDataSet; Expression:string);
     destructor Done;virtual;
     function AsString:string;virtual;
     function DisplayText:string;virtual;
     function IndexString:string;virtual;
     function Calculated:boolean;virtual;
   private
     function TestIndexField(P:PField):Boolean;virtual;
  end;

type
   PMemoField = ^TMemoField;
   TMemoField = object(TCustomMemoField)
     constructor Init(AOwner:PDataSet; AFieldNo:integer);
     function AsString:string;virtual;
     procedure SetAsString(D:string);virtual;
     function DisplayText:string;virtual;
     procedure Assign(AField:PField);virtual;
     procedure SetData(var Rec:TMemoRec);virtual;
     procedure GetData(var Rec:TMemoRec);virtual;
     function GetSize:word;virtual;
   private
     function TestIndexField(P:PField):Boolean;virtual;
   end;

type
   PCFieldRec = ^TCFieldRec;
   TCFieldRec = record
     FieldName:PString;
     FieldType:TFieldType;
     Size:integer;
     Dec:integer;
     Next:PCFieldRec;
   end;

   PIndexRec = ^TIndexRec;
   TIndexRec = record
     IndexName:PString;
     IndexExpresion:PString;
     IndexOpt:TIndexOptions;
     Next:PIndexRec;
   end;

type
   PDBFTableConstructor = ^TDBFTableConstructor;
   TDBFTableConstructor = object(TObject)
     TableName:PString;
     TableType:byte;
     SubType:byte;
     constructor Init(ATableName:FNameStr;
              ATableType:byte; ASubType:byte);
     destructor Done; virtual;
     procedure AddField(FieldName:string; FieldType:TFieldType; ASize, ADec:integer);
     procedure AddIndex(IndexName, IndexExpresion:String; IndexOptions:TIndexOptions);
     function CreateTable:PDBFTable;virtual;
   private
     FieldDefs:PCFieldRec;
     IndexDefs:PIndexRec;
   end;

const
  DisplayDeleted   : boolean = false; {砫쭮 祭  SET DELETED}
  DefaultDriver    : byte = verDBase3;

const
  CharTypes = ['C'];

function GetIndexsField(DS:PDataSet; S:string; var FieldType:TIndexOptions):PField;
implementation
uses Ext_Math, Ext_Str, Strings, VMM, Ext_Sys;

{$I DBTable.inc}

   { TDBFStream }

constructor TDBFStream.Init(AFileName:FNameStr; AMode:word);
var b:byte;
begin
   inherited Init(AFileName, AMode);
   if Status=stOk then
   if AMode<>stCreate then
   begin
     Read(b, SizeOf(Byte));
     AssignDrivers(B);
     if not Assigned(OnRead) then
     begin
       AssignDrivers(DefaultDriver);
       if not Assigned(OnRead) then
          Status:=stInitError;
     end;
     ReadHeader;
   end;
end;

constructor TDBFStream.Create(AFileName:FNameStr; AFileHeader : PDBaseHeader);
begin
  if (AFileHeader<>nil) and (AFileHeader^.Version in
                             [ verDBase2, verDBase3,
                               verDBase3Memo, verFoxPro2XMemo,
                               verDBase7, verDBase7Memo])
  then
  begin
    inherited Init(AFileName, stCreate);
    FileHeader:=AFileHeader;
    AssignDrivers(FileHeader^.Version);
    WriteHeader;
  end
  else
    Status:=stInitError;
end;

destructor TDBFStream.Done;
begin
  if FileHeader<>nil then
  begin
    DisposeStr(FileHeader^.DBCName);
    Dispose(FileHeader);
    FileHeader:=nil;
  end;
  inherited Done;
end;

procedure TDBFStream.AssignDrivers(vers:byte);
begin
  case vers of
    $02:begin
          {$IFDEF FPC}
          OnRead:=@ReadDBase2Header;
          OnWrite:=@WriteDBase2Header;
          {$ELSE}
          OnRead:=ReadDBase2Header;
          OnWrite:=WriteDBase2Header;
          {$ENDIF}
        end;
    verDBase3,
    verDBase3Memo,
    verDBase4Memo,
    verFoxPro2XMemo,
    verVFP
       :begin
          {$IFDEF FPC}
          OnRead:=@ReadDBase3Header;
          OnWrite:=@WriteDBase3Header;
          {$ELSE}
          OnRead:=ReadDBase3Header;
          OnWrite:=WriteDBase3Header;
          {$ENDIF}
        end;
    $04,
    $8C:begin
          {$IFDEF FPC}
          OnRead:=@ReadDBase7Header;
          OnWrite:=@WriteDBase7Header;
          {$ELSE}
          OnRead:=ReadDBase7Header;
          OnWrite:=WriteDBase7Header;
          {$ENDIF}
        end;
  end;
end;

procedure TDBFStream.CopyStructure(S:PStream);
begin
  if (S<>nil) and Assigned(PDBFStream(S)^.OnWrite) then
     PDBFStream(S)^.OnWrite( PDBFStream(S), FileHeader, true);
end;

procedure TDBFStream.ReadHeader;
begin
  if Assigned(OnRead) then
     OnRead(@Self, FileHeader);
end;

procedure TDBFStream.WriteHeader;
begin
  if Assigned(OnWrite) then OnWrite(@Self, FileHeader, false);
end;

procedure TDBFStream.WriteFields;
begin
  if Assigned(OnWrite) then OnWrite(@Self, FileHeader, true);
end;

{******* constructor TrueWordInit *******************}
{      뢠                              }
{*******************************************************}
constructor TDBFTable.Init(AFileName:FNameStr;AMode:word);
var
  Ind:PIndexObject;
  i:integer;
begin
  inherited Init;
  IndexDefs:=New(PCollection,Init(10,5));
  TableName:=NewStr(ReplaceExt(AFileName,sDbfExt,false));
  Mode:=AMode;
  Ind:=IndexOpen(ReplaceExt(AFileName, '', true), @Self, true);
  if Ind<>nil then
  begin
    IndexDefs^.Insert(Ind);
    for i:=1 to Ind^.IndexStream^.IndexCount-1 do
      IndexDefs^.Insert(Ind^.IndexStream^.GetIndex(I));
  end;
  if not Open then
  begin
    Done;
    fail;
  end;
  CreateFieldList;
end;


{******* constructor TDBFTable.Create *****************}
{                                 }
{*******************************************************}
constructor TDBFTable.Create(AFileName:FNameStr;
                  var AFieldDefs: array of TFieldRecord;
                  fFieldCount:integer;
                  DbfTableType:byte; SubType:byte);
var
  i:word;
  AP:PString;
  MemoPresent:boolean;
begin
   inherited Init;
   AFileName:=ReplaceExt(AFileName,sDbfExt,false);
   SetDeleted(DispDeleted);
   PStream(DbfFile):=Session^.GetDBStream(AFileName);
   if DbfFile<>nil then
   begin
     AP:=@AFileName;
     DBError( edbFileNotCreate, AP);
     exit;
   end;
   DbfFile:=New(PDBFStream,Init(AFileName,stCreate));
   if DbfFile^.Status = stOk then
   begin
{     TableName:=DbfFIle^.FileName;}
     IndexDefs:=New(PCollection,Init(10,5));
     New(FileHeader);
     DbfFile^.FileHeader:=FileHeader;
     FillChar(FileHeader^, SizeOf(TDBaseHeader), 0);
     FileHeader^.LastUpdate:=Date;
     FileHeader^.FieldCount:=fFieldCount;
     FileHeader^.LanguageDriver:=SubType;
     FileHeader^.Version:=DbfTableType;
     GetMem(FileHeader^.Fields,FileHeader^.FieldCount*SizeOf(TFieldRecord));
     move(AFieldDefs, FileHeader^.Fields^, FileHeader^.FieldCount*SizeOf(TFieldRecord));
     Mode:=stOpen;
     FileHeader^.RecordSize:=1;
     MemoPresent:=false;
     for i:=0 to FileHeader^.FieldCount-1 do
     with FileHeader^.Fields^[i] do
     begin
       Name:=NewStr(Name^);
       fOffset:=FileHeader^.RecordSize;
       fType:=chUpCase(fType);
       if fType in ['N','F'] then
       begin
         if fLength>20 then fLength:=20;
         if fDecimal>min(fLength-2,18) then fDecimal:=min(fLength-2,18);
       end;
       case fType of
         'L':begin
               fLength:=1; fDecimal:=0;
             end;
         'I':begin
               fLength:=4; fDecimal:=0;
             end;
         'D','T':begin fLength:=8; fDecimal:=0; end;
         {'T':begin fLength:=8; fDecimal:=0; end;}
         'M',
         'B':begin
               fLength:=10;
               fDecimal:=0;
               MemoPresent:=true;
             end;
       end;
       inc(FileHeader^.RecordSize, fLength);
     end;
     if MemoPresent then
     case FileHeader^.Version of
       verDBase3:FileHeader^.Version:=verDBase3Memo;
       verDBase7:FileHeader^.Version:=verDBase7Memo;
     end
     else
     if FileHeader^.Version in [verDBase4Memo,verFoxPro2XMemo] then
        FileHeader^.Version:=verDBase3;

     DbfFile^.AssignDrivers(FileHeader^.Version);
     DbfFile^.WriteFields;
     if DbfFile^.Status<>stOk then dsStatus:=dsErrorWriteHeader;
   end
   else dsStatus:=dsErrorWriteHeader;
   TableName:=NewStr(DbfFIle^.FileName^);
   DbfFile^.Free;
   if dsStatus<>dsOk then
   begin
     DBError(edbFileNotCreate, TableName);
     DisposeStr(TableName);
     fail;
   end
   else
   begin
     Open;
     CreateFieldList;
   end;
end;

{******* destructor TDBFTable.Done; *******************}
{      뢠                       }
{*******************************************************}
destructor TDBFTable.Done;
begin
  Close;
  DisposeStr(TableName);
  TableName:=nil;
  if Master<>nil then SetMaster(nil,'');
{  if MemoFile<>nil then MemoFile^.Close;}
  inherited Done;
end;

function TDBFTable.Open:boolean;
var i:integer;
    Ind:PIndexObject;
begin
  Open:=true;
  PStream(DbfFile):=Session^.GetDBStream(TableName^);
  if DbfFile=nil then DbfFile:=New(PDBFStream,init(TableName^, Mode));
  if DbfFile^.Status<>stOk then
  begin
    DBError(edbFileNotOpen, TableName);
    Open:=false;
    exit;
  end
  else
  begin
    DisposeStr(TableName);
    TableName:=DbfFIle^.FileName;
    FileHeader:=DbfFIle^.FileHeader;
    if DSStatus=dsOk then
    begin
       GetMem( RecBuf, RecordSize);
       FillChar( RecBuf^, RecordSize, 32);
       GetMem( CancelBuf, RecordSize);
       FillChar( CancelBuf^, RecordSize, 32);
       GetMem( SetKeyBuf, RecordSize);
       FillChar( SetKeyBuf^, RecordSize, 32);
    end
    else
      DBError(edbFileNotOpen, TableName);
  end;
  //
  SetDeleted(DispDeleted);
  State:=dsBrowse;
  for i:=0 to IndexDefs^.Count-1 do PIndexObject(IndexDefs^.At(i))^.Open;
  FileHeader^.MemoFile:=MemoFile<>nil;
  if MemoFile<>nil then MemoFile^.Open;
  if FileHeader^.HeaderSize>=DbfFile^.GetSize then FileHeader^.RecordCount:=0;
  for i:=0 to FieldCount-1 do Fields(i)^.SetDataPtr;
  if FileHeader^.RecordCount<>0 then First else RecNo:=0;
end;

procedure TDBFTable.Close;
var i:integer;
begin
  if RecBuf<>nil then
  begin
    FreeMem( RecBuf, RecordSize);
    RecBuf:=nil;
  end;
  if CancelBuf<>nil then
  begin
    FreeMem( CancelBuf, RecordSize);
    CancelBuf:=nil;
  end;
  if SetKeyBuf<>nil then
  begin
    FreeMem( SetKeyBuf, RecordSize);
    SetKeyBuf:=nil;
  end;
  if StartRange<>nil then
  begin
    FreeMem( StartRange, RecordSize);
    StartRange:=nil;
  end;
  if EndRange<>nil then
  begin
    FreeMem(EndRange,RecordSize);
    EndRange:=nil;
  end;
  TableName:=NewStr(TableName^);
  Session^.FreeDBStream(DbfFile);
  DbfFile:=nil;
  State:=dsInactive;
  for i:=0 to IndexDefs^.Count-1 do PIndexObject(IndexDefs^.At(i))^.Close;
  if MemoFile<>nil then MemoFile^.Close;
end;

procedure TDBFTable.CreateFieldList;
var
  i:integer;
begin
  FieldDefs^.FreeAll;
  for i:=0 to FileHeader^.FieldCount-1 do
  begin
     if ShowSysField or (FiName(i)<>LockFieldName) then
        case FiType(i) of
          'D':FieldDefs^.Insert(New(PDateField, Init(PDBFTable(@Self),i)));
          'T':FieldDefs^.Insert(New(PVFPDateField, Init(PDBFTable(@Self),i)));
          'N':if FiDecimals(i)=0 then
                 FieldDefs^.Insert(
                 New(PIntegerField,
                     Init(PDBFTable(@Self),i)))
              else FieldDefs^.Insert(New(PFloatField,Init(@Self,i)));
          'F':FieldDefs^.Insert(New(PFloatField,Init(@Self,i)));
          'C':FieldDefs^.Insert(New(PStringField, Init(@Self,i)));
          'L':FieldDefs^.Insert(New(PBooleanField, Init(@Self,i)));
          'M':FieldDefs^.Insert(New(PMemoField, Init(@Self,i)));
          'I':FieldDefs^.Insert(New(PVFPIntegerField, Init(@Self,i)));
          'B':if  FileHeader^.Version=verVFP then
                FieldDefs^.Insert(New(PFVPFloatField , Init(@Self,i)))
              else FieldDefs^.Insert(New(PMemoField, Init(@Self,i)));
          'Y':FieldDefs^.Insert(New(PVFPCurrencyField, Init(@Self,i)));
        else
           FieldDefs^.Insert(New(PField,Init(@Self,i)));
        end;
  end;
end;

{$IFDEF VER70}
function TDBFTable.FieldCount:integer;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TDBFTable.FieldDefs
        MOV     AX,ES:[DI].TCollection.Count
end;
{$ELSE}
function TDBFTable.FieldCount:integer;
begin
  FieldCount:=FieldDefs^.Count;
end;
{$ENDIF}

procedure TDBFTable.AddIndex( AIndex:PIndexObject);
begin
  if AIndex=nil then exit;
  Index:=AIndex;
  IndexDefs^.Insert(Index);
  if PIndexObject(Index)^.iStatus=iOk then First;
end;

procedure OnFunction(Owner:PObject; Sender:PParser; const name:String; var p1,p2,p3:string;var Val:TParserValue);far;
var P:PField;
begin
  if Name='DELETED' then Val.AssignBool(PDataSet(Owner)^.DeletedRec) else
  if Name='ISNULL' then
  begin
    P:=PDataSet(Owner)^.FieldByName(P1);
    if P<>nil then Val.AssignBool(P^.IsNull)
    else Val.AssignBool(false);
  end;
end;

procedure OnGetValue(Owner:PObject; Sender:PParser; const s:String; var v:TParserValue);far;
var F:PField;
    Table:PDataSet absolute Owner;
    SS:Pstring;
begin
  F:=Table^.FindField(S);
  if (F=nil) and (Table^.CodePageDrv<>nil) then
     F:=Table^.FindField(Table^.CodePageDrv^.StrConvertFrom(S));
  if F<>nil then
  case F^.FieldType of
    ftString:V.AssignStr(F^.AsString);
    ftInteger,
    ftFloat:V.AssignNum(F^.AsFloat);
    ftBoolean:V.AssignBool(F^.AsBoolean);
    ftDate,
    ftDateTime
       :begin
          Sender^.Format:=fDate;
          V.AssignNum(F^.AsDateTime);
        end;
  end
  else
  begin
    SS:=@S;
    DBError(edbFieldNotFound, SS);
  end;
end;

procedure TDBFTable.SetFilter(AFilter:string);
begin
  inherited SetFilter(AFilter);
end;

procedure TDBFTable.SetFiltred(F:boolean);
var R:TParserValue;
begin
  if f then
  begin
    if Filter<>nil then
    begin
      fParser:=New(PParser, Init(@Self));
      fParser^.OnGetValue:=OnGetValue;
      fParser^.OnFunction:=OnFunction;
      fParserOPZ:=NewStr(fParser^.Str2OPZ(Filter^));
      R.Init;
      fParser^.CalcOPZ(fParserOPZ^, R);
      R.Done;
      if HasError then
      begin
        DBError(edbfInvalidFilter, Filter);
        HasError:=false;
        SetFiltred(false);
      end;
    end;
  end
  else
    if fParser<>nil then
    begin
      fParser^.Free;
      fParser:=nil;
      DisposeStr(fParserOPZ);
      fParserOPZ:=nil;
    end;
  inherited SetFiltred(F);
  First;
end;

function TDBFTable.DoFilter:boolean;
var R:TParserValue;
{$IFDEF VER70}
    Result:boolean;
{$ENDIF}
begin
  Result:=true;
  if Filtred then
  begin
    if (fParser<>nil) and (fParserOPZ<>nil) then
    begin
      R.Init;
      fParser^.CalcOPZ(fParserOPZ^, R);
      Result:=R.AsBoolean;
      R.Done;
    end;
    if Result and Assigned(OnFilter) then Result:=OnFilter(@Self);
  end;
{$IFDEF VER70}
  DoFilter:=Result;
{$ENDIF}
end;

procedure TDBFTable.SetIndex(IndexName:string);
var
  Dir: DirStr;
  Name: NameStr;
  Ext: ExtStr;
  P:Pointer;
  i:integer;
function Test(Item:PIndexObject):Boolean;
begin
  FSplit( Item^.FileName^, Dir, Name, Ext);
  Name:=stUpCase(Name);
  Test:=Name=IndexName;
end;
begin
  IndexName:=stUpCase(IndexName);
  Index:=nil;
  for i:=0 to IndexDefs^.Count-1 do
    if Test(IndexDefs^.At(i)) then
    begin
      Index:=IndexDefs^.At(i);
      break;
    end;
  if Index=nil then
  begin
    P:=@IndexName;
    DBError(edbIndexNotFound, P)
  end
end;

{************ TDBFTable.SetDeleted(Del:boolean) *******************}
{     蠥/頥 ⮡ࠦ 㤠 ᥩ             }
{*******************************************************************}
function TDBFTable.SetDeleted(Del:boolean):boolean;
begin
  SetDeleted:=DispDeleted;
  DispDeleted:=del;
end;

{******************* TDbf.Go **********************}
{  믮 p室   RecNo               }
{**************************************************}
procedure TDBFTable.Go(ARecNo:longint);
begin
  if State=dsInactive then
  begin
    fEOF:=true;
    fBOF:=true;
    RecNo:=0;
  end
  else
  begin
    fEOF:=(ARecNo>FileHeader^.RecordCount) or (FileHeader^.RecordCount=0);
    fBOF:=ARecNo<1;
    if (ARecNo<=FileHeader^.RecordCount) and (ARecNo>0) then
    begin
       RecNo:=ARecNo;
       Get;
       inherited Go(ARecNo);
       if Index<>nil then
       begin
         PIndexObject(Index)^.FindCurent;
         fEOF:=iEofFind in PIndexObject(Index)^.iStatus;
         fBOF:=iBofFind in PIndexObject(Index)^.iStatus;
       end;
    end;
    if fEOF or fBOF then ClearBuffer;
  end;
end;

{ॢ ப   ᮮ⢥ ஢}
procedure TDBFTable.ConvertRecBuffer(Target:boolean);
var i:integer;
    P:PCharArray;
begin
  if CodePageDrv<>nil then
  begin
    for i:=0 to FileHeader^.FieldCount-1 do
    begin
      if FileHeader^.Fields^[i].fType in CharTypes then
      begin
        P:=PCharArray(Longint(RecBuf)+FileHeader^.Fields^[i].fOffset);
        if Target then
          CodePageDrv^.BufConvertTO(P^, FileHeader^.Fields^[i].fLength)
        else CodePageDrv^.BufConvertFROM(P^, FileHeader^.Fields^[i].fLength);
      end;
    end;
  end;
end;

{  ⠥                   }
procedure TDBFTable.Get;
begin
  dsStatus:=dsOk;
  with FileHeader^, DbfFile^ do
  begin
    Seek(FileHeader^.HeaderSize+Pred(RecNo)*FileHeader^.RecordSize); {Seek(p)}
    Read(RecBuf^,RecordSize);
    if Status<>stOk then dsStatus:=dsErrorReadRecord
    else ConvertRecBuffer(true);
  end;
end;

{****************** TDBFTable.Put(var RB;f:boolean) *************}
{    ᫨ isAppend=TRUE  p     䠩    }
{      蠥 ⥪                             }
{****************************************************************}
procedure TDBFTable.Put(isAppend:boolean);
var
  p:longint;
begin
  if Mode=stOpenRead then DBError(edbDataSetReadOnly, TableName)
  else
  begin
    DbfFile^.ReadHeader;
    if isAppend then
    begin
      inc(FileHeader^.RecordCount);
      RecNo:=FileHeader^.RecordCount;
      DbfFile^.WriteHeader;
    end;
    ConvertRecBuffer(false);
    DbfFile^.seek(FileHeader^.HeaderSize+Pred(RecNo)*FileHeader^.RecordSize);
    DbfFile^.write(RecBuf^,RecordSize);
    ConvertRecBuffer(true);
    if DbfFile^.Status<>0 then DSStatus:=dsErrorReadFile
    else DSStatus:=dsOk;
  end;
end;

function TDBFTable.ReadOnly:boolean;
begin
  ReadOnly:=Mode=stOpenRead;
end;

{******************* TDBFTable.RecordSize **********************}
{  頥 ࠧ                                }
{****************************************************************}
{$IFDEF VER70}
function TDBFTable.RecordSize:word;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TDBFTable.FileHeader
        MOV     AX,ES:[DI].TDBaseHeader.RecordSize
end;
{$ELSE}
function TDBFTable.RecordSize:word;
begin
  RecordSize:=FileHeader^.RecordSize;
end;
{$ENDIF}

{******************* TDBFTable.HeaderSize **********************}
{  頥 ࠧ                            }
{***************************************************************}
{$IFDEF VER70}
function TDBFTable.HeaderSize:word;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TDBFTable.FileHeader
        MOV     AX,ES:[DI].TDBaseHeader.HeaderSize
end;
{$ELSE}
function TDBFTable.HeaderSize:word;
begin
  HeaderSize:=FileHeader^.HeaderSize;
end;
{$ENDIF}

{******************* TDBFTable.FiName(ANum:word) ****************}
{  頥                                             }
{****************************************************************}
function TDBFTable.FiName(ANum:byte):string;
begin
   FiName:=FileHeader^.Fields^[ANum].Name^;
end;

{******************* TDBFTable.FiType(ANum:word) ***************}
{  頥 ࠧ                             }
{****************************************************************}
function TDBFTable.FiType(ANum:byte):char;
begin
  FiType:=FileHeader^.Fields^[ANum].fType;
end;

{******************* TDBFTable.FiOfs(ANum:word) ****************}
{  頥 ࠧ                             }
{****************************************************************}
function TDBFTable.FiOfs(ANum:integer):word;
begin
  FiOfs:=FileHeader^.Fields^[ANum].fOffset
end;

{******************* TDBFTable.FiLength(ANum:word) ***************}
{  頥                                     }
{*******************************************************************}
function TDBFTable.FiLength(ANum:integer):integer;
begin
  FiLength:=FileHeader^.Fields^[ANum].fLength
end;

{******************* TDBFTable.FiDecimals(ANum:word) ***********}
{  頥 ᫮ ⨬                             }
{****************************************************************}
function TDBFTable.FiDecimals(ANum:integer):integer;
begin
  FiDecimals:=FileHeader^.Fields^[ANum].fDecimal
end;

procedure TDBFTable.FieldInfo(NumField:integer;var Rec:TFieldInfoRecord);
begin
   Rec.Name:=FiName(NumField);
   Rec.Size:=FiLength(NumField);
   Rec.Dec:=FiDecimals(NumField);
   Rec.Offset:=FiOfs(NumField);
   Rec.Align:=alLeft;
   case FiType(NumField) of
       'C':Rec.fType:=ftString;
       'N','I':begin
             Rec.Align:=alRight;
             if Rec.Dec<>0 then Rec.fType:=ftFloat
             else Rec.fType:=ftInteger;
           end;
       'F':begin
             Rec.Align:=alRight;
             Rec.fType:=ftFloat
           end;
       'L':Rec.fType:=ftBoolean;
       'D','T':begin
             Rec.fType:=ftDateTime;
             Rec.Align:=alRight;
           end;
       'M':Rec.fType:=ftMemo;
       'Y':begin
             Rec.fType:=ftCurrency;
             Rec.Align:=alRight;
           end;
       'B':if FileHeader^.Version=verVFP then
           begin
              Rec.Align:=alRight;
              Rec.fType:=ftFloat;
           end
           else Rec.fType:=ftBlob;
   end;
end;

function TDBFTable.RecordCount:Longint;
begin
  RecordCount:=FileHeader^.RecordCount;
end;

{************ TDBFTable.GetfFields(ANum:word):string ***********}
{                p頥 祭  AName                 }
{***************************************************************}
function TDBFTable.GetFields(AName:byte):string;
var
  s:string;
begin
  if State=dsInactive then GetFields:=''
  else
  begin
    SetLength(S,FileHeader^.Fields^[AName].fLength);
    move(RecBuf^[FileHeader^.Fields^[AName].fOffset],S[1],Length(S));
    GetFields:=S;
  end;
end;

{************* TDBFTable.PutFields(ANum:word;Value:string) *********}
{   뢠 祭 Value   AName  p              }
{********************************************************************}
procedure TDBFTable.PutFields( AName:byte; Value:string);
var
  P:PByteArray;
  Len:byte;
begin
  if State=dsInactive then DBError(edbDataSetClosed, TableName) else
  if State=dsBrowse then DBError(edbDSNotEditInsert, TableName)
   else
   begin
     FieldSet:=FieldSet+[AName];
     inc(CountFiSearch);
     RecBuf^[0]:=32;
     P:=@(RecBuf^[FileHeader^.Fields^[AName].fOffset]);
     Len:=FileHeader^.Fields^[AName].fLength;
     FillChar(P^,Len,#32);
     if Len>Length(Value) then Len:=Length(Value);
     Move(Value[1],P^,Len);
   end;
end;

procedure TDBFTable.SetKey;
begin
  EditKey;
  FillChar(RecBuf^, RecordSize, 0);
end;

procedure TDBFTable.EditKey;
begin
  if State=dsInactive then DBError(edbDataSetClosed, TableName) else
  begin
    State:=dsSetKey;
    CountFiSearch:=0;
    Move(RecBuf^, CancelBuf^, RecordSize);
    Move(SetKeyBuf^, RecBuf^,  RecordSize);
    FieldSet:=[];
  end;
end;

function TDBFTable.FindKey:boolean;
var
  L:Longint;
begin
  if State<>dsSetKey then DBError(edbDSNotsetKey, DbfFile^.FileName);
  if Index=nil then DBError(edbFileNotIndexed, TableName);
  State:=dsBrowse;
  L:=Index^.FindKey(Index^.IndexField^.IndexString, -1, false);
  FindKey:=L<>-1;
  Move(RecBuf^, SetKeyBuf^, RecordSize);
  Cancel;
  if L<>-1 then Go(L) else ClearBuffer;
  if L=-1 then DSStatus:=dsNotFind;
  FieldSet:=[];
end;

function TDBFTable.GotoKey:boolean;
var
  L:Longint;
begin
  if State<>dsSetKey then DBError(edbDSNotsetKey, DbfFile^.FileName);
  if Index=nil then DBError(edbFileNotIndexed, TableName);
  State:=dsBrowse;
  L:=Index^.FindKey(Index^.IndexField^.IndexString, -1, true);
  GotoKey:=L<>-1;
  Move(RecBuf^, SetKeyBuf^, RecordSize);
  Go(L);
  Cancel;
  if L=-1 then DSStatus:=dsNotFind;
  FieldSet:=[];
end;

{*************************************************}
{    ⥪饥 ᮤন  ।஢ }
{        䠩                                   }
{*************************************************}
function TDBFTable.Location:boolean;
var {Stop:boolean;}
    I,j,jMax:integer;
    PPos:Longint;
    S1,S2:string;
    P:Pointer;
    FindSet:TByteSet;
    FindFields:PCollection;
{function DoFind()
     FieldNo:=AFieldNo;}
begin
  dsStatus:=dsOk;
  State:=dsBrowse;
  Location:=true;
  Move(RecBuf^, CancelBuf^, RecordSize);
  FindSet:=FieldSet;
  Cancel;
  jMax:=0; P:=RecBuf;
  PPos:=RecNo;
  FindFields:=New(PCollection, Init(FieldCount,0));
  Location:=false;
  dsStatus:=dsNotFind;
  for i:=0 to FieldCount-1 do
    if Fields(i)^.FieldNo in FindSet then FindFields^.Insert(Fields(i));
  if FindFields^.Count>0 then
  begin
    if soGoFirst in LocateOptions then First;
    while not Eof do
    begin
      j:=0;
      for i:=0 to FindFields^.Count-1 do
      begin
        S1:=PField(FindFields^.At(i))^.AsString;
        if not (dsStatus in [dsOk,dsNotFind]) then break;
        dsStatus:=dsFindSet;
        RecBuf:=CancelBuf;
        S2:=PField(FindFields^.At(i))^.AsString;
        if not (dsStatus in [dsOk,dsNotFind,dsFindSet]) then break;
        dsStatus:=dsNotFind;
        RecBuf:=P;
        if (soCaseInSensitive in LocateOptions) then
        begin
          S1:=stUpCase(S1);
          S2:=stUpCase(S2);
        end;
        if soRightTrim in LocateOptions then
        begin
          S1:=TrimRight(S1);
          S2:=TrimRight(S2);
        end;
        if soInclude in LocateOptions then
        begin
          if Pos(S2,S1)<>0 then inc(j)
        end
        else
        if ASCIICompare(S1,S2)=0 then inc(j)
      end;
      if not (dsStatus in [dsOk,dsNotFind]) then break
      else
      if j=CountFiSearch then
      begin
        PPos:=RecNo;
        Location:=true;
        dsStatus:=dsOk;
        break;
      end
      else
      if j>jMax then
      begin
        jMax:=j;
        PPos:=RecNo;
      end;
      Next;
    end;
  end;
  Go(PPos);
  FindFields^.DeleteAll;
  Dispose(FindFields, Done);
end;

{******* procedure TDBFTable.Delete *******************}
{      砥      㤠      }
{*******************************************************}
procedure TDBFTable.Delete;
var i:integer;
begin
  if State=dsInactive then DBError(edbDataSetClosed, TableName) else
  begin
    for i:=0 to IndexDefs^.Count-1 do PIndexObject(IndexDefs^.At(i))^.DeleteCyrrent;
    DoBeforeDelete;
    RecBuf^[0]:=Byte('*');
    Put(false);
    DoAfterDelete;
    if Index<>nil then First;
  end;
end;

{************** Recall *********************************}
{      "㤠"           }
{*******************************************************}
procedure TDBFTable.Recall;
procedure AddCurent(Item:PIndexObject);
begin
  Item^.AddKey(Item^.IndexField^.DisplayText,RecNo);
end;
var i:integer;
begin
  RecBuf^[0]:=Byte(' ');
  Put({RecBuf^,}false);
  for i:=0 to IndexDefs^.Count-1 do AddCurent(IndexDefs^.At(i));
end;

{*********************************************}
{ 㭪 頥 ਧ ᪮      }
{ 㤠                              }
{*********************************************}
{$IFDEF VER70}
function TDBFTable.DeletedRec:boolean;assembler;
asm
        MOV  AL,1
        LES  DI,Self
        LES  DI,ES:[DI].TDBFTable.RecBuf
        CMP  BYTE PTR ES:[DI],'*'
        JE   @End
        DEC  AL
@End:
end;
{$ELSE}
function TDBFTable.DeletedRec:boolean;
begin
  DeletedRec:=RecBuf^[0]=Byte('*');
end;
{$ENDIF}
{************ TDBFTable.Edit ******************}
{ ॢ    ० ।஢ }
{***********************************************}
procedure TDBFTable.Edit;
begin
  if State<>dsBrowse then
     DBError(edbDSNotBrowse, DbfFile^.FileName)
  else
  begin
    DoBeforeEdit;
    State:=dsEdit;
    CountFiSearch:=0;
    Move(RecBuf^, CancelBuf^, RecordSize);
    FieldSet:=[];
    DoAfterEdit;
  end;
end;

procedure TDBFTable.Append;
var i:integer;
begin
  if State<>dsBrowse then
       DBError(edbDSNotBrowse, DbfFile^.FileName)
  else
  begin
    DoBeforeInsert;
    CountFiSearch:=0;
    Move(RecBuf^, CancelBuf^, RecordSize);
    FieldSet:=[];
    ClearBuffer;
    State:=dsInsert;
    DoAfterInsert;
    for i:=0 to FieldDefs^.Count-1 do Fields(i)^.OnNewRecord;
  end;
end;

procedure TDBFTable.Cancel;
begin
  DoBeforeCancel;
  FieldSet:=[];
  dsStatus:=dsOk;
  State:=dsBrowse;
  DbfFile^.ReadHeader;
  if (RecNo>0) and (RecNo<=FileHeader^.RecordCount) then Get
  else FillChar(RecBuf^, RecordSize, $20);
  DoAfterCancel;
end;

procedure TDBFTable.ClearBuffer;
begin
    FillChar(RecBuf^, RecordSize, $20);
end;

procedure TDBFTable.Post;
var P:Pointer;
procedure AddCurent(Item:PIndexObject);
begin
  Item^.AddKey(Item^.IndexField^.IndexString,RecNo);
end;
procedure DelCurent(Item:PIndexObject);
begin
  Item^.DeleteKey(Item^.IndexField^.IndexString, RecNo);
end;

function ChekExistRec:boolean;

begin
  ChekExistRec:=false;
  if (Index<>nil) and PIndexObject(Index)^.Unique and
     (PIndexObject(Index)^.FindKey(PIndexObject(Index)^.IndexField^.IndexString, RecNo, false)<>-1) then
  begin
    ChekExistRec:=true;
    SingleError(edbRecAlreadyExist)
  end;
end;

procedure _Edit;
var I:integer;
begin
  if FileHeader^.RecordCount = 0 then  Put(true)
  else
  begin
    P:=RecBuf; RecBuf:=CancelBuf;
    for i:=0 to IndexDefs^.Count-1 do DelCurent(IndexDefs^.At(i));
    RecBuf:=P;
    Put(false);
  end;
  for i:=0 to IndexDefs^.Count-1 do AddCurent(IndexDefs^.At(i));
  if Index<>nil then
  begin
    PIndexObject(Index)^.FindCurent;
    fEOF:=iEofFind in PIndexObject(Index)^.iStatus;
    fBOF:=iBofFind in PIndexObject(Index)^.iStatus;
  end;
end;

procedure _Insert;
var I:integer;
begin
  if ChekExistRec then Exit;
  Put( true);
  for i:=0 to IndexDefs^.Count-1 do AddCurent(IndexDefs^.At(i));
  if Index<>nil then
  begin
    PIndexObject(Index)^.FindCurent;
    fEOF:=iEofFind in PIndexObject(Index)^.iStatus;
    fBOF:=iBofFind in PIndexObject(Index)^.iStatus;
  end;
end;

procedure _SetKey;
begin
  if Index=nil then DBError(edbFileNotIndexed, TableName);
  Move(RecBuf^, SetKeyBuf^, RecordSize);
end;

begin
  if State=dsBrowse then exit;
  DoBeforePost;
  case State of
    dsEdit:_Edit;
    dsInsert:_Insert;
    dsSetKey:_SetKey;
    dsEditRangeStart,
    dsEditRangeEnd:ApplyRange;
  else
    DBError(edbDSNotEditInsert,DbfFile^.FileName);
  end;
  FieldSet:=[];
  DoAfterPost;
  State:=dsBrowse;
end;

procedure TDBFTable.SetRangeEnd;
begin
  EditRangeEnd;
  FillChar(RecBuf^, RecordSize, 0);
end;

procedure TDBFTable.SetRangeStart;
begin
  EditRangeStart;
  FillChar(RecBuf^, RecordSize, 0);
end;

procedure TDBFTable.EditRangeStart;
begin
  if StartRange=nil then
  begin
    GetMem(StartRange,RecordSize);
    FillChar(StartRange^, RecordSize, 0);
  end;
  Move(StartRange^, RecBuf^, RecordSize);
  State:=dsEditRangeStart;
  FieldSet:=[];
  CountFiSearch:=0;
end;

procedure TDBFTable.EditRangeEnd;
begin
  if EndRange=nil then
  begin
    GetMem(EndRange,RecordSize);
    FillChar(EndRange^, RecordSize, 0);
  end;
  Move(EndRange^, RecBuf^, RecordSize);
  FieldSet:=[];
  State:=dsEditRangeEnd;
  CountFiSearch:=0;
end;

procedure TDBFTable.ApplyRange;
begin
  if State=dsEditRangeStart then
  begin
    StartRangeSet:=FieldSet;
    Move(RecBuf^,StartRange^,RecordSize);
  end
  else
  begin
    EndRangeSet:=FieldSet;
    Move(RecBuf^,EndRange^,RecordSize);
  end;
  Cancel;
  First;
end;

{*****************************************************}
{  ࠥ  ࠭祭   ६    }
{*****************************************************}
procedure TDBFTable.CancelRange;
begin
  CancelRangeStart;
  CancelRangeEnd;
end;

{*****************************************************}
{  ࠥ ࠭祭                }
{           ६                                }
{*****************************************************}
procedure TDBFTable.CancelRangeStart;
begin
  if StartRange<>nil then
  begin
    FreeMem(StartRange,RecordSize);
    StartRange:=nil;
    StartRangeSet:=[];
  end
  else DBError(edbStartRangeNotInstal,DbfFile^.FileName);
end;

{*****************************************************}
{  ࠥ ࠭祭  孥           }
{            ६                               }
{*****************************************************}
procedure TDBFTable.CancelRangeEnd;
begin
  if EndRange<>nil then
  begin
    FreeMem(EndRange,RecordSize);
    EndRange:=nil;
    EndRangeSet:=[];
  end
  else DBError(edbEndRangeNotInstal,DbfFile^.FileName);
end;

{*****************************************************}
{  ஢ 宦           }
{            ⨬ 祭                      }
{*****************************************************}
function TDBFTable.ValidRange:boolean;
var i:integer;
    S:String;
    P:Pointer;
begin
  P:=RecBuf;
  if StartRange<>nil then
     for i:=1 to FieldCount do
       if i in StartRangeSet then
       begin
          RecBuf:=StartRange;
          S:=GetFields(i);
          RecBuf:=P;
          if GetFields(i)<S then
          begin
            ValidRange:=false;
            exit;
          end;
       end;
  if EndRange<>nil then
     for i:=1 to FieldCount do
       if i in EndRangeSet then
       begin
          RecBuf:=EndRange;
          S:=GetFields(i);
          RecBuf:=P;
          if GetFields(i)>S then
          begin
            ValidRange:=false;
            exit;
          end;
       end;
  ValidRange:=true;
end;

procedure TDBFTable.EmptyTable;
var i:integer;
begin
  if DbfFile^.CountUse<>1 then DBError(edbNonExclus, DbfFile^.FileName)
  else
  begin
     DBFFile^.Seek(FileHeader^.HeaderSize);
     DBFFile^.Truncate;
     DBFFile^.Flush;
     FileHeader^.RecordCount:=0;
     DBFFile^.WriteHeader;
     for i:=0 to IndexDefs^.Count-1 do PIndexObject(IndexDefs^.At(i))^.EmptyIndex;
  end;
end;

procedure TDBFTable.Refresh;
begin
  DbfFile^.ReadHeader;
end;

function TDBFTable.RLock:boolean;
{$IFDEF VER70}
var
  R:Registers;
  P:Longint;
begin
  R.AX:=$5C00;
  R.BX:=DbfFile^.Handle;
  R.SI:=0;
  R.DI:=RecordSize;
  P:=FileHeader^.HeaderSize+Pred(RecNo)*FileHeader^.RecordSize;
  R.CX:=LongRec(P).Hi;
  R.DX:=LongRec(P).Lo;
  MsDos(R);
  RLock:=R.Flags and fCarry = 0;
  if R.Flags and fCarry <> 0 then
  begin
     DBError(edbDataSetNotLock, DbfFile^.FileName);
     byte(dsStatus):=R.AL;
     DbfFile^.Reset;
  end;
end;
{$ELSE}
begin
end;
{$ENDIF}

function TDBFTable.RUnLock:boolean;
{$IFDEF VER70}
var
  R:Registers;
  P:Longint;
begin
  R.AX:=$5C01;
  R.BX:=DbfFile^.Handle;
  R.SI:=0;
  R.DI:=RecordSize;
  P:=FileHeader^.HeaderSize+Pred(RecNo)*FileHeader^.RecordSize;
  R.CX:=LongRec(P).Hi;
  R.DX:=LongRec(P).Lo;
  MsDos(R);
  RUnLock:=R.Flags and fCarry = 0;
  if R.Flags and fCarry <> 0 then
  begin
     DBError(edbDataSetNotUnLock, DbfFile^.FileName);
     byte(dsStatus):=R.AL;
     DbfFile^.Reset;
  end;
end;
{$ELSE}
begin
end;
{$ENDIF}

function TDBFTable.LastUpdate:TDateTime;
begin
  LastUpdate:=FileHeader^.LastUpdate;
end;

procedure TDBFTable.FreeBookmark(Bookmark: PBookmark);
begin
   if Bookmark=nil then exit;
   if Bookmark^.Owner<>@Self then DBError(edbErrorFreeBk, TableName);
   Dispose(Bookmark);
end;

function TDBFTable.GetBookmark: PBookmark;
var Bookmark: PBookmark;
begin
  New(Bookmark);
  Bookmark^.RecNo:=RecNo;
  Bookmark^.fEof:=fEof;
  Bookmark^.fBof:=fBof;
  Bookmark^.Owner:=@Self;
  GetBookmark:=Bookmark;
end;

function TDBFTable.isPosBookmark(Bookmark: PBookmark):boolean;
begin
  isPosBookmark:=(Bookmark<>nil) and (Bookmark^.Owner=@Self) and (Bookmark^.RecNo=RecNo);
end;

procedure TDBFTable.GotoBookmark(Bookmark: PBookmark);
begin
  if Bookmark=nil then exit;
  if Bookmark^.Owner<>@Self then DBError(edbErrorPosBk, TableName);
  Go(Bookmark^.RecNo);
  fEof:=Bookmark^.fEof;
  fBof:=Bookmark^.fBof;
end;

function TDBFTable.GetNumberFieldByName(AName:string):integer;
var
  i:integer;
  P:Pointer;
begin
  AName:=StUpCase(AName);
  for i:=1 to (FileHeader^.HeaderSize-33) div 32 do
    if StUpCase(FiName(i))=AName then
    begin
      GetNumberFieldByName:=i;
      exit;
    end;
  P:=@AName;
  DBError( edbFieldNotFound, P);
  GetNumberFieldByName:=0;
end;

procedure TDBFTable.FieldChanged(Field:PField);
begin
  if FieldDefs^.IndexOf(Field)=-1 then exit;
  if not (Field^.FieldNo in FieldSet) then
  begin
    FieldSet:=FieldSet+[Field^.FieldNo];
    inc(CountFiSearch);
  end;
end;

function TDBFTable.GetMemoFile:PMemoFile;
var R:PMemoFilesRecord;
begin
  if MemoFile=nil then
  begin
    R:=MemoFilesCreatedList;
    while R<>nil do
      if R^.ByteID=FileHeader^.Version then break
    else R:=R^.Next;
    if (R<>nil) and Assigned(R^.CreateFunc) then
     MemoFile:=R^.CreateFunc(@Self);
  end;
  GetMemoFile:=MemoFile;
end;

procedure TDBFTable.MoveBy(Distance:longint);
var
  L,i:LongInt;
begin
  fBOF:=false;
  fEOF:=false;
  if Distance=0 then exit;
  if Index = nil then
  begin
    if FileHeader^.RecordCount < (Distance+RecNo) then Last
    else
    if (Distance+RecNo)<=0 then First
    else
    if Filtred or DispDeleted then
    begin
      if Distance>0 then for i:=1 to Distance do Next
      else for i:=Distance downto 1 do Prior
    end
    else
    begin
      FBof:=Distance+RecNo<1;
      FEof:=Distance+RecNo>FileHeader^.RecordCount;
      Go(Distance+RecNo);
    end;
  end
  else
  begin
    L:=-1;
    if Distance>0 then
       for i:=0 to Distance-1 do
       begin
           L:=PIndexObject(Index)^.NextKey;
           if PIndexObject(Index)^.iStatus<>iOk then
           begin
             L:=PIndexObject(Index)^.LastKey;
             fEOF:=true;
             break;
           end;
       end
    else
       for i:=0 downto Distance+1 do
       begin
           L:=PIndexObject(Index)^.PrevKey;
           if PIndexObject(Index)^.iStatus<>iOk then
           begin
             L:=PIndexObject(Index)^.FirstKey;
             fBOF:=true;
             break;
           end;
       end;
    if L<>-1 then Go(l);
  end;
end;

procedure TDBFTable.Next;
var
  L, TempPtr:Longint;
begin
  TempPtr:=RecNo;
  fEOF:=false;
  if Index = nil then
  begin
    Go(RecNo+1);
    while not (DoFilter or fEof) do Go(RecNo+1);
    if fEof then
    begin
      TDBFTable.Go(TempPtr);
      fEOF:=true;
    end;
  end
  else
  begin
    L:=PIndexObject(Index)^.NextKey;
    if not (iEofFind in PIndexObject(Index)^.iStatus) then
    begin
      TDBFTable.Go(L);
      while not (DoFilter or (iEofFind in PIndexObject(Index)^.iStatus)) do
      begin
        L:=PIndexObject(Index)^.NextKey;
        if not (iEofFind in PIndexObject(Index)^.iStatus) then TDBFTable.Go(L);
      end;
    end;
    if iEofFind in PIndexObject(Index)^.iStatus then
    begin
      TDBFTable.Go(TempPtr);
      fEof:=true;
    end;
  end;
end;

procedure TDBFTable.Prior;
var
  L,TempPtr:Longint;
begin
  TempPtr:=RecNo;
  fBof:=false;
  if Index = nil then
  begin
    Go(RecNo-1);
    while not (DoFilter or fBof) do Go(RecNo-1);
    if fBof then
    begin
       Go(TempPtr);
       fBOF:=true;
    end;
  end
  else
  begin
    L:=PIndexObject(Index)^.PrevKey;
    if not (iBofFind in PIndexObject(Index)^.iStatus) then
    begin
      TDBFTable.Go(L);
      while not (DoFilter or (iBofFind in PIndexObject(Index)^.iStatus)) do
      begin
        L:=PIndexObject(Index)^.PrevKey;
        if not (iBofFind in PIndexObject(Index)^.iStatus) then
          TDBFTable.Go(L);
      end;
    end;
    if iBofFind in PIndexObject(Index)^.iStatus then
    begin
      TDBFTable.Go(TempPtr);
      fBof:=true;
    end;
  end;
end;

procedure TDBFTable.First;
var L:Longint;
begin
  fBOF:=true;
  fEOF:=false;
  if Index = nil then Go(1)
  else
  begin
    L:=Index^.FirstKey;
    if Index^.iStatus=iOk then TDBFTable.Go(L)
    else
    begin
      ClearBuffer;
      fEOF:=true;
    end;
  end;
  if not DoFilter then Next;
end;

procedure TDBFTable.Last;
var L:Longint;
begin
  fEOF:=true;
  fBOF:=false;
  if Index = nil then Go(FileHeader^.RecordCount)
  else
  begin
    L:=Index^.LastKey;
    if Index^.iStatus=iOk then TDBFTable.Go(L)
    else
    begin
      ClearBuffer;
      fEOF:=true;
    end;
  end;
  if not DoFilter then Prior;
end;

procedure TDBFTable.CreateIndex(AIndexName:FNameStr; AField:string; iOpt:TIndexOptions);
begin
  AddIndex(IndexCreate(AIndexName, @Self, AField, iOpt));
end;

procedure TDBFTable.OpenIndex(AIndexName:FNameStr);
begin
  AddIndex(IndexOpen(AIndexName, @Self, false));
end;

procedure TDBFTable.RenameTable(NewTableName: string);
var Temp:Longint;
    i:integer;
begin
  Temp:=RecNo;
  Close;
  if FileExists(NewTableName) then DeleteFile(NewTableName);
  NewTableName:=ExtractAliasName(TableName^)+ExtractName(NewTableName)+sDbfExt;
  RenameFile(FullName(TableName^), NewTableName);
  AssignStr(TableName, NewTableName);
  for i:=0 to IndexDefs^.Count-1 do PIndexObject(IndexDefs^.At(i))^.RenameIndex;
  if MemoFile<>nil then MemoFile^.RenameMemo;
  Open;
  Go(Temp);
end;

procedure TDBFTable.CopyStructure(AFileName:string);
var S:PDBFStream;
    TempL:longint;
begin
  S:=New(PDbfStream, Init(AFileName, stCreate));
  S^.OnWrite:=DbfFile^.OnWrite;
  TempL:=DbfFile^.FileHeader^.RecordCount;
  DbfFile^.FileHeader^.RecordCount:=0;
  DbfFile^.CopyStructure(S);
  DbfFile^.FileHeader^.RecordCount:=TempL;
  S^.Free;
end;

procedure TDBFTable.ApplyUpdates;
begin
  DbfFile^.WriteHeader;
  DbfFile^.Flush;
end;

procedure TDBFTable.CancelUpdates;
begin
  Cancel;
end;

  { TStringField }
destructor TStringField.Done;
begin
  SetDafault('');
  inherited Done;
end;

function TStringField.AsString:string;
begin
  AsString:=TrimRight(DisplayText);
end;

procedure TStringField.SetAsString( Value:String);
begin
  Owner^.PutFields( FieldNo, Value);
  UpdateField;
end;

{$IFDEF VER70}
function TStringField.DisplayText:string;assembler;
asm
        PUSH    DS
        LDS     SI,Self
        LES     DI,@Result
        MOV     DX,DS:[SI].TField.PtrInBuf
        MOV     CX,DS:[SI].TField.Size
        LDS     SI,DS:[SI].TField.Owner
        LDS     SI,DS:[SI].TDataSet.RecBuf
        ADD     SI,DX
@@2:    MOV     AX,CX
        CLD
        STOSB
        JCXZ  @End
    REP MOVSB
@End:
        POP     DS
end;
{$ELSE}
function TStringField.DisplayText:string;
begin
  DisplayText:=Owner^.GetFields(FieldNo);
end;
{$ENDIF}
function TStringField.AsFloat:double;
begin
  AsFloat := StrToFloat( AsString);
end;

procedure TStringField.SetAsFloat(Value:Double);
begin
  SetAsString(FloatToStr(Value));
end;

function TStringField.AsInteger:Longint;
begin
  AsInteger:=StrToInt( AsString);
end;

procedure TStringField.SetAsInteger( Value:Longint);
begin
  SetAsString(IntToStr(Value));
end;

procedure TStringField.SetDafault(const s:string);
begin
  AssignStr(DefaultValue, S);
end;

procedure TStringField.Load(S:PReader);
begin
  inherited Load(S);
  AssignStr(DefaultValue, S^.ReadString('DefaultValue', DinamicStr(DefaultValue)));
end;

procedure TStringField.Store(S:PWriter);
begin
  inherited Store(S);
  S^.WriteString('DefaultValue', DinamicStr(DefaultValue));
end;

procedure TStringField.OnNewRecord;
begin
  if DefaultValue<>nil then SetAsString(DefaultValue^);
end;

{ TDateField }
function TDateField.AsString:string;
begin
  if isNull then AsString:=''
  else AsString:=DateToStr(GetValue);
end;

procedure TDateField.SetAsString(D:string);
begin
  if D='' then Clear
  else SetValue( StrToDate(D));
end;

function TDateField.AsDateTime:TDateTime;
begin
  AsDateTime:=GetValue;
end;

procedure TDateField.SetAsDateTime(D:TDateTime);
begin
  SetValue(D);
end;

procedure TDateField.Load(S:PReader);
begin
  inherited Load(S);
  AssignStr(DisplayFormat, S^.ReadString('DisplayFormat',DisplayFormat^));
end;

procedure TDateField.Store(S:PWriter);
begin
  inherited Store(S);
  S^.WriteString('DisplayFormat',DinamicStr(DisplayFormat));
end;

function TDateField.DisplayText:string;
begin
  if isNull then DisplayText:=''
  else
  if DisplayFormat<>nil then
     DisplayText:=FormatDateTime( GetValue, DisplayFormat^)
  else DisplayText:=DateToStr(GetValue);
end;

function TDateField.IsValidValue(S:String):boolean;
var Year, Month, Day: Word;
    i:integer;
    R:Boolean;
    A:array [0..1] of Pointer;
    TempS:string;
begin
  IsValidValue:=true;
  if S='' then exit;
  TempS:=S;
  i:=Pos(DateSeparator,S);
  Day:=StrToInt(Copy(S,1,i-1));
  Delete(S,1,i);
  i:=Pos(DateSeparator,S);
  Month:=StrToInt(Copy(S,1,i-1));
  Delete(S,1,i);
  Year:=StrToInt(S);
  R:=(Year>=0) and (Month>0) and (Month<13) and
                (Day>0) and (Day<=DayInMonth(Month,Year));
  if not R then
  begin
    A[0]:=@TempS;
    A[1]:=Name;
    DBError(edbInvalidDateValue, A);
  end;
  IsValidValue:=R;
end;

function TDateField.GetValue:TDateTime;
var S:string;
begin
  SetLength(S,8);
  Move(DataPtr^, S[1], 8);
  GetValue:=EncodeDate(StrToInt(Copy(S,1,4)),
                       StrToInt(Copy(S,5,2)),
                       StrToInt(Copy(S,7,2)));
end;

procedure TDateField.SetValue(AValue:TDateTime);
var
  S,S1,S2,S3:String[15];
  Year, Month, Day:Word;
begin
  DecodeDate(AValue, Year, Month, Day);
  Str(Day:2, S1);
  if S1[1]=' ' then S1[1]:='0';
  Str(Month:2, S2);
  if S2[1]=' ' then S2[1]:='0';
  Str(Year:4, S3);
  S:=S3+S2+S1;
  Move(S[1], DataPtr^, 8);
  UpdateField;
end;

function TDateField.GetWidth:word;
begin
  GetWidth:=11;
end;

  { TVFPDateField }
type
  PVFPDateTime = ^TVFPDateTime;
  TVFPDateTime = packed record
    fDate:int32;
    fTime:int32;
  end;

function TVFPDateField.GetWidth:word;
begin
  GetWidth:=20;
end;

function TVFPDateField.DisplayText:string;
var Value:TDateTime;
begin
  if isNull then DisplayText:=''
  else
  begin
    Value:=GetValue;
    if DisplayFormat<>nil then
       DisplayText:=FormatDateTime( Value, DisplayFormat^)
    else DisplayText:=DateToStr(Value)+' '+TimeToStr(Value);
  end;
end;

function TVFPDateField.AsString:string;
var Value:TDateTime;
begin
  if isNull then AsString:=''
  else
  begin
    Value:=GetValue;
    AsString:=DateToStr(Value)+' '+TimeToStr(Value);
  end;
end;

procedure TVFPDateField.SetAsString(D:string);
var V:TDateTime;
begin
  D:=Trim(D);
  V:=0;
  if Pos(' ',D)<>0 then
  begin
    V:=StrToTime(Trim(Copy(D,Pos(' ',D)+1,Length(D))));
    Delete(D,Pos(' ',D)+1,Length(D));
  end;
  V:=V+StrToDate(Trim(D));
  SetValue(V);
end;

function TVFPDateField.IsNull:boolean;
begin
  IsNull:=(PVFPDateTime(DataPtr)^.fDate=0) and (PVFPDateTime(DataPtr)^.fTime=0);
end;

function TVFPDateField.GetValue:TDateTime;
begin
  GetValue:=PVFPDateTime(DataPtr)^.fDate - FVPDateDiff +
            (PVFPDateTime(DataPtr)^.fTime/MSecsPerDay);
end;

procedure TVFPDateField.SetValue(AValue:TDateTime);
begin
  PVFPDateTime(DataPtr)^.fDate:=FVPDateDiff + Round(AValue);
  PVFPDateTime(DataPtr)^.fTime:=Round(Frac(AValue)*MSecsPerDay);
  UpdateField;
end;

 {TIntegerField}

constructor TIntegerField.Init(AOwner:PDataSet; AFieldNo:integer);
begin
  inherited Init( AOwner, AFieldNo);
  if Size>9 then
  begin
    MinValue:=-MaxLongint-1;
    MaxValue:=MaxLongint;
  end
  else
  begin
    MinValue:=StrToInt('-'+DupChar('9',Size-1));
    MaxValue:=StrToInt(DupChar('9',Size));
  end;
  CheckRange:=true;
end;

function TIntegerField.AsString:string;
begin
  if IsNull then AsString:=''
  else AsString:=IntToStr(GetValue);
end;

function TIntegerField.AsInteger:Longint;
begin
  if IsNull then AsInteger:=0
  else AsInteger:=GetValue;
end;

function TIntegerField.AsFloat:double;
begin
  AsFloat:=AsInteger;
end;

function TIntegerField.AsBoolean:boolean;
begin
  AsBoolean:=(not IsNull) and (GetValue<>0);
end;

procedure TIntegerField.SetAsString(D:string);
var V:Longint;
    Code: Integer;
begin
  if D='' then Clear
  else
  begin
    Val(Trim(D), V, Code);
    if Code<>0 then V:=0;
    SetValue(V);
  end;
end;

procedure TIntegerField.SetAsInteger(D:Longint);
begin
  if _CheckRange(D) then SetValue(D);
end;

function TIntegerField.DisplayText:string;
begin
  if IsNull then DisplayText:=''
  else DisplayText:=IntToStr(GetValue);
end;

function TIntegerField.IsValidValue(S:String):boolean;
var V:Longint;
    Code: Integer;
    A:array [0..3] of pointer;
begin
  IsValidValue:=true;
  if S='' then Exit;
  Val(S, V, Code);
  if Code=0 then IsValidValue:=_CheckRange(V)
  else
  begin
    A[0]:=@S;
    A[1]:=Name;
    DBError(edbInvalidIntegerValue, A);
    IsValidValue:=false;
  end;
end;

function TIntegerField._CheckRange(Value:Longint):boolean;
var A:array [0..3] of Pointer;
begin
  if CheckRange then
    if (Value>=MinValue) and (Value<=MaxValue) then _CheckRange:=true
    else
    begin
      Longint(A[0]):=Value;
      A[1]:=Name;
      Longint(A[2]):=MinValue;
      Longint(A[3]):=MaxValue;
      DBError(edbIntValueOutOfRange, A);
    end
  else _CheckRange:=true;
end;

procedure TIntegerField.Load(S:PReader);
begin
  inherited Load(S);
  MinValue:=S^.ReadInteger('MinValue',MinValue);
  MaxValue:=S^.ReadInteger('MaxValue',MaxValue);
  CheckRange:=S^.ReadBoolean('CheckRange',CheckRange);
end;

procedure TIntegerField.Store(S:PWriter);
begin
  inherited Store(S);
  S^.WriteInteger('MinValue',MinValue);
  S^.WriteInteger('MaxValue',MaxValue);
  S^.WriteBoolean('CheckRange',CheckRange);
end;

function TIntegerField.GetValue:integer;
var S:string;
begin
  SetLength(S, Size);
  Move(DataPtr^,S[1], Size);
  GetValue:=StrToInt(S);
end;

procedure TIntegerField.SetValue(AValue:integer);
var S:string;
begin
  Str(AValue:Size,S);
  Move(S[1], DataPtr^, Size);
  UpdateField;
end;

  {TVFPIntegerField}
constructor TVFPIntegerField.Init(AOwner:PDataSet; AFieldNo:integer);
begin
  TField.Init( AOwner, AFieldNo);
  MinValue:=-MaxLongint-1;
  MaxValue:=MaxLongint;
{  Size:=8;}
end;

function TVFPIntegerField.GetValue:integer;
begin
  GetValue:=PLongint(DataPtr)^;
end;

procedure TVFPIntegerField.SetValue(AValue:integer);
begin
  PLongint(DataPtr)^:=AValue;
  UpdateField;
end;

   {TFloatField}

constructor TFloatField.Init(AOwner:PDataSet; AFieldNo:integer);
begin
  inherited Init( AOwner, AFieldNo);
{  MinValue:=StrToFloat('-'+DupChar('9',Size-Dec-2)+'.'+DupChar('9',Dec));
  MaxValue:=StrToFloat(DupChar('9',Size-Dec-1)+'.'+DupChar('9',Dec));
  CheckRange:=true;}
end;

function TFloatField.GetWidth:word;
begin
  GetWidth:=Size+1;
end;

function TFloatField.AsString:string;
begin
  if IsNull then AsString:=''
  else AsString:=FloatToStr(GetValue);
end;

function TFloatField.AsBoolean:boolean;
begin
  AsBoolean:=(not IsNull) and (GetValue<>0);
end;

procedure TFloatField.SetAsString(Value:string);
var V:Double;
    Code: Integer;
begin
  if Value='' then Clear
  else
  begin
    Code:=Pos(DecimalSeparator,Value);
    if Code<>0 then Value[Code]:='.';
    Val(Trim(Value), V, Code);
    if Code<>0 then V:=0;
    SetValue(V);
  end;
end;

function TFloatField.AsFloat:double;
begin
  if IsNull then AsFloat:=0
  else AsFloat:=GetValue;
end;

function TFloatField._CheckRange(Value:Double):boolean;
var A:array [0..3] of Pointer;
    S1,S2,S3,S4:string[30];
begin
  _CheckRange:=true;
  if CheckRange and ((Value < MinValue) or (Value > MaxValue)) then
  begin
    if Owner^.CodePageDrv<>nil then S4:=Owner^.CodePageDrv^.StrConvertTO(Name^)
    else S4:=Name^;
    A[0]:=@S3;
    A[1]:=@S4;
    A[2]:=@S1;
    A[3]:=@S2;
    S1:=FloatToStr(MinValue);
    S2:=FloatToStr(MaxValue);
    S3:=FloatToStr(Value);
    DBError(edbFValueOutOfRange, A);
    _CheckRange:=false
  end;
end;

procedure TFloatField.SetAsFloat(Value:Double);
var S:string;
begin
  if _CheckRange(Value) then SetValue(Value);
end;

function TFloatField.DisplayText:string;
{$IFDEF VER70}
var Result:string;
{$ENDIF}
begin
  if isNull then DisplayText:='' else
  if isCurrency then DisplayText:=CurrencyToStr(GetValue)
  else
  begin
    Str(GetValue:Size:Precision,Result);
    Result[Pos('.',Result)]:=DecimalSeparator;
    {$IFDEF VER70}
    DisplayText:=Result;
    {$ENDIF}
  end;
end;

function TFloatField.IsValidValue(S:String):boolean;
var V:Double;
    Code: Integer;
    i:integer;
    A:array [0..1] of Pointer;
begin
  IsValidValue:=true;
  if S='' then exit;
  i:=Pos(DecimalSeparator,S);
  if i<>0 then S[i]:='.';
  if Trim(S)='.' then exit;
  Val(S, V, Code);
  if Code=0 then IsValidValue:=_CheckRange( V)
  else
  begin
    A[0]:=@S;
    A[1]:=Name;
    DBError(edbInvalidFloatValue, A);
    IsValidValue:=false;
  end;
end;

procedure TFloatField.Load(S:PReader);
begin
  inherited Load(S);
  MinValue:=S^.ReadFloat('MinValue',MinValue);
  MaxValue:=S^.ReadFloat('MaxValue',MaxValue);
  CheckRange:=S^.ReadBoolean('CheckRange',CheckRange);
end;

procedure TFloatField.Store(S:PWriter);
begin
  inherited Store(S);
  S^.WriteFloat('MinValue',MinValue);
  S^.WriteFloat('MaxValue',MaxValue);
  S^.WriteBoolean('CheckRange',CheckRange);
end;

function TFloatField.GetValue:Double;
var S:string;
begin
  SetLength(S, Size);
  Move(DataPtr^,S[1], Size);
  GetValue:=StrToFloat(S);
end;

procedure TFloatField.SetValue(AValue:Double);
var S:string;
begin
  Str(AValue:Size:Precision,S);
  Move(S[1], DataPtr^, Size);
  UpdateField;
end;

  {TFVPFloatField}
constructor TFVPFloatField.Init(AOwner:PDataSet; AFieldNo:integer);
begin
  inherited Init(AOwner,AFieldNo);
{  MinValue:=-1.79E307;
  MaxValue:=1.79769313486232E308;}
end;

function TFVPFloatField.GetValue:Double;
begin
  GetValue:=PDouble(DataPtr)^;
end;

procedure TFVPFloatField.SetValue(AValue:Double);
begin
  PDouble(DataPtr)^:=AValue;
  UpdateField;
end;

function TFVPFloatField.GetWidth:word;
begin
  GetWidth:=12;
end;

  {TVFPCurrencyField}
constructor TVFPCurrencyField.Init(AOwner:PDataSet; AFieldNo:integer);
begin
  inherited Init(AOwner,AFieldNo);
  isCurrency:=true;
end;

function TVFPCurrencyField.GetWidth:word;
begin
  GetWidth:=12+Length(CurrencyString);
end;

  {TBooleanField}
constructor TBooleanField.Init(AOwner:PDataSet; AFieldNo:integer);
begin
  inherited Init(AOwner, AFieldNo);
  TextValuesTrue:=NewStr(BooleanText[true]);
  TextValuesFalse:=NewStr(BooleanText[false]);
end;

destructor TBooleanField.Done;
begin
  DisposeStr(TextValuesTrue);
  DisposeStr(TextValuesFalse);
  inherited Done;
end;

function TBooleanField.AsString:string;
begin
  AsString:=DisplayText;
end;

function TBooleanField.GetWidth:word;
begin
  GetWidth:=Max(3,Max(length(DinamicStr(TextValuesTrue)),
                length(DinamicStr(TextValuesFalse))));
end;

procedure TBooleanField.SetAsBoolean(D:boolean);
var C:char;
begin
  if D then C:='Y' else C:='N';
  Owner^.PutFields(FieldNo, C);
end;

{$IFDEF VER70}
function TBooleanField.AsBoolean:boolean;assembler;
asm
        LES     DI,Self
        MOV     BX,ES:[DI].TField.PtrInBuf
        LES     DI,ES:[DI].TField.Owner
        LES     DI,ES:[DI].TDataSet.RecBuf
        MOV     AH,ES:[DI+BX]
        AND     AH,$DF
        MOV     AL,1
        CMP     AH,'Y'
        JE      @End
        CMP     AH,'T'
        JE      @End
        DEC     AL
@End:
end;
{$ELSE}
function TBooleanField.AsBoolean:boolean;
var S:string;
begin
  S:=TrimLeft(Owner^.GetFields(FieldNo));
  AsBoolean:=(S[1]='Y') or (S[1]='T');
end;
{$ENDIF}
procedure TBooleanField.SetAsString(D:string);
var S:string;
begin
  if D='' then Clear
  else
  begin
    S:=stUpCase(TextValuesTrue^);
    D:=stUpCase(D);
    SetAsBoolean((Pos(S,D)<>0) or (Pos(D,S)<>0));
    UpdateField;
  end;
end;

function TBooleanField.AsInteger:Longint;
begin
  AsInteger:=byte(AsBoolean)
end;

procedure TBooleanField.SetAsInteger(D:Longint);
begin
  SetAsBoolean(D<>0)
end;

function TBooleanField.DisplayText:string;{$IFDEF VER70}assembler;
asm
        PUSH    DS
        LDS     SI,Self
        MOV     BX,DS:[SI].TField.PtrInBuf
        LES     DI,DS:[SI].TField.Owner
        LES     DI,ES:[DI].TDataSet.RecBuf
        MOV     AH,ES:[DI+BX]
        LES     DI,@Result
        AND     AH,$DF
        MOV     AL,1
        CMP     AH,'Y'
        JE      @1
        CMP     AH,'T'
        JE      @1
        LDS     SI,DS:[SI].TBooleanField.TextValuesFalse
        JMP     @2
@1:
        LDS     SI,DS:[SI].TBooleanField.TextValuesTrue
@2:
        XOR     AX,AX
        LODSB
        MOV     CX,AX
        STOSB
    REP MOVSB
        POP     DS
end;
{$ELSE}
begin
  if isNull then DisplayText:='' else
  if AsBoolean then DisplayText:=TextValuesTrue^
  else DisplayText:=TextValuesFalse^;
end;
{$ENDIF}

procedure TBooleanField.Load(S:PReader);
begin
  inherited Load(S);
  AssignStr(TextValuesTrue,S^.ReadString('TextValuesTrue',TextValuesTrue^));
  AssignStr(TextValuesFalse,S^.ReadString('TextValuesFalse',TextValuesFalse^));
end;

procedure TBooleanField.Store(S:PWriter);
begin
  inherited Store(S);
  S^.WriteString('TextValuesTrue',DinamicStr(TextValuesTrue));
  S^.WriteString('TextValuesFalse',DinamicStr(TextValuesFalse));
end;

 { TLockUpField }
constructor TDBLockUpField.Init(AOwner, ALockUp:PDataSet;
                  AFieldName,
                  ALockUpIndex, ALockUpName, ACalcName:string);
var P:POinter;
begin
  TObject.Init;
  Owner:=ALockUp;
  MainOwner:=AOwner;
  ResultField:=Owner^.FieldByName(ALockUpName);
  IndexField:=Owner^.FieldByName(ALockUpIndex);
  MainField:=AOwner^.FieldByName(AFieldName);
  if MainField^.FieldType<>IndexField^.FieldType then
  begin
     P:=@ACalcName;
     DBError(edbLockUpFieldType, P);
     Done;
     Fail;
  end;
  AssignStr(Name, ACalcname);
  Align:=ResultField^.Align;
  FieldType:=ResultField^.FieldType;
  Size:=ResultField^.Size;
end;

function TDBLockUpField.AsString:string;
begin
  Update;
  AsString:=ResultField^.AsString;
end;

function TDBLockUpField.AsInteger:Longint;
begin
  Update;
  AsInteger:=ResultField^.AsInteger;
end;

function TDBLockUpField.AsFloat:Double;
begin
  Update;
  AsFloat:=ResultField^.AsFloat;
end;

function TDBLockUpField.AsBoolean:boolean;
begin
  Update;
  AsBoolean:=ResultField^.AsBoolean;
end;

function TDBLockUpField.AsDateTime:TDateTime;
begin
  Update;
  AsDateTime:=ResultField^.AsDateTime;
end;

procedure TDBLockUpField.Update;
begin
  Owner^.SetKey;
  case IndexField^.FieldType of
    ftString:IndexField^.SetAsString(MainField^.AsString);
    ftFloat:IndexField^.setAsFloat(MainField^.AsFloat);
    ftInteger:IndexField^.setAsInteger(MainField^.AsInteger);
  end;
{  Owner^.GotoKey;}
  Owner^.FindKey;
end;

function TDBLockUpField.Calculated:boolean;
begin
  Calculated:=true;
end;

  { TLockUpField }
constructor TLockUpField.Init(AOwner:PDataSet; AFieldNo:integer;
                                       AItems:PSItem; AErrorValue:string);
var P:PSItem;
begin
  inherited Init(AOwner, AFieldNo);
  ErrorValue:=NewStr(AErrorValue);
  Items:=New(PStrings, Init);
  while AItems<>nil do
  begin
    P:=AItems^.Next;
    Items^.Insert(AItems^.Value);
    Dispose(AItems);
    AItems:=P;
  end;
  Align:=alLeft;
end;

destructor TLockUpField.Done;
begin
  DisposeStr(ErrorValue);
  Dispose(Items, Done);
  inherited Done;
end;

function TLockUpField.DisplayText:string;
var L:Longint;
begin
  L:=AsInteger;
  if (L>=0) and (L<Items^.Count) then DisplayText:=Items^.Str(L)
  else if ErrorValue<>nil then DisplayText:=ErrorValue^ else DisplayText:='';
end;

{for expresion field}
function CheckFunc(S:string):TDBaseFunc;
begin
  CheckFunc:=funError;
  S:=stUpCase(S);
  if S='STR' then CheckFunc:=funStr;
  if S='DTOS' then CheckFunc:=funDtoS;
end;

type
  PFieldRec = ^TFieldRec;
  TFieldRec = object(TObject)
    Field:PField;
    Start:integer;
    Len:integer;
    constructor Init(DS:PDataSet; S:string);
  end;

constructor TFieldRec.Init(DS:PDataSet; S:string);
var I,j,k:integer;
begin
  inherited Init;
  Start:=1;
  Len:=-1;
  i:=Pos('(',S);
  if I<>0 then
  begin
    j:=Pos(')',S);
    S:=copy(S,I+1,j-i-1);
    K:=Pos(':',S);
    if K<>0 then
    begin
      Len:=StrToInt(Trim(Copy(S,K+1,255)));
      Delete(S,k,255);
    end;
  end;
  Field:=DS^.FindField(S);
  if Len=-1 then Len:=Field^.Size;
end;

  { TExpressionField }

constructor TExpressionField.Init(AOwner:PDataSet; Expression:string);
var k:integer;
function CheckExpres:boolean;
var S,SS:string;
    I,J,Len:integer;
    P:PField;
begin
  S:=Expression;
  K:=0;
  repeat
    i:=Pos('+',S);
    if i<>0 then
    begin
      SS:=copy(S,1,i-1);
      Delete(S,1,i);
    end
    else
    begin
      SS:=S;
      S:='';
    end;
    i:=Pos('(',SS);
    if I<>0 then
    begin
      J:=Pos(':',SS);
      if J<>0 then
      begin
        Len:=Pos(')',SS)-J-1;
        if CheckNum(Copy(SS,J+1,Len)) then Delete(SS,J,Len+1)
        else SS:='';
      end;
      j:=Pos(')',SS);
      if (j=0) or (CheckFunc(copy(SS,1,i-1))<>funError) then
      SS:=copy(SS,I+1,j-i-1)
      else SS:='';
    end;
    P:=Owner^.FindField(SS);
    if (P=nil) or P^.Calculated then
    begin
      S:='';
      k:=0;
    end
    else inc(k);
  until S='';
  CheckExpres:=K<>0;
end;
var i,j:integer;
    P:Pointer;
begin
  TObject.Init;
  Owner:=AOwner;
  if not CheckExpres then
  begin
    P:=@Expression;
    DBError(edbIndexFieldTypeMismatch, P);
    fail;
  end;
  Name:=NewStr(Expression);
  FieldDefs:=New(PCollection, Init(10,5));
  for i:=1 to k do
  begin
    j:=pos('+',Expression);
    if j=0 then j:=length(Expression)+1;
    FieldDefs^.Insert(New(PFieldRec, Init(Owner, copy(Expression,1,j-1))));
    Delete(Expression,1,j);
  end;
  for I:=0 to FieldDefs^.Count-1 do
     Size:=Size+PFieldRec(FieldDefs^.At(i))^.Len;
end;

destructor TExpressionField.Done;
begin
  Dispose(FieldDefs, Done);
  inherited Done;
end;

function TExpressionField.AsString:string;
begin
  AsString:=TrimRight(IndexString);
end;

function TExpressionField.DisplayText:string;
begin
  DisplayText:=IndexString;
end;

function TExpressionField.IndexString:string;
var i:integer;
    F:PFieldRec;
{$IFDEF VER70}
    Result:string;
{$ENDIF}
begin
  Result:='';
  for i:=0 to FieldDefs^.Count-1 do
  begin
    F:=FieldDefs^.At(i);
    Result:=Result+Copy(F^.Field^.IndexString,F^.Start,F^.Len);
  end;
{$IFDEF VER70}
  IndexString:=Result;
{$ENDIF}
end;

function TExpressionField.Calculated:boolean;
begin
  Calculated:=true;
end;

function TExpressionField.TestIndexField(P:PField):Boolean;
function DoTest(PRec:PFieldRec):Boolean;{$IFNDEF FPC}far;{$ENDIF}
begin
  DoTest:=PRec^.Field=P;
end;
{$IFDEF PPC_DELPHI}
var i:integer;
{$ENDIF}
begin
{$IFNDEF PPC_DELPHI}
  TestIndexField:=FieldDefs^.FirstThat(@DoTest)<>nil;
{$ELSE}
  for i:=0 to FieldDefs^.Count-1 do
    if DoTest(FieldDefs^.At(i)) then
    begin
      TestIndexField:=true;
      exit;
    end;
  TestIndexField:=false;
{$ENDIF}
end;

function GetIndexsField(DS:PDataSet; S:string; var FieldType:TIndexOptions):PField;
begin
  if (Pos('+',S)<>0) or (Pos('(',S)<>0) or (Pos(')',S)<>0) then
  begin
    GetIndexsField:=New(PExpressionField, Init(DS, S));
    FieldType:=[ixExpression];
  end
  else
  begin
    GetIndexsField:=DS^.FieldByName(S);
    FieldType:=[];
  end
end;

 { TMemoField }
constructor TMemoField.Init(AOwner:PDataSet; AFieldNo:integer);
begin
  inherited Init(AOwner, AFieldNo);
  if PDBFTable(Owner)^.MemoFile=nil then
     PDBFTable(Owner)^.MemoFile:=Owner^.GetMemoFile;
end;

function TMemoField.AsString:string;
var P:PMemoRec;
    S:string;
    L:word;
begin
  L:=GetSize;
  GetMem(P, L+SizeOf(Word));
  GetData(P^);
  SetLength(S,Min(255,L));
  Move(P^.Text, S[1], Length(S));
  FreeMem(P, L+SizeOf(Word));
  AsString:=S;
end;

procedure TMemoField.SetAsString(D:string);
var P:PMemoRec;
begin
  GetMem(P, Length(D)+SizeOf(Word));
  Move( D[1], P^.Text, Length(D));
  P^.Length:=Length(D);
  SetData(P^);
  FreeMem(P, Length(D)+SizeOf(Word));
end;

function TMemoField.DisplayText:string;
begin
   if ShowMemoValue then DisplayText:=Owner^.GetFields(FieldNo)
   else DisplayText:=DefMemoString
end;

procedure TMemoField.Assign(AField:PField);
var P:PMemoRec;
    L:word;
begin
  L:=AField^.GetSize;
  GetMem(P, L);
  P^.Length:=L;
  AField^.GetData(P^);
  SetData(P^);
  FreeMem(P, L);
end;

procedure TMemoField.SetData(var Rec:TMemoRec);
var S:string;
begin
  with PDBFTable(Owner)^ do
    if MemoFile<>nil then
    begin
      if MemoFile^.CalcOffset(DataPtr)<>0 then
         MemoFile^.AtFree(MemoFile^.CalcOffset(DataPtr));
      if Rec.Length<>0 then
      begin
        if CodePageDrv<>nil then
          CodePageDrv^.BufConvertFROM(Rec.Text,Rec.Length);
        MemoFile^.StoreOffset(DataPtr,MemoFile^.Put(Rec));
        if CodePageDrv<>nil then
          CodePageDrv^.BufConvertTO(Rec.Text,Rec.Length);
      end;
      UpdateField;
    end;
end;

procedure TMemoField.GetData(var Rec:TMemoRec);
begin
  with PDBFTable(Owner)^ do
    if MemoFile<>nil then
    begin
      MemoFile^.Get(MemoFile^.CalcOffset(DataPtr), Rec);
      if CodePageDrv<>nil then CodePageDrv^.BufConvertTO(Rec.Text,Rec.Length);
  end;
end;

function TMemoField.GetSize:word;
begin
  with PDBFTable(Owner)^ do
    if MemoFile<>nil then
      GetSize:=MemoFile^.GetSize(MemoFile^.CalcOffset(DataPtr))
    else GetSize:=0;
end;


function TMemoField.TestIndexField(P:PField):Boolean;
begin
  TestIndexField:=False;
end;

{}
   { TDBFTableConctructor }
constructor TDBFTableConstructor.Init(ATableName:FNameStr;
              ATableType:byte; ASubType:byte);
begin
  inherited Init;
  if Session=nil then SingleError(edbSesionNotInit)
  else
  begin
    TableName:=NewStr(ATableName);
    TableType:=ATableType;
    SubType:=ASubType;
  end;
end;

destructor TDBFTableConstructor.Done;
var PF:PCFieldRec;
    IRec:PIndexRec;
begin
  DisposeStr(TableName);
  PF:=FieldDefs;
  while PF<>nil do
  begin
   FieldDefs:=PF^.Next;
   DisposeStr(PF^.FieldName);
   Dispose(PF);
   PF:=FieldDefs;
  end;
  IRec:=IndexDefs;
  while IRec<>nil do
  begin
   IndexDefs:=IRec^.Next;
   DisposeStr(IRec^.IndexName);
   DisposeStr(IRec^.IndexExpresion);
   Dispose(IRec);
   IRec:=IndexDefs;
  end;
  inherited Done;
end;

procedure TDBFTableConstructor.AddField(FieldName:string; FieldType:TFieldType; ASize, ADec:integer);
var PF:PCFieldRec;
begin
  New(PF);
  PF^.FieldName:=NewStr(FieldName);
  PF^.FieldType:=FieldType;
  PF^.Size:=ASize;
  PF^.Dec:=ADec;
  PF^.Next:=FieldDefs;
  FieldDefs:=PF;
end;

procedure TDBFTableConstructor.AddIndex(IndexName, IndexExpresion:String; IndexOptions:TIndexOptions);
var IRec:PIndexRec;
begin
  New(IRec);
  IRec^.IndexName:=NewStr(IndexName);
  IRec^.IndexExpresion:=NewStr(IndexExpresion);
  IRec^.IndexOpt:=IndexOptions;
  IRec^.Next:=IndexDefs;
  IndexDefs:=IRec;
end;

function TDBFTableConstructor.CreateTable:PDBFTable;
var PF:PCFieldRec;
    IRec:PIndexRec;
    i,j:integer;
    FieldArray:PFieldArray;
    D:PDbfTable;
begin
{  CreateTable:=nil;}
  j:=0;
  PF:=FieldDefs;
  while PF<>nil do
  begin
    inc(j);
    PF:=PF^.Next;
  end;
  GetMem(FieldArray, SizeOf(TFieldRec)*j);

  i:=1;
  PF:=FieldDefs;
  while PF<>nil do
  begin
    with FieldArray^[j-i] do
    begin
      Name:=NewStr(PF^.FieldName^);
      fType:=fTypeToChar(PF^.FieldType);
      fLength:=PF^.Size;
      fDecimal:=PF^.Dec;
    end;
    PF:=PF^.Next;
    inc(i);
  end;
  D:=New(PDbfTable, Create(TableName^, FieldArray^, j, TableType, SubType));
  if D<>nil then
  begin
    IRec:=IndexDefs;
    while IRec<>nil do
    begin
      D^.CreateIndex(IRec^.IndexName^, IRec^.IndexExpresion^, IRec^.IndexOpt);
      IRec:=IRec^.Next;
    end;
  end;
  FreeMem(FieldArray, SizeOf(TFieldRec)*j);
  CreateTable:=D;
end;

var
  OldDBOpen:TDBOpenProc;

function DBOpenProc(AFileName:FNameStr):PDataSet;{$IFNDEF FPC}far;{$ENDIF}
var P:PDataSet;
    FName:FNameStr;
begin
  P:=nil;
  FName:=ReplaceExt(FullName(AFileName), sDbfExt, false);
  if (ExtractExt(FName)=sDbfExt) and FileExists(FName) then
     P:=New(PDBFTable, Init(AFileName, stOpen));
  if (P=nil) and Assigned(OldDBOpen) then P:=OldDBOpen(AFileName);
  DBOpenProc:=P;
end;

begin
  OldDBOpen:=DBOpen;
  {$IFDEF FPC}
  DBOpen:=@DBOpenProc;
  {$ELSE}
  DBOpen:=DBOpenProc;
  {$ENDIF}
end.
