unit DBUtils;
interface
uses Objects, DAE, DBFTable, Ext_Type;

type
  PReindexItems = ^TReindexItems;
  TReindexItems = record
    Next:PReindexItems;
    NameIndex:PString;
    IndexExpression:PString;
    IndexOptions:TIndexOptions;
  end;

  PReindexRecord = ^TReindexRecord;
  TReindexRecord = record
    NameDbf:PString;
    Items:PReindexItems;
    Next:PReindexRecord;
  end;

type
  TCheckIndexProc = procedure(Dbf:PDbfTable; IndexName:FNameStr);
procedure dbuPackTable(var Table:PDbfTable; Proc:TProgresEvent);
procedure CheckReindex(Item:PReindexRecord; DirectReindex:boolean;
          InfoProc:TCheckIndexProc; DoReindexProc:TProgresEvent);

function NewReindexRec(ADbfName:FNameStr; Items:PReindexItems;
                       Next:PReindexRecord):PReindexRecord;
function NewReindexItems(IndexName:FNameStr; IndexExpression:string;
                IndexOptinos:TIndexOptions;
                Next:PReindexItems):PReindexItems;
{頥  ⠡   ப}
function TableVersion(Table:PDbfTable; DefString:string):string;
function TableExists(FileName:FNameStr):boolean;
implementation
uses Ext_Dos, Clases, Dos, Ext_Math;

procedure dbuPackTable(var Table:PDbfTable; Proc:TProgresEvent);
var NewTable:PDbfTable;
    OldName, NewName:FNameStr;
    Mode:word;
    L:Longint;

procedure DoCreateFile;
var SS:PDosStream;
    L:Longint;
begin
   Table^.CopyStructure(NewName);
   NewTable:=New(PDbfTable, Init(NewName, stOpen));
   if NewTable^.DbfFile^.Status<>stOk then
      NewTable^.DbfFile^.Reset;
   NewTable^.FileHeader^.RecordCount:=0;
   NewTable^.CodePageDrv:=Table^.CodePageDrv;
end;

begin
  if Table<>nil then
  begin
    if Table^.DbfFile^.CountUse<2 then
    begin
      Mode:=Table^.Mode;
      OldName:=Table^.TableName^;
      NewName:=GetTempName(FullName(Table^.TableName^));
      DoCreateFile;
      for L:=1 to Table^.RecordCount do
      begin
        Table^.Go(L);
        if not Table^.DeletedRec then
        begin
          NewTable^.Append;
          Move(Table^.RecBuf^,NewTable^.RecBuf^, NewTable^.RecordSize);
          NewTable^.Post;
        end;
        if Assigned(Proc) then Proc(Table^.RecordCount,L,Table^.TableName);
      end;
      Table^.Close;
      NewTable^.RenameTable(Table^.TableName^);
      NewTable^.Free;
      Table^.Open;
    end
    else
      DBError(edbNonExclus, Table^.TableName);
  end
  else
     SingleError(edbNoError);
end;

function NewReindexRec(ADbfName:FNameStr; Items:PReindexItems;
                       Next:PReindexRecord):PReindexRecord;
var P:PReindexRecord;
begin
  New(P);
  P^.Next:=Next;
  P^.NameDbf:=NewStr(ADbfName);
  P^.Items:=Items;
  NewReindexRec:=P;
end;

function NewReindexItems(IndexName:FNameStr; IndexExpression:string;
               IndexOptinos:TIndexOptions;
               Next:PReindexItems):PReindexItems;
var P:PReindexItems;
begin
  New(P);
  P^.Next:=Next;
  P^.NameIndex:=NewStr(IndexName);
  P^.IndexExpression:=NewStr(IndexExpression);
  P^.IndexOptions:=IndexOptinos;
  NewReindexItems:=P;
end;

procedure CheckReindex(Item:PReindexRecord; DirectReindex:boolean;
          InfoProc:TCheckIndexProc; DoReindexProc:TProgresEvent);
function IndexValid:boolean;
var
    L:PReindexRecord;
    P:PReindexItems;
begin
  IndexValid:=false;
  L:=Item;
  while L<>nil do
  begin
    P:=L^.Items;
    while P<>nil do
    begin
      if (P^.NameIndex=nil) or (not TableExists(P^.NameIndex^)) then exit;
      P:=P^.Next;
    end;
    L:=L^.Next;
  end;
  IndexValid:=true;
end;

var P:TProgresEvent;
    R,R1:PReindexItems;
    L:PReindexRecord;
    D:PDbfTable;
begin
  if (not DirectReindex) and IndexValid then exit;
  P:=DoReindex;
  DoReindex:=DoReindexProc;
  while Item<>nil do
  begin
    D:=New(PDbfTable, Init(Item^.NameDbf^, stOpenRead));
    R:=Item^.Items;
    while R<>nil do
    begin
      if Assigned(InfoProc) then  InfoProc(D, R^.NameIndex^);
      D^.CreateIndex(R^.NameIndex^, R^.IndexExpression^, R^.IndexOptions);
      R1:=R^.Next;
      DisposeStr(R^.NameIndex);
      DisposeStr(R^.IndexExpression);
      Dispose(R);
      R:=R1;
    end;
    D^.Free;
    L:=Item^.Next;
    DisposeStr(Item^.NameDbf);
    Dispose(Item);
    Item:=L;
  end;
  DoReindex:=P;
end;

function TableVersion(Table:PDbfTable; DefString:string):string;
var Ver:PMemoFilesRecord;
begin
  Ver:=MemoFilesCreatedList;
  while Ver<>nil do
    if Ver^.ByteID=Table^.FileHeader^.Version then break
    else Ver:=Ver^.Next;
  if (Ver<>nil) then TableVersion:=Ver^.SystemName^
  else TableVersion:=DefString;
end;

function TableExists(FileName:FNameStr):boolean;
begin
  TableExists:=FileExists(FullName(FileName));
end;
end.
