{*******************************************************}
{                                                       }
{       Turbo Pascal 7.0                                }
{       FreePacal                                       }
{       Delphi 4                                        }
{       New index object                                }
{                                                       }
{       Copyright (c) 1995-99 by Alexs                  }
{                                                       }
{*******************************************************}

{$IFNDEF FPC}
{$O-}
{$ENDIF}
{$I Define.inc}
unit MIX_Unit;
interface
uses Objects, DB, IND_Unit;
const
  defMixExt = '.MIX';
  Sign = $58494D78;
  PageSize = $800;

type
  PMIXGlobalHeader = ^TMIXGlobalHeader;
  TMIXGlobalHeader = record
    Signature:Longint;
    FileName:FNameStr;
    FirstFree:Longint;
    FistNode:Longint;
    FirstTag:Longint;
  end;

type
  PMIXIndexTag = ^TMIXIndexTag;
  TMIXIndexTag = record
    ItemSize:integer;
    KeySize:word;
    PageSize:integer;
    HalfPage:integer;
    Descending:Boolean;
    Unique:boolean;
    CaseSensitive:Boolean;
    KeyExp:string;
    NextTag:Longint;
    {!!!this field in memory only!!!}
    NextTagPtr:PMIXIndexTag;
  end;

type
  TMIXStream = object(TIndexStream)
     Header:TMIXGlobalHeader;
     IndexTag:PMIXIndexTag;
     constructor Init(AFileName:FNameStr; Mode:word);
     destructor Done;virtual;
     function NewPage:longint;virtual;
     procedure FreePage(L:Longint);virtual;
     procedure WriteHeader;virtual;
     procedure ReadHeader;
     procedure DoDisposeTags;
     procedure AddIndexTag(P:PMIXIndexTag);
     procedure RemoveIndexTag(P:PMIXIndexTag);
  end;

implementation

constructor TMIXStream.Init(AFileName:FNameStr; Mode:word);
begin
  inherited Init(AFileName, Mode);
  if Mode<>stCreate then ReadHeader
  else
  begin
    Header.Signature:=Sign;
    NewPage;
    WriteHeader;
  end;
end;

destructor TMIXStream.Done;
begin
  inherited Done;
  DoDisposeTags;
end;

function TMIXStream.NewPage:longint;
var P:Pointer;
begin
  Modified:=true;
  if Header.FirstFree<>0 then
  begin
    NewPage:=Header.FirstFree;
    Seek(Header.FirstFree);
    Read(Header.FirstFree, SizeOf(Longint));
  end
  else
  begin
    NewPage:=GetSize;
    Seek(GetSize);
    GetMem(P, PageSize);
    FillChar(P^, PageSize, 0);
    Write(P^, PageSize);
    FreeMem(P, PageSize);
  end;
end;

procedure TMIXStream.FreePage(L:Longint);
begin
end;

procedure TMIXStream.WriteHeader;
var L:Longint;
    P,PP:PMIXIndexTag;
begin
  Seek(0);
  Write(Header, SizeOf(TMIXGlobalHeader));
  L:=Header.FirstTag;
  P:=IndexTag;
  while P<>nil do
  begin
     Seek(L);
     Write(P^, SizeOf(TMIXIndexTag) - SizeOf(PMIXIndexTag));
     L:=P^.NextTag;
     P:=P^.NextTagPtr;
  end;
end;

procedure TMIXStream.ReadHeader;
var L:Longint;
    P,PP:PMIXIndexTag;
begin
  Seek(0);
  Read(Header, SizeOf(TMIXGlobalHeader));
  L:=Header.FirstTag;
  while L<>0 do
  begin
     New(P);
     Seek(L);
     Read(P^, SizeOf(TMIXIndexTag) - SizeOf(PMIXIndexTag));
     P^.NextTagPtr:=nil;
     L:=P^.NextTag;
     if IndexTag<>nil then
     begin
       PP:=IndexTag;
       while PP^.NextTagPtr<>nil do PP:=PP^.NextTagPtr;
       PP^.NextTagPtr:=PP;
     end
     else IndexTag:=P;
  end;
end;

procedure TMIXStream.DoDisposeTags;
var P:PMIXIndexTag;
begin
  if IndexTag=nil then exit;
  while IndexTag<>nil do
  begin
    P:=IndexTag^.NextTagPtr;
    Dispose(IndexTag);
    IndexTag:=P;
  end;
end;

procedure TMIXStream.AddIndexTag(P:PMIXIndexTag);
var L:Longint;
    PP:PMIXIndexTag;
begin
  if IndexTag<>nil then
  begin
    PP:=IndexTag;
    while PP^.NextTagPtr<>nil do PP:=PP^.NextTagPtr;
    PP^.NextTagPtr:=P;
    PP^.NextTag:=NewPage;
  end
  else
  begin
    Header.FirstTag:=NewPage;
    IndexTag:=P;
  end;
end;

procedure TMIXStream.RemoveIndexTag(P:PMIXIndexTag);
begin
end;

end.