{*******************************************************}
{                                                       }
{       Turbo Pascal 7.0                                }
{       FreePacal                                       }
{       Delphi 4                                        }
{       TV index object                                 }
{                                                       }
{       Copyright (c) 1995-99 by Alexs                  }
{                                                       }
{*******************************************************}

{$IFNDEF FPC}
{$O+}
{$ENDIF}
{$I Define.inc}
unit NTX_Unit;
interface
uses Objects, DAE, Dos, Ext_Type, DBFTable, DAE_Type;
const
  NTXPageSize = 1024;
  defNtxExt:string[4] = '.NTX';

type
  PNTXHeader = ^TNTXHeader;
  TNTXHeader = packed record
    Sign:word;
    Version:word;
    RootNode:Longint;
    EmptyPage:Longint;
    ItemSize:word;
    KeySize:word;
    KeyDec:word;
    MaxItem:word;
    HalfPage:word;
    KeyExp:array [1..256] of char;
    Unique:boolean;
  end;

type
  PNTXStream = ^TNTXStream;
  TNTXStream = object(TIndexStream)
     Header:TNTXHeader;
     constructor Init(AFileName:FNameStr; Mode:word);
     destructor Done;virtual;
     function NewPage:longint;virtual;
     procedure FreePage(L:Longint);virtual;
     procedure WriteHeader;virtual;
  end;

type
  PNtxNode = ^TNtxNode;
  TNtxNode = object(TNode)
     procedure Store;virtual;
     procedure Load(ARef:Longint);virtual;
     function Count:word;
     function Key(Num:integer):string;
     function PageRef(Num:integer):longint;
     function RecordNumber(Num:integer):longint;
     function LastRef:longint;
     function Search(S:string; PtrRecord:longint; var Num:integer):boolean;
     function Delete( Num:integer):boolean;

     procedure SetLastRef(L:longint);
     procedure SetCount(ACount:word);

     procedure GetRecord(Num:integer; var Rec:TPageRecord);
     procedure SetRecord(Num:integer; var Rec:TPageRecord);
  end;

type
  PNTXIndex = ^TNTXIndex;
  TNTXIndex = object(TIndexObject)
     constructor Init(ALink:PDataSet; AFileName:FNameStr);
     constructor Create(ALink:PDataSet; AFileName:FNameStr; AField:string;
                        AOptions:TIndexOptions);
     function KeySize:word;virtual;
     function NextKey:Longint;virtual;
     function PrevKey:Longint;virtual;
     function Unique:Boolean;virtual;
     function FindKey(Key:string; PtrRec:longint; SoftFind:Boolean):LongInt;virtual;
     function FirstKey:Longint;virtual;
     function LastKey:Longint;virtual;
     function IndexExpresion:string;virtual;
     function GetNode(ARef:Longint):PNode;virtual;
{     function Curentkey:longint;virtual;}
     procedure EmptyIndex; virtual;
     procedure AddKey( AKey:String; ALink:LongInt);virtual;
     procedure DeleteKey( AKey:String; ALink:longint);virtual;
  end;

function NTXOpen(FileName:FNameStr; var Ext:ExtStr; DS:PDataSet):PIndexObject;far;
function NTXCreate( FileName:FNameStr; var Ext:ExtStr; DS:PDataSet;
                    AField:string; AOptions:TIndexOptions):PIndexObject;far;
implementation
uses Strings, Ext_Str, Ext_Dos;

  { TNTXStream }
constructor TNTXStream.Init(AFileName:FNameStr; Mode:word);
begin
  inherited Init(AFileName, Mode);
  if Mode<>stCreate then Read(Header, SizeOf(TNTXHeader));
end;

destructor TNTXStream.Done;
begin
  WriteHeader;
  inherited Done;
end;

procedure TNTXStream.WriteHeader;
var P:Pointer;
begin
  if Modified then
  begin
    GetMem(P, NTXPageSize);
    FillChar(P^, NTXPageSize, 0);
    Move(Header, P^, SizeOf(TNTXHeader));
    Seek(0);
    Write(P^, NTXPageSize);
    FreeMem(P, NTXPageSize);
  end;
end;
{$IFDEF VER70}
function TNTXStream.NewPage:longint;
var P:Pointer;
begin
  Modified:=true;
  if Header.EmptyPage<>0 then
  begin
    NewPage:=Header.EmptyPage;
    Seek(Header.EmptyPage);
    Read(Header.EmptyPage, SizeOf(Longint));
  end
  else
  begin
    NewPage:=GetSize;
    Seek(GetSize);
    GetMem(P, NTXPageSize);
    FillChar(P^, NTXPageSize, 0);
    asm
        LES     DI,Self
        MOV     CX,ES:[DI].TNTXStream.Header.MaxItem
        MOV     BX,ES:[DI].TNTXStream.Header.ItemSize
        INC     CX
        MOV     AX,CX
        SHL     AX,1
        ADD     AX,2
        LES     DI,P
        ADD     DI,2
@@1:    STOSW
        ADD     AX,BX
        LOOP    @@1
    end;
    Write(P^, NTXPageSize);
    FreeMem(P, NTXPageSize);
  end;
end;
{$ELSE}
function TNTXStream.NewPage:longint;
begin
  abstract;
end;
{$ENDIF}
procedure TNTXStream.FreePage(L:Longint);
begin
  Seek(L);
  Write(Header.EmptyPage, SizeOf(Longint));
  Header.EmptyPage:=L;
end;

    { TNtxNode }
procedure TNtxNode.Store;
begin
  inherited Store;
  Stream^.Seek(Handle);
  Stream^.Write(Buffer^, NTXPageSize)
end;

procedure TNtxNode.Load(ARef:Longint);
begin
  inherited Load(ARef);
  Stream^.Seek(ARef);
  Stream^.Read(Buffer^, NTXPageSize);
end;
{$IFDEF VER70}
function TNtxNode.Count:word;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     AX,ES:[DI]
end;

procedure TNtxNode.SetCount(ACount:word);assembler;
asm
        MOV     AX,ACount
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     ES:[DI],AX
end;

function TNtxNode.Key(Num:integer):string;assembler;
asm
        PUSH    DS
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Stream
        MOV     CX,ES:[DI].TNTXStream.Header.KeySize
        MOV     BX,Num
        SHL     BX,1
        ADD     BX,2
        LDS     SI,Self
        LDS     SI,DS:[SI].TNode.Buffer
        MOV     AX,DS:[SI+BX]
        ADD     AX,8
        ADD     SI,AX
        LES     DI,@Result
        CLD
        MOV     AL,CL
        STOSB
    REP MOVSB
        POP     DS
end;

function TNtxNode.PageRef(Num:integer):longint;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     BX,Num
        SHL     BX,1
        ADD     BX,2
        ADD     DI,ES:[DI+BX]
        MOV     AX,ES:[DI]
        MOV     DX,ES:[DI+2]
end;

function TNtxNode.RecordNumber(Num:integer):longint;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     BX,Num
        SHL     BX,1
        ADD     BX,2
        MOV     AX,ES:[DI+BX]
        ADD     AX,4
        ADD     DI,AX
        MOV     AX,ES:[DI]
        MOV     DX,ES:[DI+2]
end;

function TNtxNode.LastRef:longint;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     BX,ES:[DI]
        SHL     BX,1
        ADD     BX,2
        MOV     AX,ES:[DI+BX]
        ADD     DI,AX
        MOV     AX,ES:[DI]
        MOV     DX,ES:[DI+2]
end;


procedure TNtxNode.SetLastRef(L:longint);assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     BX,ES:[DI]
        SHL     BX,1
        ADD     BX,2
        MOV     AX,ES:[DI+BX]
        ADD     DI,AX
        MOV     AX,L.WORD[0]
        MOV     ES:[DI],AX
        MOV     AX,L.WORD[2]
        MOV     ES:[DI+2],AX
end;

procedure TNtxNode.GetRecord(Num:integer; var Rec:TPageRecord);assembler;
asm
        PUSH    DS
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Stream
        MOV     CX,ES:[DI].TNTXStream.Header.ItemSize
        LES     DI,Rec
        LDS     SI,Self
        LDS     SI,DS:[SI].TNode.Buffer
        MOV     BX,Num
        SHL     BX,1
        ADD     BX,2
        MOV     AX,DS:[SI+BX]
        ADD     SI,AX
    REP MOVSB
        POP     DS
end;

procedure TNtxNode.SetRecord(Num:integer; var Rec:TPageRecord);assembler;
asm
        PUSH    DS
        LDS     SI,Rec
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Stream
        MOV     CX,ES:[DI].TNTXStream.Header.ItemSize
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     BX,Num
        SHL     BX,1
        ADD     BX,2
        MOV     AX,ES:[DI+BX]
        ADD     DI,AX
    REP MOVSB
        POP     DS
end;
{$ELSE}
function TNtxNode.Count:word;
begin
end;
procedure TNtxNode.SetCount(ACount:word);
begin
end;
function TNtxNode.Key(Num:integer):string;
begin
end;
function TNtxNode.PageRef(Num:integer):longint;
begin
end;
function TNtxNode.RecordNumber(Num:integer):longint;
begin
end;
function TNtxNode.LastRef:longint;
begin
end;
procedure TNtxNode.SetLastRef(L:longint);
begin
end;
procedure TNtxNode.GetRecord(Num:integer; var Rec:TPageRecord);
begin
end;
procedure TNtxNode.SetRecord(Num:integer; var Rec:TPageRecord);
begin
end;
{$ENDIF}
function TNtxNode.Search(S:string; PtrRecord:longint; var Num:integer):boolean;
var
  L,H:integer;
  C:Longint;
begin
  L := 0;
  H := Count - 1;
  while L <= H do
  begin
    Num := (L + H) shr 1;
    C := ASCIICompare(Key(Num),S);
    if (C=0) and (PtrRecord<>-1) then C:=RecordNumber(Num)-PtrRecord;
    if C<=0 then L := Num + 1;
    if C>=0 then H := Num - 1;
  end;
  Search:=C=0;
  if (C<>0) then Num:=L;
end;

function TNtxNode.Delete( Num:integer):boolean;
var i:integer;
    PgRec:TPageRecord;
    _LastRef:Longint;
begin
  _LastRef:=LastRef;
  for i:=Num to Count-2 do
  begin
    GetRecord(i+1,PgRec);
    SetRecord(i,PgRec);
  end;
  SetCount(Count-1);
  SetLastRef(_LastRef);
end;

  {TNTXIndex}
constructor TNTXIndex.Init(ALink:PDataSet; AFileName:FNameStr);
var
  AOptions:TIndexOptions;
begin
  inherited Init(ALink);
  AFileName:=ReplaceExt(AFileName, defNtxExt,false);
  IndexStream:=PNTXStream(Session^.GetDBStream(AFileName));
  if IndexStream=nil then
     IndexStream:=New(PNTXStream, Init(AFileName, stOpen));
  if IndexStream=nil then Fail;
  FileName:=PIndexStream(IndexStream)^.FileName;
  IndexField:=GetIndexsField(Owner, IndexExpresion, AOptions);
  Options:=Options+AOptions;
  if Unique then Options:=Options+[ixUnique];
end;

constructor TNTXIndex.Create(ALink:PDataSet; AFileName:FNameStr; AField:string;
                        AOptions:TIndexOptions);
begin
  AFileName:=ReplaceExt(AFileName, defNtxExt, false);
  inherited Create(ALink, AOptions);
  IndexField:=GetIndexsField(Owner, AField, AOptions);
  Options:=Options+AOptions;
  IndexStream:=PNTXStream(Session^.GetDBStream(AFileName));
  if (IndexStream<>nil) or (IndexField=nil) then
  begin
     Session^.FreeDBStream(PIndexStream(IndexStream));
     if ixExpression in Options then IndexField^.Free;
     if IndexStream<>nil then DBError(edbFileAlreadyOpen, AFileName);
     fail;
   end;
   IndexStream:=New(PNTXStream, Init(AFileName, stCreate));
   FileName:=PIndexStream(IndexStream)^.FileName;
   with PNTXStream(IndexStream)^.Header do
   begin
    Sign:=$0006;
    Version:=$0001;
    RootNode:=0;
    KeySize:=IndexField^.Size;
    ItemSize:=KeySize+SizeOf(Longint)*2;
    KeyDec:= IndexField^.Dec;
    MaxItem:= (NTXPageSize-2) div (ItemSize+2) and $FFFE;
    HalfPage:= MaxItem div 2;
    StrPCopy(@KeyExp, AField);
    Unique:=ixUnique in Options;
   end;
   PIndexStream(IndexStream)^.Modified:=true;
   PNTXStream(IndexStream)^.WriteHeader;
   Reindexs;
end;

function TNTXIndex.FindKey(Key:string; PtrRec:longint; SoftFind:Boolean):LongInt;
var
  Node:PNtxNode;
  L:integer;
  Ref:longint;
begin
  ClearKey;
  FindKey:=-1;
  Ref:=PNTXStream(IndexStream)^.Header.RootNode;
  while Ref<>0 do
  begin
    Node:=GetPage(@Self, Ref);
    inc(pp);
    PagePtrArray[PP].Ref:=Ref;
    if Node^.Search(Key, PtrRec, L) then
    begin
      FindKey:=Node^.RecordNumber(L);
      PagePtrArray[PP].CurItem:=L;
      if EndRange(Node^.Key(PagePtrArray[PP].CurItem)) then iStatus:=iStatus+[iBofFind];
      if StartRange(Node^.Key(PagePtrArray[PP].CurItem)) then iStatus:=iStatus+[iEofFind];
      exit;
    end
    else
    begin
      PagePtrArray[PP].CurItem:=L;
      if L>=Node^.Count then Ref:=Node^.LastRef
      else Ref:=Node^.PageRef(L);
    end;
  end;
  if SoftFind then FindKey:=Node^.RecordNumber(L);
  if StartRange(Node^.Key(PagePtrArray[PP].CurItem)) then iStatus:=iStatus+[iEofFind];
  if EndRange(Node^.Key(PagePtrArray[PP].CurItem)) then iStatus:=iStatus+[iBofFind];
end;

function TNTXIndex.FirstKey:Longint;
var
  Node:PNtxNode;
  Ref:longint;
  S:string;
begin
  ClearKey;
  S:=StartRangeValue;
  if S<>'' then FirstKey:=FindKey(S,0, true)
  else
  begin
     Ref:=PNTXStream(IndexStream)^.Header.RootNode;
     while Ref<>0 do
     begin
       Node:=GetPage(@Self, Ref);
       inc(pp);
       PagePtrArray[PP].Ref:=Ref;
       PagePtrArray[PP].CurItem:=0;
       Ref:=Node^.PageRef(0);
     end;
     FirstKey:=Node^.RecordNumber(0);
  end;
end;

function TNTXIndex.LastKey:Longint;
var
  Node:PNtxNode;
  L:integer;
  Ref:longint;
  S:String;
begin
  ClearKey;
  S:=EndRangeValue;
  if S<>'' then LastKey:=FindKey(S, MaxLongint, true)
  else
  begin
    Ref:=PNTXStream(IndexStream)^.Header.RootNode;
    while Ref<>0 do
    begin
      Node:=GetPage(@Self, Ref);
      inc(pp);
      PagePtrArray[PP].Ref:=Ref;
      PagePtrArray[PP].CurItem:=Node^.Count;
      Ref:=Node^.LastRef;
    end;
    LastKey:=Node^.RecordNumber(Node^.Count-1);
  end;
end;

function TNTXIndex.IndexExpresion:string;
begin
  IndexExpresion:=StrPas(@PNTXStream(IndexStream)^.Header.KeyExp)
end;

function TNTXIndex.KeySize:word;
begin
  KeySize:=PNTXStream(IndexStream)^.Header.KeySize;
end;

function TNTXIndex.NextKey:Longint;
var Node:PNtxNode;
    L:longint;
begin
  if iEofFind in iStatus then exit;
  Move(PagePtrArray,OldPagePtrArray,SizeOf(TPagePtrArray));
  OldPP:=PP;
  iStatus:=iStatus - [iEofFind];
  if PP = 0 then L:=PNTXStream(IndexStream)^.Header.RootNode
  else
  begin
    Inc(PagePtrArray[PP].CurItem);
    Node:=GetPage(@Self, PagePtrArray[PP].Ref);
    if PagePtrArray[PP].CurItem<Node^.Count then
       L:=Node^.PageRef(PagePtrArray[PP].CurItem)
    else L:=Node^.LastRef;
  end;
  while L<>0 do
  begin
    Node:=GetPage(@Self, L);
    inc(PP);
    PagePtrArray[PP].Ref:=L;
    PagePtrArray[PP].CurItem:=0;
    L:=Node^.PageRef(0);
  end;
  if PP<>0 then
  begin
    while (PP>1) and (PagePtrArray[PP].CurItem>Node^.Count-1) do
    begin
      Dec(PP);
      Node:=GetPage(@Self, PagePtrArray[PP].Ref);
    end;
    if PagePtrArray[PP].CurItem<Node^.Count then
    begin
       NextKey:=Node^.RecordNumber(PagePtrArray[PP].CurItem);
       if EndRange(Node^.Key(PagePtrArray[PP].CurItem)) then PP:=0;
    end
    else PP:=0;
  end;
  if PP=0 then
  begin
    iStatus:=iStatus + [iEofFind];
    Move(OldPagePtrArray,PagePtrArray,SizeOf(TPagePtrArray));
    PP:=OldPP;
  end
  else iStatus:=iStatus - [iBofFind];
end;

function TNTXIndex.PrevKey:Longint;
var Node:PNtxNode;
    L:longint;
begin
  if iBofFind in iStatus then exit;
  Move(PagePtrArray,OldPagePtrArray,SizeOf(TPagePtrArray));
  OldPP:=PP;
  iStatus:=iStatus - [iBofFind];
  if PP = 0 then L:=PNTXStream(IndexStream)^.Header.RootNode
  else
  begin
    Node:=GetPage(@Self, PagePtrArray[PP].Ref);
    if PagePtrArray[PP].CurItem<Node^.Count then
       L:=Node^.PageRef(PagePtrArray[PP].CurItem)
    else L:=Node^.LastRef;
  end;
  while L<>0 do
  begin
    Node:=GetPage(@Self, L);
    inc(PP);
    PagePtrArray[PP].Ref:=L;
    PagePtrArray[PP].CurItem:=Node^.Count;
    L:=Node^.LastRef;
  end;
  if PP<>0 then
  begin
    while (PP>1) and (PagePtrArray[PP].CurItem<1) do
    begin
      Dec(PP);
      Node:=GetPage(@Self, PagePtrArray[PP].Ref);
    end;
    if PagePtrArray[PP].CurItem>0 then
    begin
      Dec(PagePtrArray[PP].CurItem);
      PrevKey:=Node^.RecordNumber(PagePtrArray[PP].CurItem);
      if StartRange(Node^.Key(PagePtrArray[PP].CurItem)) then PP:=0;
    end else PP:=0;
  end;
  if PP=0 then
  begin
    iStatus:=iStatus + [iBofFind];
    Move(OldPagePtrArray,PagePtrArray,SizeOf(TPagePtrArray));
    PP:=OldPP;
  end
  else iStatus:=iStatus - [iEofFind];
end;

function TNTXIndex.Unique:Boolean;
begin
  Unique:=PNTXStream(IndexStream)^.Header.Unique;
end;

function TNTXIndex.GetNode(ARef:Longint):PNode;
begin
  GetNode:=New(PNtxNode, Init(IndexStream, ARef, NTXPageSize));
end;

procedure TNTXIndex.EmptyIndex;
begin
  CloseStream(IndexStream);
  PNTXStream(IndexStream)^.Header.RootNode:=0;
  PNTXStream(IndexStream)^.Header.EmptyPage:=0;
  PNTXStream(IndexStream)^.WriteHeader;
  IndexStream^.Truncate;
end;

procedure TNTXIndex.AddKey( AKey:String; ALink:LongInt);
var Node, NewNode:PNtxNode;
    isInsert:boolean;
    PgRec1,PgRec2, PgRec3:TPageRecord;

procedure DoAdd(Ref:Longint);
var L:integer;

procedure Insert;
var LastRef:Longint;
    i:integer;
begin
  Node:=PNtxNode(GetPage(@Self, Ref));
  LastRef:=Node^.LastRef;
  if Node^.Count<PNTXStream(IndexStream)^.Header.MaxItem then
  begin
    for i:=Node^.Count-1 downto L do
    begin
       Node^.GetRecord(i, PgRec3);
       Node^.SetRecord(i+1, PgRec3);
    end;
    Node^.SetRecord(L, PgRec1);
    Node^.SetCount(Node^.Count+1);
    isInsert:=false;
  end
  else
  with PNTXStream(IndexStream)^.Header do
  begin
    NewNode:=PNtxNode(GetPage(@Self, PIndexStream(IndexStream)^.NewPage));
    for i:=0 to HalfPage-1 do
    begin
      Node^.GetRecord(i, PgRec3);
      NewNode^.SetRecord(i, PgRec3);
      Node^.GetRecord(i+HalfPage, PgRec3);
      Node^.SetRecord(i, PgRec3);
    end;
    if L<=HalfPage then
    begin
      if L=HalfPage then Move(PgRec1,PgRec2, SizeOf(TPageRecord))
      else
      begin
        NewNode^.GetRecord(HalfPage-1,PgRec2);
        for i:=HalfPage-1 downto L+1 do
        begin
          NewNode^.GetRecord(i-1,PgRec3);
          NewNode^.SetRecord(i,PgRec3);
        end;
        NewNode^.SetRecord(L,PgRec1);
      end;
    end
    else
    begin
       L:=L-HalfPage-1;
       Node^.GetRecord(0,PgRec2);
       for i:=0 to L-1 do
       begin
         Node^.GetRecord(i+1,PgRec3);
         Node^.SetRecord(i,PgRec3);
       end;
       Node^.SetRecord(L,PgRec1);
    end;
    Node^.SetCount(HalfPage);
    NewNode^.SetCount(HalfPage);
    NewNode^.SetLastRef(PgRec2.LeftPage);
    PgRec2.LeftPage:=NewNode^.Handle;
    PgRec1:=PgRec2;
    NewNode^.Modified:=true;
  end;
  Node^.SetLastRef(LastRef);
  Node^.Modified:=true;
end;

begin
  if Ref=0 then
  begin
    isInsert:=true;
    PgRec1.LeftPage:=0;
    PgRec1.PtrRecord:=ALink;
    move(AKey[1], PgRec1.Key, byte(AKey[0]));
  end
  else
  begin
    Node:=PNtxNode(GetPage(@Self, Ref));
    Node^.Search(AKey, ALink, L);
    inc(PP);
    if L>=Node^.Count then PagePtrArray[PP].Ref:=Node^.LastRef
    else PagePtrArray[PP].Ref:=Node^.PageRef(L);
    PagePtrArray[PP].CurItem:=L;
    DoAdd(PagePtrArray[PP].Ref);
    if isInsert then Insert;
  end;
end;

begin
  FillChar(PgRec1, SizeOf(TPageRecord), 0);
  FillChar(PgRec2, SizeOf(TPageRecord), 0);
  FillChar(PgRec3, SizeOf(TPageRecord), 0);
  PP:=0;
  DoAdd(PNTXStream(IndexStream)^.Header.RootNode);
  if isInsert then
  begin
    Node:=PNtxNode(GetPage(@Self, PIndexStream(IndexStream)^.NewPage));
    Node^.SetCount(1);
    Node^.SetRecord(0,PgRec1);
    Node^.SetLastRef(PNTXStream(IndexStream)^.Header.RootNode);
    PNTXStream(IndexStream)^.Header.RootNode:=Node^.Handle;
    Node^.Modified:=true;
  end;
end;

procedure TNTXIndex.DeleteKey( AKey:String; ALink:longint);
var
  PageTooSmall:Boolean;

procedure DelB(Ref:Longint);
var Node:PNtxNode;
    isFind:Boolean;
    L:integer;
    ARef:Longint;

procedure UnderFlow(_Ref:Longint);
begin
  Node:=PNtxNode(GetPage(@Self, _Ref));
  if Node^.Count=0 then FreePage(Node);
end;

procedure DelA(LRef:Longint);
var LNode:PNtxNode;
    PgRec:TPageRecord;
begin
  while LRef<>0 do
  begin
    inc(PP);
    PagePtrArray[PP].Ref:=LRef;
    LNode:=PNtxNode(GetPage(@Self, LRef));
    LRef:=LNode^.LastRef;
  end;
  LNode^.GetRecord(LNode^.Count-1, PgRec);
  Node^.SetRecord(L, PgRec);
  LNode^.SetCount(LNode^.Count-1);
  LNode^.SetLastRef(0);
  PageTooSmall:=LNode^.Count<PNTXStream(IndexStream)^.Header.HalfPage;
end;

begin
  if Ref = 0 then PageTooSmall:=false
  else
  begin
    inc(PP);
    PagePtrArray[PP].Ref:=Ref;
    Node:=PNtxNode(GetPage(@Self, Ref));
    isFind:=Node^.Search(AKey, ALink, L);
    if L>=Node^.Count then ARef:=Node^.LastRef
    else ARef:=Node^.PageRef(L);
    if isFind then
       if ARef=0 then
       begin
         Node^.Delete(L);
         PageTooSmall:=Node^.Count<PNTXStream(IndexStream)^.Header.HalfPage;
       end
       else
       begin
         DelA(ARef);
         if PageTooSmall then UnderFlow(Ref);
       end
    else
    begin
      DelB(ARef);
      if PageTooSmall then UnderFlow(Ref);
    end;
  end;
end;

begin
end;

const
   OldOnOpenIndex:TIndexOpen=nil;
   OldOnCreateIndex:TIndexCreate=nil;

function NTXOpen(FileName:FNameStr; var Ext:ExtStr; DS:PDataSet):PIndexObject;
begin
  if Ext=defNtxExt then
     NtxOpen:=New(PNTXIndex, Init(DS, FileName))
  else
  if Assigned(OldOnOpenIndex) then
     NTXOpen:=OldOnOpenIndex(FileName,Ext,DS)
  else NTXOpen:=nil;
end;

function NTXCreate( FileName:FNameStr; var Ext:ExtStr; DS:PDataSet;
                    AField:string; AOptions:TIndexOptions):PIndexObject;
begin
  if Ext=defNtxExt then
     NTXCreate:=New(PNtxIndex, Create(DS, FileName, AField, AOptions))
  else
  if Assigned(OldOnCreateIndex) then
     NTXCreate:=OldOnCreateIndex(FileName,Ext, DS, AField, AOptions)
  else NTXCreate:=nil;
end;

begin
   OldOnOpenIndex:=OnOpenIndex;
   OldOnCreateIndex:=OnCreateIndex;
   OnCreateIndex:=NTXCreate;
   OnOpenIndex:=NTXOpen;
end.
