{************************************************}
{                                                }
{   Turbo Pascal 7.0                             }
{   FreePacal                                    }
{   Delphi 4                                     }
{   Data access unit                             }
{   Copyright (c) 1995-2000 by Alexs             }
{                                                }
{************************************************}
{$IFDEF VER70}
{$O+}
{$ENDIF}
{$I Define.inc}

unit Querys;
interface
uses Objects, DAE, Ext_Type;
type
   PParam = ^TParam;
   TParam = object(TObject)
     constructor Init(const AName:string);
     destructor Done;virtual;
     constructor Load(var S:TStream);
     procedure Store(var S:TStream);
     function Name:string;
     function AsString: string;
     function AsBoolean: Boolean;
     function AsDateTime: TDateTime;
     function AsFloat: Double;
     function AsInteger: Longint;
     function IsEqual(Value: PParam): Boolean;
     function Text: string;
     procedure SetAsString(const Value: string);
     procedure SetAsBoolean(const Value: Boolean);
     procedure SetAsDateTime(const Value: TDateTime);
     procedure SetAsFloat(const Value: Double);
     procedure SetAsInteger(const Value: Longint);
     procedure SetText(const Value: string);
   private
     fName:PString;
     fValue:PString;
   end;

type
   PParamList = ^TParamList;
   TParamList = object(TSortedCollection)
      function Compare(Key1, Key2: Pointer): Integer; virtual;
      function KeyOf(Item: Pointer): Pointer; virtual;
   end;

type
  PQuery = ^TQuery;
  TQuery = object(TDataSet)
    ResultSet:PDataSet;
    Text:PChar;
    Params:PParamList;
    AllRecordFetched:boolean;
    Prepared:boolean;
    constructor Init(AText:PChar);
    destructor Done;virtual;
    function Open:boolean;virtual;
    procedure Close;virtual;
    procedure Conect;virtual;
    procedure Disconect;virtual;
    function ParamCount:integer;
    function ParamByName(Name:string):PParam;
    procedure ExecSql;virtual;
    procedure Prepare;virtual;
    procedure UnPrepare;virtual;
    function FieldCount:integer;virtual;

    procedure Go( APos:Longint);virtual;
    procedure Last;virtual;
    procedure First;virtual;
    procedure Next;virtual;
    procedure Prior;virtual;
    procedure MoveBy(Distance:longint);virtual;
    procedure FreeBookmark(Bookmark: PBookmark);virtual;
    function GetBookmark:PBookmark;virtual;
    procedure GotoBookmark(Bookmark: PBookmark);virtual;
    procedure FetchNextBlock;virtual;{$IFDEF EAM} abstract;{$ENDIF}
  end;

implementation
uses Ext_Str, Ext_Math, Ext_Dos, Strings;

   { TParam }
constructor TParam.Init(const AName:string);
begin
  inherited Init;
  fName:=NewStr(AName);
end;

destructor TParam.Done;
begin
  DisposeStr(fName);
  DisposeStr(fValue);
  inherited Done;
end;

constructor TParam.Load(var S:TStream);
begin
  inherited Init;
  fName:=S.ReadStr;
  fValue:=S.ReadStr;
end;

procedure TParam.Store(var S:TStream);
begin
  S.WriteStr(fName);
  S.WriteStr(fValue);
end;

function TParam.Name:string;
begin
  if fName<>nil then Name:=fName^
  else Name:='';
end;

function TParam.AsString: string;
begin
  if fValue<>nil then AsString:=fValue^
  else AsString:='';
end;

function TParam.AsBoolean: Boolean;
begin
  AsBoolean:=stUpCase(Copy(AsString,1,1))='T';
end;

function TParam.AsDateTime: TDateTime;
begin
  AsDateTime:=StrToDate(AsString);
end;

function TParam.AsFloat: Double;
begin
  AsFloat:=StrToFloat(AsString);
end;

function TParam.AsInteger: Longint;
begin
  AsInteger:=StrToInt(AsString);
end;

function TParam.IsEqual(Value: PParam): Boolean;
begin
  IsEqual:=AsString=Value^.AsString
end;

function TParam.Text: string;
begin
  Text:=AsString;
end;

procedure TParam.SetAsString(const Value: string);
begin
  AssignStr(fValue, Value);
end;

procedure TParam.SetAsBoolean(const Value: Boolean);
begin
  SetAsString(BooleanToStr(Value));
end;

procedure TParam.SetAsDateTime(const Value: TDateTime);
begin
  SetAsString(DateToStr(Value));
end;

procedure TParam.SetAsFloat(const Value: Double);
begin
  SetAsString(FloatToStr(Value));
end;

procedure TParam.SetAsInteger(const Value: Longint);
begin
  SetAsString(IntToStr(Value));
end;

procedure TParam.SetText(const Value: string);
begin
  SetAsString(Value);
end;

   { TParamList }
function TParamList.Compare(Key1, Key2: Pointer): Integer;
var P1:PString absolute Key1;
    P2:PString absolute Key2;
begin
  if stUpCase(P1^)<stUpCase(P2^) then Compare:=-1
  else
  if stUpCase(P1^)>stUpCase(P2^) then Compare:=1
  else Compare:=0;
end;

function TParamList.KeyOf(Item: Pointer): Pointer;
var P:PParam absolute Item;
begin
  KeyOf := P^.fName;
end;

  {TQuery}
constructor TQuery.Init(AText:PChar);
begin
  inherited Init;
  Text:=StrNew(AText);
  Params:=New(PParamList, Init(10,5));
end;

destructor TQuery.Done;
begin
  Close;
  StrDispose(Text);
  if Params<>nil then Dispose(Params, Done);
  Inherited Done;
end;

function TQuery.Open:boolean;
begin
  Conect;
end;

procedure TQuery.Close;
begin
  Disconect;
end;

procedure TQuery.Conect;
begin
  Prepare;
  ExecSql;
end;

procedure TQuery.Disconect;
begin
   AllRecordFetched:=false;
end;

function TQuery.ParamCount:integer;
begin
  ParamCount:=Params^.Count;
end;

procedure TQuery.ExecSql;
begin
end;

procedure TQuery.Prepare;
begin
end;

procedure TQuery.UnPrepare;
begin
end;

{function TQuery.BOF:boolean;
begin
  if ResultSet<>nil then BOF:=ResultSet^.Bof
  else BOF:=ResultSet^.Bof;
end;

function TQuery.EOF:boolean;
begin
  if ResultSet<>nil then EOF:=ResultSet^.Eof
  else EOF:=ResultSet^.Eof;
end;}

procedure TQuery.Go( APos:Longint);
begin
end;


procedure TQuery.Last;
begin
  if ResultSet=nil then exit;
  if ResultSet^.Eof then
  while not AllRecordFetched do FetchNextBlock;
  ResultSet^.Last;
end;

procedure TQuery.First;
begin
end;

procedure TQuery.Next;
begin
  if ResultSet=nil then exit;
  if ResultSet^.Eof then FetchNextBlock;
  ResultSet^.Last;
end;

procedure TQuery.Prior;
begin
end;

procedure TQuery.MoveBy(Distance:longint);
begin
end;

procedure TQuery.FreeBookmark(Bookmark: PBookmark);
begin
end;

function TQuery.GetBookmark:PBookmark;
begin
end;

procedure TQuery.GotoBookmark(Bookmark: PBookmark);
begin
end;

function TQuery.ParamByName(Name:string):PParam;
var P:PParam;
    i:integer;
begin
  ParamByName:=nil;
  if Params=nil then exit;
  if Params^.Search(@Name, i) then
     ParamByName:=Params^.At(i)
  else
     DBError(edbParamNotFound,Name);
end;

function TQuery.FieldCount:integer;
begin
  if ResultSet<>nil then FieldCount:=ResultSet^.FieldCount
  else FieldCount:=0;
end;

{$IFNDEF EAM}
procedure TQuery.FetchNextBlock;
begin abstract;
end;
{$ENDIF}

end.
