{************************************************}
{                                                }
{   Clipboard objects                            }
{                                                }
{   Turbo Pascal 7                               }
{   FreePacal                                    }
{   Delphi 2 - 5                                 }
{   Virtual Pascal 2.1                           }
{                                                }
{   Copyright (c) 1995-2000 by Alexs             }
{                                                }
{************************************************}

{$IFDEF VER70}
{$O+}
{$ENDIF}
{$I Define.inc}

unit Ext_Clp;
interface
uses Objects;
const
  winclpText            = $01;
  winclpBitmap          = $02;
  winclpMetafilePic     = $03;
  winclpSYLK            = $04;
  winclpDIF             = $05;
  winclpTIFF            = $06;
  winclpOEMText         = $07;
  winclpDIBBitmap       = $08;
  winclpSpecial         = $80;
  winclpDSPText         = $81;
  winclpDSPBitmap       = $82;

type
  PClipObject = ^TClipObject;
  TClipObject = object(TObject)
    constructor Init;
    destructor Done;virtual;
    procedure CopyToClp(var Buf; BufType:word; Len:word);virtual;{$IFDEF EAM}abstract;{$ENDIF}
    procedure PasteFromClp(var Buf; BufType:word);virtual;{$IFDEF EAM}abstract;{$ENDIF}
    procedure ClearClp;virtual;{$IFDEF EAM}abstract;{$ENDIF}
    function DataSize(BufType:word):word;virtual;{$IFDEF EAM}abstract;{$ENDIF}
    procedure CopyString(S:string);
    function PasteString:string;
  end;

type
  PSingleClip = ^TSingleClip;
  TSingleClip = object(TClipObject)
    constructor Init;
    destructor Done;virtual;
    procedure CopyToClp(var Buf; BufType:word; Len:word);virtual;
    procedure PasteFromClp(var Buf; BufType:word);virtual;
    procedure ClearClp;virtual;
    function DataSize(BufType:word):word;virtual;
  private
    ClipBuf:string;
  end;

const
  Clp:PClipObject = nil;

procedure InitSingleClp;
procedure DoneClipboard;

implementation
uses Ext_Math, Ext_Str;

  { TClipObject }
constructor TClipObject.Init;
begin
  inherited Init;
  Clp:=@Self;
end;

destructor TClipObject.Done;
begin
  Clp:=nil;
  inherited Done;
end;

{$IFNDEF EAM}
procedure TClipObject.CopyToClp(var Buf; BufType:word; Len:word);
begin
  abstract;
end;
procedure TClipObject.PasteFromClp(var Buf; BufType:word);
begin
  abstract;
end;
procedure TClipObject.ClearClp;
begin
  abstract;
end;
function TClipObject.DataSize(BufType:word):word;
begin
  abstract;
end;
{$ENDIF}
procedure TClipObject.CopyString(S:string);
begin
  CopyToClp(s[1], winclpOEMText, Length(S));
end;

function TClipObject.PasteString:string;
var P:Pointer;
    L:word;
    S:string;
begin
  L:=DataSize(winclpOEMText);
  GetMem(P,L);
  PasteFromClp(P^,winclpOEMText);
  SetLength(S,Min(L,255));
  move(P^,s[1],Length(s));
  PasteString:=s;
  FreeMem(P,L);
end;

  { TSingleClip }
constructor TSingleClip.Init;
begin
  inherited Init;
end;

destructor TSingleClip.Done;
begin
  inherited Done;
end;

procedure TSingleClip.CopyToClp(var Buf; BufType:word; Len:word);
begin
  {$IFDEF VER70}
  if Len>255 then len:=255;
  {$ENDIF}
  SetLength(ClipBuf, Len);
  Move(buf, ClipBuf[1], Len);
end;

procedure TSingleClip.PasteFromClp(var Buf; BufType:word);
begin
  Move(ClipBuf[1], buf, Length(ClipBuf));
end;

procedure TSingleClip.ClearClp;
begin
  FillChar(ClipBuf, SizeOf(String), 0);
end;

function TSingleClip.DataSize(BufType:word):word;
begin
  DataSize:=Length(ClipBuf);
end;

procedure InitSingleClp;
begin
  Clp:=New(PSingleClip, Init);
end;

procedure DoneClipboard;
begin
  if Clp<>nil then Clp^.Free;
  Clp:=nil;
end;

end.