unit EXT_DOCS;
interface
uses Objects, Ext_Type, Ext_Dos, Ext_Str;
type
   PFormatRecord = ^TFormatRecord;
   TFormatRecord = record
     FormType:word;
     Next:PFormatRecord;
   end;

   PParagraph = ^TParagraph;
   TParagraph = object(TObject)
     Left:word;
     Right:word;
     Abs:word;
     Format:PFormatRecord;
     Size:word;
     Text:PCharArray;
     constructor Init(S,A,l,R:word);
     destructor Done;virtual;
     constructor CopyFrom(AP:PParagraph);
   end;
type
   PDocsObjects = ^TDocsObjects;
   TDocsObjects = object(TObject)
     FileName:FNameStr;
     DocsFile:PBufStream;
     PC:PCollection;
     PCurent:integer;
     modifi:boolean;
     constructor Init(AFileName:FNameStr);
     destructor Done;virtual;
     function FirstParagraph:PParagraph;
     function NextParagraph:PParagraph;
     function GoParagraph(ANum:Longint):PParagraph;
     function EndOfParagraph:boolean;
     procedure AddParagraph(AParagraph:PParagraph);
     procedure InsertParagraph(AParagraph:PParagraph);
     procedure WriteParagraph(P:PParagraph);virtual;
     procedure WriteHeader;virtual;
     procedure WriteEnd;virtual;
   end;
type
   PText = ^TText;
   TText = object(TDocsObjects)
     procedure WriteParagraph(P:PParagraph);virtual;
   end;
type
   PRTFText = ^TRTFText;
   TRTFText = object(TDocsObjects)
     function FirstParagraph:PParagraph;virtual;
     function NextParagraph:PParagraph;virtual;
   end;

implementation

constructor TParagraph.Init(S,A,l,R:word);
begin
  inherited Init;
  Left:=l;
  Size:=S;
  Right:=R;
  Abs:=A;
  if Size<>0 then
  begin
     GetMem(Text,Size);
     FillChar(Text^,Size,0);
  end;
end;

destructor TParagraph.Done;
begin
  if Text<>nil then FreeMem(Text, Size);
  inherited Done;
end;

constructor TParagraph.CopyFrom(AP:PParagraph);
begin
  Left:=AP^.Left;
  Size:=AP^.Size;
  Right:=AP^.Right;
  Abs:=AP^.Abs;
  if Size<>0 then
  begin
     GetMem(Text,Size);
     Move( AP^.Text^, Text^,Size)
  end;
end;

constructor TDocsObjects.Init(AFileName:FNameStr);
begin
  inherited Init;
  FileName:=AFileName;
  if FExists(FileName) then
     DocsFile:=New(PBufStream,Init(FileName,stOpen,2048))
  else
     DocsFile:=New(PBufStream,Init(FileName,stCreate,2048));
  PC:=New(PCollection,Init(100,100));
end;

destructor TDocsObjects.Done;

procedure DoWrite(P:PParagraph);far;
begin
  WriteParagraph(P);
end;

begin
  if modifi then
  begin
    DocsFile^.Seek(0);
    DocsFile^.Truncate;
    WriteHeader;
    PC^.ForEach(@DoWrite);
    WriteEnd;
  end;
  PC^.FreeAll;
  Dispose(PC,Done);
  if DocsFile<>nil then DocsFile^.Free;
  inherited Done;
end;

procedure TDocsObjects.WriteParagraph(P:PParagraph);
begin
end;

procedure TDocsObjects.WriteHeader;
begin
end;

procedure TDocsObjects.WriteEnd;
begin
end;

function TDocsObjects.FirstParagraph:PParagraph;
begin
  FirstParagraph:=GoParagraph(0)
end;

function TDocsObjects.NextParagraph:PParagraph;
begin
  NextParagraph:=GoParagraph(PCurent+1)
end;

function TDocsObjects.GoParagraph(ANum:Longint):PParagraph;
begin
  if (ANum>PC^.Count-1) or (ANum<0) then exit;
  GoParagraph:=PC^.At(ANum);
  PCurent:=ANum;
end;

function TDocsObjects.EndOfParagraph:boolean;
begin
  EndOfParagraph:=PCurent=PC^.Count-1;
end;

procedure TDocsObjects.AddParagraph(AParagraph:PParagraph);
var
  P:PParagraph;
begin
  P:=New(PParagraph,CopyFrom(AParagraph));
  PC^.Insert(P);
  PCurent:=PC^.Count;
  modifi:=true;
end;

procedure TDocsObjects.InsertParagraph(AParagraph:PParagraph);
var
  P:PParagraph;
begin
  P:=New(PParagraph,CopyFrom(AParagraph));
  PC^.AtInsert(PCurent,P);
  modifi:=true;
end;

{TText}
function SkipSpase(S:PStream):integer;
var L:integer;
    C:byte;
begin
  l:=0;
  S^.Read(C,1);
  while C<33 do
  begin
    if c=8 then l:=l or 8
    else  inc(l);
    if C=13 then
    begin
      S^.Read(C,1);
      l:=0;
    end;
    S^.Read(C,1);
  end;
  S^.Seek(S^.GetPos-1);
  SkipSpase:=l;
end;

function ReadLine(S:PStream):String;
var St:String;
    C:Char;
begin
   St:='';
   S^.Read(C,1);
   while (C<>#13) and (S^.Status=stOk) do
   begin
     St:=St+C;
     S^.Read(C,1);
   end;
   if S^.Status<>stOk then S^.Reset
   else S^.Read(C,1);
   ReadLine:=St;
end;
{
function TText.FirstParagraph:PParagraph;
var
  P:PParagraph;
  Temp:PStream;
  S:string;
begin
  New(P);
  FillChar(P^,SizeOf(TParagraph),0);
  Temp:=TempStream(inMemory);
  DocsFile^.Reset;
  DocsFile^.Seek(0);
  P^.left:=SkipSpase(DocsFile);
  S:=ReadLine(DocsFile);
  Temp^.Write(S[1],byte(S[0]));
  P^.Size:=Temp^.GetSize;
  GetMem(P^.Text,P^.Size);
  Temp^.Seek(0);
  Temp^.Read(P^.Text^,P^.Size);
  Temp^.Free;
end;

function TText.NextParagraph:PParagraph;
begin
end;
}
procedure TText.WriteParagraph(P:PParagraph);
var
  S,S1:string;
  Len:byte;
  Pos:word;
  Stop:boolean;
  C,OldC:integer;
begin
  Pos:=0;
  S:=DupChar(#$20,P^.Abs);
  while Pos<P^.Size do
  begin
    if P^.Size-Pos<P^.Right then
    begin
       S1[0]:=Char(P^.Size-Pos);
       Move(P^.Text^[Pos],S1[1],Byte(S1[0]));
       S:=S+S1;
       Pos:=P^.Size;
    end
    else
    begin
      C:=0;
      repeat
         S:=S+P^.Text^[Pos];
         inc(Pos);
         Stop:=Pos>=P^.Size;
         if S[Length(S)]=#$20 then
         begin
           C:=OldC;
           OldC:=Length(S);
         end;
      until ((Length(S)>P^.Right) and (S[Length(s)]=#$20)) or Stop;
      if Length(S)>P^.Right then
      begin
        Pos:=Pos-(Length(S)-C);
        S[0]:=Char(C);
      end;
     end;
    S:=S+#13#10;
    DocsFile^.Write(S[1],Length(S));
    S:='';
  end;
end;

{TRTFText}
function TRTFText.FirstParagraph:PParagraph;
begin
end;

function TRTFText.NextParagraph:PParagraph;
begin
end;
end.
