{*******************************************************}
{                                                       }
{       Turbo Pascal 7.0                                }
{       FreePacal                                       }
{       Delphi 4                                        }
{       Printer manager                                 }
{                                                       }
{       Copyright (c) 1995-99 by Alexs                  }
{                                                       }
{*******************************************************}

{$IFNDEF FPC}
{$O+}
{$ENDIF}
{$I Define.inc}

unit Ext_Prn;

interface

uses Objects, Dos, Ext_Dos, Ext_Type, Register;

const
  pnLpt1 = 0;
  pnLpt2 = 1;
  pnLpt3 = 2;
type
  PPrinterFormatCode = ^TPrinterFormatCode;
  TPrinterFormatCode = record
  end;

type
   PPageBreak = ^TPageBreak;
   TPageBreak = record
      Next:PPageBreak;
      Pos:Longint;
   end;

   PPrinterRec = ^TPrinterRec;
   TPrinterRec = record
     Name:String[20];
     FileName:FNameStr;
     Position:Longint;
     Size:Longint;
     PrnNum:word;
     S:PBufStream;
     PageBreak:PPageBreak;
   end;

type
   PPrinterManager = ^TPrinterManager;
   TPrinterManager = object(TCollection)
     PrintFile:PBufStream;
     Complete:Boolean;
     PrinterNumber:word;
     FreeAllItem:boolean;
     Stop:boolean;
     PageBreak:PPageBreak;
     constructor Init(APrinterNumber:word);
     destructor Done;virtual;
     procedure FreeItem(Item: Pointer); virtual;
     function GetItem(var S: TStream): Pointer; virtual;
     procedure PutItem(var S: TStream; Item: Pointer); virtual;
     function OpenPrinter(AName:String):PPrinterRec;
     procedure ClosePrinter(Rec:PPrinterRec);
     procedure Write(const Rec:PPrinterRec;S:String);
     procedure WriteLn(const Rec:PPrinterRec; S:String);
     procedure WriteChar(const Rec:PPrinterRec; C:Char);
     procedure WriteInteger(const Rec:PPrinterRec; L:Longint);
     procedure WriteFloat(const Rec:PPrinterRec; R:Real);
     procedure WriteTime(const Rec:PPrinterRec; D:TDateTime);
     procedure WriteDate(const Rec:PPrinterRec; D:TDateTime);
     procedure WriteBoolean(const Rec:PPrinterRec; L:Boolean);
     procedure SetCondensed(const Rec:PPrinterRec; Enable:boolean);
     procedure SetNLQMode(const Rec:PPrinterRec; Enable:boolean);
     procedure Set216InchLF(const Rec:PPrinterRec; Margin:byte);
     procedure SetLeftMargin(const Rec:PPrinterRec; Margin:byte);
     procedure SetRightMargin(const Rec:PPrinterRec; Margin:byte);
     procedure EndPage(const Rec:PPrinterRec);
     procedure NewPage(const Rec:PPrinterRec);
     procedure ResetPrinter(const Rec:PPrinterRec);
     procedure Update;
     procedure Look;
     procedure UnLook;
   private
     LookFlag:integer;
     PrnNum:word;
   end;

const
  PrinterManager:PPrinterManager = nil;

function TestPrn(PrinterNumber:word):Boolean;
function PrintChar(PrinterNumber:word;C:byte):Boolean;

implementation

uses Ext_Math, Ext_Sys;

function PrinterError(Sender:PPrinterManager; Code:TMessegeCode; Info:Pointer; Flags:Word):word;
var C:Char;
begin
  Sender^.Look;
  PrinterError:=ErrorMessage(Code, Info, Flags);
  Sender^.UnLook;
end;

constructor TPrinterManager.Init(APrinterNumber:word);
var
  ACount,i:integer;

begin
  inherited Init(10,5);
  Complete:=true;
  PrinterNumber:=APrinterNumber;
  FreeAllItem:=True;
  if FileExists('APP_PRN.DSK') then
  begin
    PrintFile:=New(PBufStream,Init('APP_PRN.DSK',stOpen,2048));
    PrintFile^.Read(ACount, SizeOf(Integer));
    for i:=0 to ACount-1 do Insert(GetItem(PrintFile^));
    Dispose(PrintFile,Done);PrintFile:=nil;
    DeleteFile('APP_PRN.DSK');
  end;
end;

destructor TPrinterManager.Done;
var
  i:integer;
begin
  if PrintFile<>nil then Dispose(PrintFile,Done);
  if (Count<>0) and not FreeAllItem then
  begin
    PrintFile:=New(PBufStream,Init('APP_PRN.DSK',stCreate,2048));
    PrintFile^.Write(Count, SizeOf(Integer));
    for i:=0 to Count-1 do PutItem(PrintFile^, At(i));
    Dispose(PrintFile,Done);
  end
  else DeleteFile('APP_PRN.DSK');
  inherited Done;
  PrinterManager:=nil;
end;

procedure TPrinterManager.FreeItem(Item: Pointer);
var
  AItem: PPrinterRec absolute Item;
  P:Pointer;
begin
  if AItem^.S<>nil then Dispose(AItem^.S,Done);
  if FreeAllItem then DeleteFile(AItem^.FileName);
  while AItem^.PageBreak<>nil do
  begin
    P:=AItem^.PageBreak^.Next;
    Dispose(AItem^.PageBreak);
    AItem^.PageBreak:=P;
  end;
  FreeMem( AItem, SizeOf(TPrinterRec));
end;

function TPrinterManager.GetItem(var S: TStream): Pointer;
var
  AItem: PPrinterRec;
  f:boolean;
begin
  GetMem( AItem, SizeOf(TPrinterRec));
  S.Read( f, SizeOf(Boolean));
  FillChar( AItem^, SizeOf(TPrinterRec), 0);
  S.Read( Aitem^, SizeOf(TPrinterRec)-4);
  if f then AItem^.S:=New(PBufStream, Init(AItem^.FileName, stOpen, 1024));
  GetItem:=AItem;
end;

procedure TPrinterManager.PutItem(var S: TStream; Item: Pointer);
var
  AItem: PPrinterRec absolute Item;
  f:boolean;
begin
  f:=AItem^.S<>nil;
  S.Write( f, SizeOf(Boolean));
  S.Write( Aitem^, SizeOf(TPrinterRec)-4);
end;

function TPrinterManager.OpenPrinter(AName:String):PPrinterRec;
var
  AItem: PPrinterRec;
  Year, Month, Day, DayOfWeek: Word;
  PBreak:PPageBreak;
begin
  GetMem( AItem, SizeOf(TPrinterRec));
  FillChar( AItem^, SizeOf(TPrinterRec), 0);
  AItem^.Name:=AName;
  repeat
    AItem^.FileName:=TempDir+'_'+IntToStr(Random(200))+'.$$$';
  until not FileExists(AItem^.FileName);
  AItem^.S:=New(PBufStream,Init(AItem^.FileName, stCreate, 1024));
  AItem^.PrnNum:=PrinterNumber;
  OpenPrinter:=AItem;
  New(PBreak);
  PBreak^.Next:=nil;
  PBreak^.Pos:=-1;
  AItem^.PageBreak:=PBreak;
end;

procedure TPrinterManager.ClosePrinter(Rec:PPrinterRec);
begin
  Rec^.Size:=Rec^.S^.GetSize;
  Dispose(Rec^.S,Done);
  Rec^.S:=nil;
  Insert(Rec);
end;

procedure TPrinterManager.Write(const Rec:PPrinterRec;S:String);
begin
  Rec^.S^.Write(S[1], Byte(S[0]));
end;

procedure TPrinterManager.WriteLn(const Rec:PPrinterRec; S:String);
begin
  Rec^.S^.Write(S[1], Byte(S[0]));
  S:=#13#10;
  Rec^.S^.Write(S[1], Byte(S[0]));
end;

procedure TPrinterManager.WriteInteger(const Rec:PPrinterRec; L:Longint);
begin
  Write(Rec, IntToStr(L));
end;

procedure TPrinterManager.WriteFloat(const Rec:PPrinterRec; R:Real);
begin
  Write(Rec, FloatToStr(R));
end;

procedure TPrinterManager.WriteTime(const Rec:PPrinterRec; D:TDateTime);
begin
  Write(Rec, TimeToStr(D));
end;

procedure TPrinterManager.WriteDate(const Rec:PPrinterRec; D:TDateTime);
begin
  Write(Rec, DateToStr(D));
end;

procedure TPrinterManager.WriteBoolean(const Rec:PPrinterRec; L:Boolean);
begin
  Write(Rec, BooleanToStr(L));
end;

procedure TPrinterManager.WriteChar(const Rec:PPrinterRec; C:Char);
begin
  Rec^.S^.Write(C, SizeOf(Char));
end;

procedure TPrinterManager.SetCondensed(const Rec:PPrinterRec; Enable:boolean);
begin
  if Enable then write(Rec, #$1B#$0F) else write(Rec, #$12);
end;

procedure TPrinterManager.SetNLQMode(const Rec:PPrinterRec; Enable:boolean);
begin
  write(Rec,#$1B#$78+Char(Enable));
end;

procedure TPrinterManager.Set216InchLF(const Rec:PPrinterRec; Margin:byte);
begin
  write(Rec, #$1B#$33+Char(Margin));
end;

procedure TPrinterManager.SetLeftMargin(const Rec:PPrinterRec; Margin:byte);
begin
  write(Rec, #$1B#$6C+Char(Margin));
end;

procedure TPrinterManager.SetRightMargin(const Rec:PPrinterRec; Margin:byte);
begin
  write(Rec, #$1B#$51+Char(Margin));
end;

procedure TPrinterManager.NewPage(const Rec:PPrinterRec);
begin
end;

procedure TPrinterManager.EndPage(const Rec:PPrinterRec);
var
  P,PBreak:PPageBreak;
begin
  WriteLn(Rec, #12);
  new(PBreak);
  PBreak^.Next:=nil;
  PBreak^.Pos:=Rec^.S^.GetPos;
  if Rec^.PageBreak=nil then Rec^.PageBreak:=PBreak
  else
  begin
    P:=Rec^.PageBreak;
    while (P^.Next<>nil) do P:=P^.Next;
    P^.Next:=PBreak
  end;
end;

procedure TPrinterManager.ResetPrinter(const Rec:PPrinterRec);
begin
  write(Rec, #$1B#$40);
end;

procedure TPrinterManager.Look;
begin
  inc(LookFlag);
end;

procedure TPrinterManager.UnLook;
begin
  Dec(LookFlag);
end;

procedure TPrinterManager.Update;
var
  Item:PPrinterRec;
  C:byte;
begin
  if Stop then exit;
  if (Count=0) or (LookFlag<>0) then exit;
  if not TestPrn(PrinterNumber) then
     if PrinterError(@Self, mcPrinterNotReady,
        nil, mfError+MfOkCancel)<>cmOk then Complete:=true
     else exit;
  if Complete then
     begin
       if PrintFile<>nil then
       begin
         Dispose(PrintFile,Done);
         PrintFile:=nil;
         AtFree(0);
       end;
       if Count=0 then exit;
       Complete:=false;
       Item:=At(0);
       PrnNum:=Item^.PrnNum;
       PrintFile:=New(PBufStream,Init(Item^.FileName, stOpenRead, 1024));
       PageBreak:=Item^.PageBreak;
     end
  else
  begin
    if PageBreak<>nil then
       if PageBreak^.Pos=PrintFile^.GetPos then
       begin
          PrinterError(@Self, mcPrinterNewPage, nil, mfConfirmation+mfOkButton);
          Stop:=true;
          PageBreak:=PageBreak^.Next;
          exit;
       end;
    PrintFile^.Read(C,SizeOf(Byte));
    Complete:= PrintFile^.GetSize=PrintFile^.GetPos;
    if not PrintChar(PrnNum,C) then
    begin
      PrintFile^.Seek(PrintFile^.GetPos-1);
      if PrinterError(@Self, mcPrinterNotReady,
         nil, mfError+MfOkCancel)<>cmOk then Stop:=true
      else exit;
    end;
  end;
end;

function TestPrn(PrinterNumber:word):Boolean;
{$IFDEF DOS}
var
  R:Registers;
begin
  R.DX:=PrinterNumber;
  R.AH:=02;
  Intr($17, R);
  TestPrn:= (R.AH and $8) <> $8;
end;
{$ELSE}
begin
end;
{$ENDIF}


function PrintChar(PrinterNumber:word;C:byte):Boolean;
{$IFDEF DOS}
var
  R:Registers;
begin
  R.DX:=PrinterNumber;
  R.AH:=0;
  R.Al:=C;
  Intr($17, R);
  PrintChar:= (R.AH and $8) <> $8;
end;
{$ELSE}
begin
end;
{$ENDIF}
end.
