{************************************************}
{                                                }
{   Extended system functions                    }
{                                                }
{   Turbo Pascal 7                               }
{   FreePacal                                    }
{   Delphi 2 - 5                                 }
{   Virtual Pascal 2.1                           }
{                                                }
{   Copyright (c) 1995-2000 by Alexs             }
{                                                }
{************************************************}


{$IFDEF VER70}
{$O-}
{$ENDIF}
{$I Define.inc}

unit EXT_SYS;

interface

uses Ext_Type, Register, Std_Str;

type
  TMessegeCode = (mcInvalidCharacter, mcInvalidPicture, mcInvalidValue,
                  mcInvalidFValue, mcNotInList, mcInvalidAdres,
                  mcNotInListParam, mcTooManyFiles, mcInvalidDrive,
                  mcInvalidFileName, mcDataError, mcErrorString,
                  mcSaveModify, mcSaveUntitled, mcPrinterNotReady,
                  mcPrinterNewPage);

type
  TErorrMessage = function(MessageCode:TMessegeCode; Info:Pointer; Flags:word):word;

{misk routins}
procedure Error(Message:string);
procedure PrintStr(const S: String);
procedure PrintStrLn(const S: String);


function StdErrorString(MessageCode:TMessegeCode):string;
function DefErrorMessage(MessageCode:TMessegeCode; Info:Pointer; Flags:word):word;{$IFNDEF FPC}far;{$ENDIF}

const
  ErrorMessage:TErorrMessage = DefErrorMessage;

const
  LongMonthNames: array [1..12] of string[10] =
  ('','ࠫ','',    '५', '',   '',
   '',  '', '','','','');
const
  ShortMonthNames: array[1..12] of string[10] =
  ('ﭢ','䥢ࠫ','',    '५', '',   '',
   '',  '', 'ᥭ','','','');

const
  ShortDayNames: array [1..7] of string[3] =
  ('','','','','','','');
const
  LongDayNames : array [1..7] of string[11] =
  ('쭨','୨','।','⢥',
   '⭨','㡮','ᥭ');
const
  ShortNameDayOfW{:string[20]}='      ';
const
   DaysInMonth: array[1..12] of Byte =
     (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);

const
  DateSeparator:char = '/';
  TimeSeparator:char = ':';
  ExpandDate:boolean = true;
  CurrencyString:string[5] = '.';
  ThousandSeparator:Char = ' ';
  DecimalSeparator:char = '.';
  CurrencyDecimals:byte = 2;
  ShortTimeFormat:string[20] = 'hh:mm';
  LongTimeFormat :string[20] = 'hh:mm:ss';
  ShortDateFormat:string[20] = 'd/mm/yy';
  LongDateFormat :string[20] = 'dd mm yyyy';
  TwoDigitYearCenturyWindow: Word = 30;

const
  BooleanText:array [boolean] of string[10] = ('false','true');

const
  ButtonName: array[0..4] of string[6] =
    (strYes, strNo, strOk, strCancel, strHelp);

const
  Commands: array[0..3] of word =
    (cmYes, cmNo, cmOK, cmCancel);

const
  Titles: array[0..3] of string[14] =
    (strWarning,strError,strInformation,strConfirmation);

const
  DrivesS: String[10] = strDrives;

implementation
uses Ext_Str
{$IFDEF COMPACT_EXE}
, Drivers
{$ENDIF COMPACT_EXE};

procedure PrintStrLn(const S: String);
begin
  PrintStr(S+#13#10);
end;

{$IFDEF VER70}
procedure PrintStr(const S: String); assembler;
asm
        PUSH    DS
        LDS     SI,S
        CLD
        LODSB
        XOR     AH,AH
        XCHG    AX,CX
        MOV     AH,40H
        MOV     BX,1
        MOV     DX,SI
        INT     21H
        POP     DS
end;
{$ELSE}
procedure PrintStr(const S: String);
begin
  writeln(S);
end;
{$ENDIF}
procedure Error(Message:string);
begin
  PrintStr(Message+#13#10);
  Halt;
end;

function StdErrorString(MessageCode:TMessegeCode):string;
begin
  case MessageCode of
    mcInvalidCharacter : StdErrorString:=strInvalidCharacter;
    mcInvalidPicture   : StdErrorString:=strInvalidPicture;
    mcInvalidValue     : StdErrorString:=strInvalidValue;
    mcInvalidFValue    : StdErrorString:=strInvalidFValue;
    mcNotInList        : StdErrorString:=strNotInList;
    mcInvalidAdres     : StdErrorString:=strInvalidAdres;
    mcNotInListParam   : StdErrorString:=strNotInListParam;
    mcTooManyFiles     : StdErrorString:=strTooManyFiles;
    mcInvalidDrive     : StdErrorString:=strInvalidDrive;
    mcInvalidFileName  : StdErrorString:=strInvalidFileName;
    mcDataError        : StdErrorString:=strDataError;
    mcSaveModify       : StdErrorString:=strSaveModify;
    mcPrinterNotReady  : StdErrorString:=strPrinterNotReady;
    mcPrinterNewPage   : StdErrorString:=strPrinterNewPage;
  else
    StdErrorString:='';
  end;
end;

function DefErrorMessage(MessageCode:TMessegeCode; Info:Pointer; Flags:word):word;
begin
  DefErrorMessage:=cmCancel;
end;

(*
{$IFDEF COMPACT_EXE}
procedure ExitApplication;far;
var
  pm:array[0..2] of longint;
  s:string;
begin
  if ExitCode <> 0 then
  begin
     if ExitCode = 203 then
        PrintStr('Error: Out of memory.')
     else
     begin
       pm[0]:=ExitCode;
       pm[1]:=PtrRec(ErrorAddr).Seg;
       pm[2]:=PtrRec(ErrorAddr).Ofs;
       FormatStr(S,'Error %d at %04x:%04x. Report to Alexs (㭮 ..).', Pm );
       PrintStr(s);
     end;
  end;
  ErrorAddr:=nil;
  ExitCode:=0;
end;
begin
  ExitProc:=@ExitApplication;
{$ENDIF}
*)
end.