{************************************************}
{                                                }
{   Turbo Vision Unit                            }
{   Copyright (c) 1992 Borland International     }
{                                                }
{   Turbo Pascal 7                               }
{   FreePacal                                    }
{   Delphi 2 - 5                                 }
{   Virtual Pascal 2.1                           }
{                                                }
{   Copyright (c) 1995-2000 by Alexs             }
{                                                }
{************************************************}

{$IFDEF VER70}
{$O+}
{$ENDIF}

{$I Define.inc}

unit HistList;

{****************************************************************************
   History buffer structure:

    Byte Byte String          Byte Byte String
    +-------------------------+-------------------------+--...--+
    | 0 | Id | History string | 0 | Id | History string |       |
    +-------------------------+-------------------------+--...--+

 ***************************************************************************}

interface

uses Objects;

const
  HistoryBlock: Pointer = nil;
  HistorySize: Word = 1024;
  HistoryUsed: Word = 0;

procedure HistoryAdd(Id: Byte; const Str: String);
function HistoryCount(Id: Byte): Word;
function HistoryStr(Id: Byte; Index: Integer): String;
procedure ClearHistory;

procedure InitHistory;
procedure DoneHistory;

procedure StoreHistory(var S: TStream);
procedure LoadHistory(var S: TStream);
{$IFNDEF VER70}
function HistoryRemove(Id: Byte; Index: Integer): boolean;
{$ENDIF}

implementation

var
  CurId: Byte;
  CurString: PString;

procedure StartId(Id: Byte);{$IFDEF VER70} near;{$ENDIF}
begin
  CurId := Id;
  CurString := HistoryBlock;
end;

{$IFDEF VER70}
{ Advance CurString to next string with an ID of CurId }

procedure AdvanceStringPointer; near; assembler;
asm
        PUSH    DS
        MOV     CX,HistoryUsed
        MOV     BL,CurId
        LDS     SI,CurString
        MOV     DX,DS
        MOV     AX,DS
        OR      AX,SI
        JZ      @@3
        CLD
        JMP     @@2
@@1:    LODSW
        CMP     AH,BL { BL = CurId }
        JE      @@3
@@2:    LODSB
        XOR     AH,AH
        ADD     SI,AX
        CMP     SI,CX { CX = HistoryUsed }
        JB      @@1
        XOR     SI,SI
        MOV     DX,SI
@@3:    POP     DS
        MOV     CurString.Word[0],SI
        MOV     CurString.Word[2],DX
end;

{ Deletes the current string from the table }

procedure DeleteString; near; assembler;
asm
        PUSH    DS
        MOV     CX,HistoryUsed
        CLD
        LES     DI,CurString
        MOV     SI,DI
        DEC     DI
        DEC     DI
        PUSH    ES
        POP     DS
        MOV     AL,BYTE PTR [SI]
        XOR     AH,AH
        INC     AX
        ADD     SI,AX
        SUB     CX,SI
        REP     MOVSB
        POP     DS
        MOV     HistoryUsed,DI
end;

{ Insert a string into the table }

procedure InsertString(Id: Byte; const Str: String); near; assembler;
asm
        PUSH    DS
        STD

        { Position ES:DI to the end the buffer  }
        {          ES:DX to beginning of buffer }
        LES     DX,HistoryBlock
        MOV     DI,HistoryUsed
        LDS     SI,Str
        MOV     BL,[SI]
        XOR     BH,BH
        INC     BX
        INC     BX
        INC     BX
        POP     DS
        PUSH    DS
@@1:    MOV     AX,DI
        ADD     AX,BX
        SUB     AX,DX { DX = HistoryBlock.Word[0] }
        CMP     AX,HistorySize
        JB      @@2

        { Drop the last string off the end of the list }
        DEC     DI
        XOR     AL,AL
        MOV     CX,0FFFFH
        REPNE   SCASB
        INC     DI
        JMP     @@1

        { Move the table down the size of the string }
@@2:    MOV     SI,DI
        ADD     DI,BX
        MOV     HistoryUsed,DI
        PUSH    ES
        POP     DS
        MOV     CX,SI
        SUB     CX,DX { DX = HistoryBlock.Word[0] }
        REP     MOVSB

        { Copy the string into the position }
        CLD
        MOV     DI,DX { DX = HistoryBlock.Word[0] }
        INC     DI
        MOV     AH,Id
        XOR     AL,AL
        STOSW
        LDS     SI,Str
        LODSB
        STOSB
        MOV     CL,AL
        XOR     CH,CH
        REP     MOVSB

        POP     DS
end;


function HistoryCount(Id: Byte): Word;
var
  Count: Word;
begin
  StartId(Id);
  Count := 0;
  AdvanceStringPointer;
  while CurString <> nil do
  begin
    Inc(Count);
    AdvanceStringPointer;
  end;
  HistoryCount := Count;
end;

procedure HistoryAdd(Id: Byte; const Str: String);
begin
  if Str = '' then Exit;

  StartId(Id);

  { Delete duplicates }
  AdvanceStringPointer;
  while CurString <> nil do
  begin
    if Str = CurString^ then DeleteString;
    AdvanceStringPointer;
  end;

  InsertString(Id, Str);
end;

function HistoryStr(Id: Byte; Index: Integer): String;
var
  I: Integer;
begin
  StartId(Id);
  for I := 0 to Index do AdvanceStringPointer;
  if CurString <> nil then
    HistoryStr := CurString^ else
    HistoryStr := '';
end;

procedure ClearHistory;
begin
  PChar(HistoryBlock)^ := #0;
  HistoryUsed := PtrRec(HistoryBlock).Ofs + 1;
end;

procedure StoreHistory(var S: TStream);
var
  Size: Word;
begin
  Size := HistoryUsed - PtrRec(HistoryBlock).Ofs;
  S.Write(Size, SizeOf(Word));
  S.Write(HistoryBlock^, Size);
end;

procedure LoadHistory(var S: TStream);
var
  Size: Word;
begin
  S.Read(Size, SizeOf(Word));
  S.Read(HistoryBlock^, Size);
  HistoryUsed := PtrRec(HistoryBlock).Ofs + Size;
end;

{$ELSE}

type
  PByte=^Byte;
  THistRec=packed record
    Zero : byte;
    Id   : byte;
    Str  : String;
  end;
  PHistRec=^THistRec;

{ Advance CurString to next string with an ID of CurId }
procedure AdvanceStringPointer;
var
  p : PHistRec;
begin
  while (CurString<>nil) do
  begin
    inc(pchar(CurString),pbyte(CurString)^+1);
    if pchar(CurString)-pchar(HistoryBlock)>=HistoryUsed then
    begin
      CurString:=nil;
      exit;
    end;
    p:=PHistRec(CurString);
    inc(pchar(CurString),2);
    if (p^.Id=CurId) then  exit;
  end;
end;


procedure DeleteString;
var
  len  : integer;
  p,p2 : pchar;
begin
  p:=pchar(CurString);
  p2:=pchar(CurString);
  len:=pbyte(p2)^+3;
  dec(p,2);
  inc(p2,pbyte(p2)^+1);
  Move(p2^,p^,HistoryUsed-(p2-pchar(HistoryBlock)));
  dec(HistoryUsed,len);
end;


procedure InsertString(Id: Byte; const Str: String);
var
  p1,p2 : pchar;
begin
  while (HistoryUsed+Length(Str)>HistorySize) do
   begin
     runerror(199);
   end;
  p1:=pchar(HistoryBlock)+1;
  p2:=p1+Length(Str)+3;
  Move(p1^,p2^,HistoryUsed-1);
  PHistRec(p1)^.Zero:=0;
  PHistRec(p1)^.Id:=Id;
  PHistRec(p1)^.Str:=Str;
{  Move(Str[0],PHistRec(p1)^.Str,Length(Str)+1);}
  inc(HistoryUsed,Length(Str)+3);
end;

function HistoryCount(Id: Byte): Word;
var
  Count: Word;
begin
  StartId(Id);
  Count := 0;
  AdvanceStringPointer;
  while CurString <> nil do
  begin
    Inc(Count);
    AdvanceStringPointer;
  end;
  HistoryCount := Count;
end;


procedure HistoryAdd(Id: Byte; const Str: String);
begin
  if Str = '' then Exit;
  StartId(Id);
  { Delete duplicates }
  AdvanceStringPointer;
  while CurString <> nil do
  begin
    if Str = CurString^ then
       DeleteString;
     AdvanceStringPointer;
   end;

  InsertString(Id, Str);
end;

function HistoryStr(Id: Byte; Index: Integer): String;
var
  I: Integer;
begin
  StartId(Id);
  for I := 0 to Index do
   AdvanceStringPointer;
  if CurString <> nil then
    HistoryStr := CurString^
  else
    HistoryStr := '';
end;

function HistoryRemove(Id: Byte; Index: Integer): boolean;
var
  I: Integer;
begin
  StartId(Id);
  for I := 0 to Index do
   AdvanceStringPointer;
  if CurString <> nil then
    begin
       DeleteString;
       HistoryRemove:=true;
    end
  else
    HistoryRemove:=false;
end;

procedure ClearHistory;
begin
  PChar(HistoryBlock)^ := #0;
  HistoryUsed := 1;
end;


procedure StoreHistory(var S: TStream);
var
  Size: Word;
begin
  Size := HistoryUsed;
  S.Write(Size, SizeOf(Word));
  S.Write(HistoryBlock^, Size);
end;


procedure LoadHistory(var S: TStream);
var
  Size: Word;
begin
  S.Read(Size, SizeOf(Word));
  S.Read(HistoryBlock^, Size);
  HistoryUsed := Size;
end;

{$ENDIF}
procedure InitHistory;
begin
  GetMem(HistoryBlock, HistorySize);
  ClearHistory;
end;

procedure DoneHistory;
begin
  FreeMem(HistoryBlock, HistorySize);
end;

end.
