{************************************************}
{                                                }
{   Virtual memory manager                       }
{                                                }
{   Turbo Pascal 7                               }
{   FreePacal                                    }
{   Delphi 2 - 5                                 }
{   Virtual Pascal 2.1                           }
{                                                }
{   Copyright (c) 1995-2000 by Alexs             }
{                                                }
{************************************************}

{$IFDEF VER70}
{$O+}
{$ENDIF}
{$I Define.inc}

unit VMM;

interface

uses Objects, Ext_Type;

type
  TPtrHandle = Longint;
type
  PVMMPage = ^TVMMPage;
  TVMMPage = object(TObject)
     Buffer:PByteArray;
     BufferSize:word;
     Handle:TPtrHandle;
     Modified:boolean;
     LockCount:integer;
     constructor Init(AHandle:Longint; BufSize:word; isNew:Boolean);
     destructor Done;virtual;
     procedure Lock;
     procedure UnLock;
  private
     isVMMPage:boolean;
  end;

{$IFDEF VER70}
type
  PSysPageFileStream = ^TSysPageFileStream;
  TSysPageFileStream = object(TStream)
    PageList: PLongArray;
    PageListSize: Integer;
    PageSize: Integer;
    Size: Longint;

    Position: Longint;
    DeltaPage:integer;
    constructor Init(APageSize: Word; ADeltaPage:integer);
    destructor Done; virtual;
    function GetPos: Longint; virtual;
    function GetSize: Longint; virtual;
    procedure Read(var Buf; Count: Word); virtual;
    procedure Seek(Pos: Longint); virtual;
    procedure Truncate; virtual;
    procedure Write(var Buf; Count: Word); virtual;
  private
    function ChangeListSize(NewSize: integer): Boolean;
  end;
{$ENDIF}

const
  defVMMPageSize   = $2000;

const
  isMultiUser:Boolean = false;
  PrevSucces:Boolean = true;
  Nodes:PCollection = nil;
  TempStream:PStream = nil;
  MaxPagesStackSize:integer = 40; {   㠫 ࠭}
const
  VMMFistFreePage:Longint = 0;


function GetVMMPage(var AHandle:TPtrHandle; ASize:word; var Page:PVMMPage):Pointer;
function UpdateVMMPage(AHandle:TPtrHandle; var Page:PVMMPage):Pointer;
procedure FreeVMMPage(AHandle:TPtrHandle);

procedure ReorderPagesForPage;

procedure InitVMM;
procedure DoneVMM;
procedure SetLimit(ALimit:integer);

implementation
uses Ext_Dos, Ext_Math;

procedure ReorderPagesForPage;
var i:integer;
    Page:PVMMPage;
begin
   i:=Nodes^.Count;
   repeat
     Dec(i);
     Page:=Nodes^.At(i);
   until (Page^.LockCount=0) or (I=0);
   if I=0 then i:=Nodes^.Count-1;
   Nodes^.AtFree(i);
end;

function GetVMMPage(var AHandle:TPtrHandle; ASize:word; var Page:PVMMPage):Pointer;
var L,LL, TempL:Longint;
    S:word;
begin
  if VMMFistFreePage<>0 then
  begin
    LL:=VMMFistFreePage;
    S:=$FFFF;
    L:=-1;
    TempL:=L;
    while (S<>ASize) and (LL<>0) do
    begin
      TempL:=L;
      L:=LL;
      TempStream^.Seek(L);
      TempStream^.Read(S, SizeOf(Word));
      TempStream^.Read(LL,SizeOf(Longint));
    end;
    if S=ASize then
    begin
      AHandle:=L;
      if TempL<>-1 then
      begin
        TempStream^.Seek(TempL+SizeOf(Word));
        TempStream^.Write(LL, SizeOf(Longint));
      end
      else VMMFistFreePage:=LL;
    end
    else AHandle:=TempStream^.GetSize;
  end
  else AHandle:=TempStream^.GetSize;
  Page:=New(PVMMPage, Init(AHandle, ASize, true));
  if Nodes^.Count=Nodes^.Limit then
     ReorderPagesForPage;
  Nodes^.AtInsert(0, Page);
  GetVMMPage:=Page^.Buffer;
end;

function FindPage(AHandle:TPtrHandle):PVMMPage;
function DoFind(P:PVMMPage):boolean;
begin
  DoFind:=P^.isVMMPage and (AHandle=P^.Handle);
end;
var i:integer;
begin
  FindPage:=nil;
  for i:=0 to Nodes^.Count-1 do
    if DoFind(Nodes^.At(i)) then
    begin
      FindPage:=Nodes^.At(i);
      break;
    end;
end;

function UpdateVMMPage(AHandle:TPtrHandle; var Page:PVMMPage):Pointer;
begin
  Page:=FindPage(AHandle);
  if Page=nil then
  begin
     if Nodes^.Count=Nodes^.Limit then  ReorderPagesForPage;
     Page:=New(PVMMPage, Init(AHandle, 0, false));
  end
  else Nodes^.Delete(Page);
  Nodes^.AtInsert(0, Page);
  UpdateVMMPage:=Page^.Buffer;
end;

{type
  TNextPageRec = record
    Size:word;
    NextPage:Longint;
  end;}

procedure FreeVMMPage(AHandle:TPtrHandle);
var Page:PVMMPage;
begin
  Page:=FindPage(AHandle);
  if Page<>nil then
  begin
    Nodes^.Delete(Page);
    Page^.Free;
  end;
  TempStream^.Seek(AHandle+SizeOf(Word));
  TempStream^.Write(VMMFistFreePage, SizeOf(Longint));
  VMMFistFreePage:=AHandle;
end;

constructor TVMMPage.Init(AHandle:Longint; BufSize:word; isNew:Boolean);
begin
  TObject.Init;
  Handle:=AHandle;
  TempStream^.Seek(Handle);
  if isNew then
  begin
    BufferSize:=BufSize;
    GetMem(Buffer, BufSize);
    FillChar(Buffer^, BufferSize, 0);
    TempStream^.Write(BufferSize, SizeOf(word));
    TempStream^.Write(Buffer^, BufferSize);
  end
  else
  begin
    TempStream^.Read(BufferSize, SizeOf(word));
    GetMem(Buffer, BufferSize);
    TempStream^.Read(Buffer^, BufferSize);
  end;
  isVMMPage:=true;
  Modified:=true;
end;

destructor TVMMPage.Done;
begin
  TempStream^.Seek(Handle);
  TempStream^.Write(BufferSize, SizeOf(word));
  TempStream^.Write(Buffer^, BufferSize);
  FreeMem(Buffer, BufferSize);
  TObject.Done;
end;

procedure TVMMPage.Lock;
begin
  inc(LockCount);
end;

procedure TVMMPage.UnLock;
begin
  if LockCount>0 then Dec(LockCount);
  if LockCount=0 then
     if Nodes^.IndexOf(@Self)<>-1 then
     begin
       Nodes^.Delete(@Self);
       Nodes^.AtInsert(Nodes^.Count,@Self)
     end;
end;

{$IFDEF VER70}
type
  THeapProc = function(Size: Word):Integer;

const
  SaveHeapError:pointer = nil;

function HeapFunc(Size: Word): Integer; far;
var i,c:integer;
    Page:PVMMPage;
    ASize:Longint;
begin
  if Size=0 then
  begin
    HeapFunc:=0;
    exit;
  end;
  ASize:=0;
  C:=Nodes^.Count-1;
  for i:=c downto 0 do
  begin
    Page:=Nodes^.At(i);
    if (Page^.LockCount=0) then
    begin
      inc(ASize, Page^.BufferSize);
      Nodes^.AtFree(i);
      if ASize>=Size then break;
    end;
  end;
  if ASize>=Size then HeapFunc:=2
  else HeapFunc:=THeapProc(SaveHeapError)(Size)
end;
{$ENDIF}

procedure InitVMM;
var S:FNameStr;
begin
{$IFDEF VER70}
  SaveHeapError:=HeapError;
  HeapError:=@HeapFunc;
{$ENDIF}
  if Nodes=nil then
     Nodes:=New(PCollection, Init(MaxPagesStackSize, 0));

  if TempStream=nil then
  begin
    if isMultiUser then TempStream:=GetTempStream
    else
    begin
      S:=TempDir+ExeName+'.tmp';
      PrevSucces:=not FileExists(S);
      TempStream:=New(PTempStream, Init(S));
    end;
  end;
end;

procedure DoneVMM;
begin
  if Nodes<>nil then
  begin
    Dispose(Nodes, Done);
    Nodes:=nil;
  end;
  if TempStream<>nil then
  begin
    TempStream^.Free;
    TempStream:=nil;
  end;
{$IFDEF VER70}
  HeapError:=SaveHeapError;
{$ENDIF}
end;

procedure SetLimit(ALimit:integer);
begin
  Nodes^.SetLimit(Max(ALimit, 20));
end;

{$IFDEF VER70}
  { TSysPageFileStream }
constructor TSysPageFileStream.Init(APageSize: Word; ADeltaPage:integer);
begin
  inherited Init;
  PageSize:=APageSize;
  PageListSize:=$100;
  DeltaPage:=ADeltaPage;
  GetMem(PageList, PageListSize*SizeOf(Longint));
  FillChar( PageList^, PageListSize*SizeOf(Longint), $FF);
end;

destructor TSysPageFileStream.Done;
begin
  FreeMem(PageList, PageListSize);
  inherited Done;
end;

function TSysPageFileStream.GetPos: Longint;
begin
  if Status=stOk then GetPos:=Position
  else GetPos:=-1;
end;

function TSysPageFileStream.GetSize: Longint;
begin
  if Status=stOk then GetSize:=Size
  else GetSize:=-1;
end;

procedure TSysPageFileStream.Read(var Buf; Count: Word);
begin
  if Count+Position>Size then
  begin
  end
  else
  begin
    Status:=stReadError;
    ErrorInfo:=0;
    Error(Status, ErrorInfo);
  end;
end;

procedure TSysPageFileStream.Seek(Pos: Longint);
begin
  if Position>=Pos then Position:=Pos
  else
  begin
    Status:=stError;
    ErrorInfo:=0;
    Error(Status, ErrorInfo);
  end;
end;

procedure TSysPageFileStream.Truncate;
begin
  Size:=Position;
end;

procedure TSysPageFileStream.Write(var Buf; Count: Word);
var P:PByteArray;
    I,j:word;
    L:Longint;
    Page:PVMMPage;
begin
  while Count>0 do
  begin
    i:=Position div PageSize;
    j:=Position mod PageSize;
    L:=PageList^[i];
    if L=-1 then
    begin
      if I>PageListSize then
         ChangeListSize(DeltaPage);
      P:=GetVMMPage(PageList^[i], PageSize, Page)
    end
    else P:=UpdateVMMPage(L, Page);
    move(Buf, P^[PageSize-j], Min(PageSize-j, Count));
    inc(Position, Min(PageSize-j, Count));
    Count:=Count-(PageSize-j);
    j:=0;
  end;
end;

function TSysPageFileStream.ChangeListSize(NewSize: integer): Boolean;
var P:pointer;
begin
  GetMem(P, NewSize);
  Move(PageList^, P^, min(PageListSize,NewSize));
  FreeMem(PageList, PageListSize);
  PageList:=P;
  PageListSize:=NewSize;
end;
{$ENDIF}

end.